// ToolGood.Words.Translate.js
// 2020, Lin Zhijun, https://github.com/toolgood/ToolGood.Words
// Licensed under the Apache License 2.0
function Translate() {
    function WordsSearch() {
        function TrieNode() {
            this.Index = 0;
            this.Layer = 0;
            this.End = false;
            this.Char = '';
            this.Results = [];
            this.m_values = {};
            this.Failure = null;
            this.Parent = null;

            this.Add = function (c) {
                if (this.m_values[c] != null) {
                    return this.m_values[c];
                }
                var node = new TrieNode();
                node.Parent = this;
                node.Char = c;
                this.m_values[c] = node;
                return node;
            }

            this.SetResults = function (index) {
                if (this.End == false) {
                    this.End = true;
                }
                this.Results.push(index)
            }
        }
        function TrieNode2() {
            this.End = false;
            this.Results = [];
            this.m_values = {};
            this.minflag = 0xffff;
            this.maxflag = 0;

            this.Add = function (c, node3) {
                if (this.minflag > c) { this.minflag = c; }
                if (this.maxflag < c) { this.maxflag = c; }
                this.m_values[c] = node3;
            }
            this.SetResults = function (index) {
                if (this.End == false) {
                    this.End = true;
                }
                if (this.Results.indexOf(index) == -1) {
                    this.Results.push(index);
                }
            }
            this.HasKey = function (c) {
                return this.m_values[c] != undefined;
            }
            this.TryGetValue = function (c) {
                if (this.minflag <= c && this.maxflag >= c) {
                    return this.m_values[c];
                }
                return null;
            }
        }

        var _first = [];
        var _keywords = [];
        this._others = [];

        this.SetKeywords = function (keywords) {
            _keywords = keywords;
            var root = new TrieNode();

            var allNodeLayer = {};
            for (var i = 0; i < _keywords.length; i++) {
                var p = _keywords[i];
                var nd = root;
                for (var j = 0; j < p.length; j++) {
                    nd = nd.Add(p.charCodeAt(j));
                    if (nd.Layer == 0) {
                        nd.Layer = j + 1;
                        if (allNodeLayer[nd.Layer]) {
                            allNodeLayer[nd.Layer].push(nd)
                        } else {
                            allNodeLayer[nd.Layer] = [];
                            allNodeLayer[nd.Layer].push(nd)
                        }
                    }
                }
                nd.SetResults(i);
            }

            var allNode = [];
            allNode.push(root);
            for (var key in allNodeLayer) {
                var nds = allNodeLayer[key];
                for (var i = 0; i < nds.length; i++) {
                    allNode.push(nds[i]);
                }
            }
            allNodeLayer = null;

            for (var i = 1; i < allNode.length; i++) {
                var nd = allNode[i];
                nd.Index = i;
                var r = nd.Parent.Failure;
                var c = nd.Char;
                while (r != null && !r.m_values[c])
                    r = r.Failure;
                if (r == null)
                    nd.Failure = root;
                else {
                    nd.Failure = r.m_values[c];
                    for (const key2 in nd.Failure.Results) {
                        if (nd.Failure.Results.hasOwnProperty(key2) == false) { continue; }
                        var result = nd.Failure.Results[key2];
                        nd.SetResults(result);
                    }
                }
            }
            root.Failure = root;

            var allNode2 = [];
            for (var i = 0; i < allNode.length; i++) {
                allNode2.push(new TrieNode2());
            }
            for (var i = 0; i < allNode2.length; i++) {
                var oldNode = allNode[i];
                var newNode = allNode2[i];

                for (const key in oldNode.m_values) {
                    if (oldNode.m_values.hasOwnProperty(key) == false) { continue; }
                    var index = oldNode.m_values[key].Index;
                    newNode.Add(key, allNode2[index]);
                }
                for (let index = 0; index < oldNode.Results.length; index++) {
                    const item = oldNode.Results[index];
                    newNode.SetResults(item);
                }

                if (oldNode.Failure != root) {
                    for (const key in oldNode.Failure.m_values) {
                        if (oldNode.Failure.m_values.hasOwnProperty(key) == false) { continue; }
                        if (newNode.HasKey(key) == false) {
                            var index = oldNode.Failure.m_values[key].Index;
                            newNode.Add(key, allNode2[index]);
                        }
                    }
                    for (let index = 0; index < oldNode.Failure.Results.length; index++) {
                        const item = oldNode.Failure.Results[index];
                        newNode.SetResults(item);
                    }
                }
            }
            allNode = null;
            root = null;

            var first = [];
            for (let index = 0; index < 0xffff; index++) {
                first.push(null);
            }
            for (const key in allNode2[0].m_values) {
                if (allNode2[0].m_values.hasOwnProperty(key) == false) { continue; }
                first[key] = allNode2[0].m_values[key];
            }
            _first = first;
        }

        this.FindAll = function (text) {
            var ptr = null;
            var list = [];

            for (let i = 0; i < text.length; i++) {
                const t = text.charCodeAt(i);
                var tn = null;
                if (ptr == null) {
                    tn = _first[t];
                } else {
                    tn = ptr.TryGetValue(t);
                    if (!tn) {
                        tn = _first[t];
                    }
                }
                if (tn != null) {
                    if (tn.End) {
                        for (let j = 0; j < tn.Results.length; j++) {
                            const item = tn.Results[j];
                            var keyword = _keywords[item];
                            list.push({
                                Keyword: keyword,
                                Success: true,
                                End: i,
                                Start: i + 1 - _keywords[item].length,
                                Index: item,
                            });
                        }
                    }
                }
                ptr = tn;
            }
            return list;
        }
    }
    //---------------------------
    var _s2t_s = ["㐷", "㐹", "㐽", "㑇", "㑈", "㑔", "㑩", "㓆", "㓥", "㓰", "㔉", "㖊", "㖞", "㗷", "㘎", "㚯", "㛀", "㛟", "㛠", "㛣", "㛤", "㛿", "㟆", "㟜", "㟥", "㡎", "㤘", "㤽", "㥪", "㧏", "㧐", "㧑", "㧟", "㧰", "㨫", "㭎", "㭏", "㭣", "㭤", "㭴", "㱩", "㱮", "㲿", "㳔", "㳕", "㳠", "㳡", "㳢", "㳽", "㴋", "㶉", "㶶", "㶽", "㺍", "㻅", "㻏", "㻘", "䀥", "䁖", "䂵", "䃅", "䅉", "䅟", "䅪", "䇲", "䉤", "䌶", "䌷", "䌸", "䌹", "䌺", "䌻", "䌼", "䌽", "䌾", "䌿", "䍀", "䍁", "䍠", "䎬", "䏝", "䑽", "䓓", "䓕", "䓖", "䓨", "䗖", "䘛", "䘞", "䙊", "䙌", "䙓", "䜣", "䜤", "䜥", "䜧", "䜩", "䝙", "䞌", "䞍", "䞎", "䞐", "䟢", "䢀", "䢁", "䢂", "䥺", "䥽", "䥾", "䥿", "䦀", "䦁", "䦂", "䦃", "䦅", "䦆", "䦶", "䦷", "䩄", "䭪", "䯃", "䯄", "䯅", "䲝", "䲞", "䲟", "䲠", "䲡", "䲢", "䲣", "䴓", "䴔", "䴕", "䴖", "䴗", "䴘", "䴙", "䶮", "万", "与", "丑", "专", "业", "丛", "东", "丝", "丢", "两", "严", "丧", "个", "丰", "临", "为", "丽", "举", "么", "义", "乌", "乐", "乔", "习", "乡", "书", "买", "乱", "了", "争", "于", "亏", "云", "亘", "亚", "产", "亩", "亲", "亵", "亸", "亿", "仅", "仆", "仇", "从", "仑", "仓", "仪", "们", "价", "仿", "众", "优", "伙", "会", "伛", "伞", "伟", "传", "伡", "伣", "伤", "伥", "伦", "伧", "伪", "伫", "体", "余", "佛", "佣", "佥", "侠", "侣", "侥", "侦", "侧", "侨", "侩", "侪", "侬", "侭", "俊", "俣", "俦", "俨", "俩", "俪", "俫", "俭", "修", "借", "债", "倾", "偬", "偻", "偾", "偿", "傤", "傥", "傧", "储", "傩", "僵", "儿", "克", "兑", "兖", "党", "兰", "关", "兴", "兹", "养", "兽", "冁", "内", "冈", "册", "写", "军", "农", "冬", "冯", "冲", "决", "况", "冻", "净", "凄", "准", "凉", "凌", "减", "凑", "凛", "几", "凤", "凫", "凭", "凯", "凶", "出", "击", "凿", "刍", "划", "刘", "则", "刚", "创", "删", "别", "刬", "刭", "刮", "制", "刹", "刽", "刾", "刿", "剀", "剂", "剐", "剑", "剥", "剧", "劝", "办", "务", "劢", "动", "励", "劲", "劳", "势", "勋", "勚", "匀", "匦", "匮", "区", "医", "千", "升", "华", "协", "单", "卖", "卜", "占", "卢", "卤", "卧", "卫", "却", "卷", "卺", "厂", "厅", "历", "厉", "压", "厌", "厍", "厐", "厕", "厘", "厢", "厣", "厦", "厨", "厩", "厮", "县", "叁", "参", "叆", "叇", "双", "发", "变", "叙", "叠", "只", "台", "叶", "号", "叹", "叽", "吁", "吃", "合", "吊", "同", "后", "向", "吓", "吕", "吗", "吣", "吨", "听", "启", "吴", "呐", "呒", "呓", "呕", "呖", "呗", "员", "呙", "呛", "呜", "周", "咏", "咙", "咛", "咝", "咤", "咨", "咸", "咽", "哄", "响", "哑", "哒", "哓", "哔", "哕", "哗", "哙", "哜", "哝", "哟", "唇", "唛", "唝", "唠", "唡", "唢", "唤", "啧", "啬", "啭", "啮", "啯", "啰", "啴", "啸", "喂", "喷", "喽", "喾", "嗫", "嗳", "嘘", "嘤", "嘱", "噜", "噪", "嚣", "回", "团", "园", "困", "囱", "围", "囵", "国", "图", "圆", "圣", "圹", "场", "坂", "坏", "坐", "块", "坚", "坛", "坜", "坝", "坞", "坟", "坠", "垄", "垅", "垆", "垒", "垦", "垩", "垫", "垭", "垯", "垱", "垲", "垴", "埘", "埙", "埚", "埯", "堑", "堕", "塆", "墙", "壮", "声", "壳", "壶", "壸", "处", "备", "复", "够", "夫", "头", "夸", "夹", "夺", "奁", "奂", "奋", "奖", "奥", "奸", "妆", "妇", "妈", "妩", "妪", "妫", "姗", "姜", "姹", "娄", "娅", "娆", "娇", "娈", "娘", "娱", "娲", "娴", "婳", "婴", "婵", "婶", "媪", "媭", "嫒", "嫔", "嫱", "嬷", "孙", "学", "孪", "宁", "宝", "实", "宠", "审", "宪", "宫", "家", "宽", "宾", "寝", "对", "寻", "导", "寿", "将", "尔", "尘", "尝", "尧", "尴", "尸", "尽", "局", "层", "屃", "屉", "届", "属", "屡", "屦", "屿", "岁", "岂", "岖", "岗", "岘", "岙", "岚", "岛", "岩", "岭", "岳", "岽", "岿", "峃", "峄", "峡", "峣", "峤", "峥", "峦", "峰", "崂", "崃", "崄", "崭", "嵘", "嵚", "嵝", "巅", "巨", "巩", "巯", "币", "布", "帅", "师", "帏", "帐", "帘", "帜", "带", "帧", "席", "帮", "帱", "帻", "帼", "幂", "幞", "干", "并", "幸", "广", "庄", "庆", "床", "庐", "庑", "库", "应", "庙", "庞", "废", "庵", "庼", "廪", "开", "异", "弃", "弑", "张", "弥", "弦", "弪", "弯", "弹", "强", "归", "当", "录", "彟", "彦", "彨", "彩", "彻", "征", "径", "徕", "御", "忆", "忏", "志", "忧", "念", "忾", "怀", "态", "怂", "怃", "怄", "怅", "怆", "怜", "总", "怼", "怿", "恋", "恒", "恤", "恳", "恶", "恸", "恹", "恺", "恻", "恼", "恽", "悦", "悫", "悬", "悭", "悮", "悯", "惊", "惧", "惨", "惩", "惫", "惬", "惭", "惮", "惯", "愈", "愠", "愤", "愦", "愿", "慑", "慭", "懑", "懒", "懔", "戆", "戋", "戏", "戗", "战", "戚", "戬", "戯", "户", "才", "扎", "扑", "托", "扣", "执", "扩", "扪", "扫", "扬", "扰", "折", "抚", "抛", "抟", "抠", "抡", "抢", "护", "报", "抬", "抵", "担", "拐", "拟", "拢", "拣", "拥", "拦", "拧", "拨", "择", "挂", "挚", "挛", "挜", "挝", "挞", "挟", "挠", "挡", "挢", "挣", "挤", "挥", "挦", "挨", "挽", "捝", "捞", "损", "捡", "换", "捣", "据", "掳", "掴", "掷", "掸", "掺", "掼", "揽", "揾", "揿", "搀", "搁", "搂", "搄", "搅", "搜", "携", "摄", "摅", "摆", "摇", "摈", "摊", "撄", "撑", "撵", "撷", "撸", "撺", "擜", "擞", "攒", "敌", "敚", "敛", "敩", "数", "斋", "斓", "斗", "斩", "断", "无", "旧", "时", "旷", "旸", "昆", "昙", "昵", "昼", "昽", "显", "晋", "晒", "晓", "晔", "晕", "晖", "暂", "暅", "暗", "暧", "曲", "术", "朱", "朴", "机", "杀", "杂", "权", "杆", "杠", "条", "来", "杨", "杩", "杯", "杰", "松", "板", "极", "构", "枞", "枢", "枣", "枥", "枧", "枨", "枪", "枫", "枭", "柜", "柠", "柽", "栀", "栅", "标", "栈", "栉", "栊", "栋", "栌", "栎", "栏", "树", "栖", "栗", "样", "核", "栾", "桠", "桡", "桢", "档", "桤", "桥", "桦", "桧", "桨", "桩", "桪", "梁", "梦", "梼", "梾", "梿", "检", "棁", "棂", "椁", "椝", "椟", "椠", "椢", "椤", "椫", "椭", "椮", "楼", "榄", "榅", "榇", "榈", "榉", "榝", "槚", "槛", "槟", "槠", "横", "樯", "樱", "橥", "橱", "橹", "橼", "檩", "欢", "欤", "欧", "欲", "歼", "殁", "殇", "残", "殒", "殓", "殚", "殡", "殴", "毁", "毂", "毕", "毙", "毡", "毵", "毶", "氇", "气", "氢", "氩", "氲", "汇", "汉", "汤", "汹", "沈", "沟", "没", "沣", "沤", "沥", "沦", "沧", "沨", "沩", "沪", "泛", "泞", "注", "泪", "泶", "泷", "泸", "泺", "泻", "泼", "泽", "泾", "洁", "洒", "洼", "浃", "浅", "浆", "浇", "浈", "浉", "浊", "测", "浍", "济", "浏", "浐", "浑", "浒", "浓", "浔", "浕", "涂", "涚", "涛", "涝", "涞", "涟", "涠", "涡", "涢", "涣", "涤", "润", "涧", "涨", "涩", "淀", "渊", "渌", "渍", "渎", "渐", "渑", "渔", "渖", "渗", "温", "游", "湾", "湿", "溁", "溃", "溅", "溆", "溇", "滗", "滚", "滞", "滟", "滠", "满", "滢", "滤", "滥", "滦", "滨", "滩", "滪", "漓", "潆", "潇", "潋", "潍", "潜", "潴", "澛", "澜", "濑", "濒", "灏", "灭", "灯", "灵", "灶", "灾", "灿", "炀", "炉", "炖", "炜", "炝", "点", "炼", "炽", "烁", "烂", "烃", "烛", "烟", "烦", "烧", "烨", "烩", "烫", "烬", "热", "焕", "焖", "焘", "煴", "熏", "爱", "爷", "牍", "牦", "牵", "牺", "犊", "状", "犷", "犸", "犹", "狈", "狝", "狞", "独", "狭", "狮", "狯", "狰", "狱", "狲", "猃", "猎", "猕", "猡", "猪", "猫", "猬", "献", "獭", "玑", "玙", "玚", "玛", "玮", "环", "现", "玱", "玺", "珐", "珑", "珰", "珲", "琎", "琏", "琐", "琼", "瑶", "瑷", "瑸", "璎", "瓒", "瓮", "瓯", "电", "画", "畅", "畴", "疖", "疗", "疟", "疠", "疡", "疬", "疭", "疮", "疯", "疱", "疴", "症", "痈", "痉", "痒", "痖", "痨", "痪", "痫", "痴", "瘅", "瘆", "瘗", "瘘", "瘪", "瘫", "瘾", "瘿", "癞", "癣", "癫", "皂", "皑", "皱", "皲", "盏", "盐", "监", "盖", "盗", "盘", "眍", "眦", "眬", "睁", "睐", "睑", "瞆", "瞒", "瞩", "矩", "矫", "矶", "矾", "矿", "砀", "码", "砖", "砗", "砚", "砜", "砺", "砻", "砾", "础", "硁", "硕", "硖", "硗", "硙", "硚", "确", "硵", "硷", "碍", "碛", "碜", "碱", "礼", "祃", "祎", "祢", "祯", "祷", "祸", "禀", "禄", "禅", "离", "私", "秃", "秆", "秋", "种", "秘", "积", "称", "秽", "秾", "稆", "税", "稣", "稳", "穑", "穗", "穞", "穷", "窃", "窍", "窎", "窑", "窜", "窝", "窥", "窦", "窭", "竖", "竞", "笃", "笋", "笔", "笕", "笺", "笼", "笾", "筑", "筚", "筛", "筜", "筝", "筹", "筼", "签", "筿", "简", "箓", "箦", "箧", "箨", "箩", "箪", "箫", "篑", "篓", "篮", "篯", "篱", "簖", "籁", "籴", "类", "籼", "粜", "粝", "粤", "粪", "粮", "粽", "糁", "糇", "糍", "系", "紧", "絷", "緼", "縆", "纟", "纠", "纡", "红", "纣", "纤", "纥", "约", "级", "纨", "纩", "纪", "纫", "纬", "纭", "纮", "纯", "纰", "纱", "纲", "纳", "纴", "纵", "纶", "纷", "纸", "纹", "纺", "纻", "纼", "纽", "纾", "线", "绀", "绁", "绂", "练", "组", "绅", "细", "织", "终", "绉", "绊", "绋", "绌", "绍", "绎", "经", "绐", "绑", "绒", "结", "绔", "绕", "绖", "绗", "绘", "给", "绚", "绛", "络", "绝", "绞", "统", "绠", "绡", "绢", "绣", "绤", "绥", "绦", "继", "绨", "绩", "绪", "绫", "绬", "续", "绮", "绯", "绰", "绱", "绲", "绳", "维", "绵", "绶", "绷", "绸", "绹", "绺", "绻", "综", "绽", "绾", "绿", "缀", "缁", "缂", "缃", "缄", "缅", "缆", "缇", "缈", "缉", "缊", "缋", "缌", "缍", "缎", "缏", "缐", "缑", "缒", "缓", "缔", "缕", "编", "缗", "缘", "缙", "缚", "缛", "缜", "缝", "缞", "缟", "缠", "缡", "缢", "缣", "缤", "缥", "缦", "缧", "缨", "缩", "缪", "缫", "缬", "缭", "缮", "缯", "缰", "缱", "缲", "缳", "缴", "缵", "罂", "网", "罗", "罚", "罢", "罴", "羁", "羟", "羡", "群", "翘", "翙", "翚", "耢", "耧", "耸", "耻", "聂", "聋", "职", "聍", "联", "聩", "聪", "肃", "肠", "肤", "肮", "肴", "肾", "肿", "胀", "胁", "胄", "胆", "背", "胜", "胡", "胧", "胨", "胪", "胫", "胶", "脉", "脍", "脏", "脐", "脑", "脓", "脔", "脚", "脱", "脶", "脸", "腊", "腌", "腘", "腭", "腻", "腼", "腽", "腾", "膑", "膻", "臜", "致", "舆", "舍", "舣", "舰", "舱", "舻", "艰", "艳", "艺", "节", "芈", "芗", "芜", "芦", "芲", "芸", "苁", "苇", "苈", "苋", "苌", "苍", "苎", "苏", "苔", "苘", "苧", "苹", "范", "茎", "茏", "茑", "茔", "茕", "茧", "荆", "荐", "荙", "荚", "荛", "荜", "荝", "荞", "荟", "荠", "荡", "荣", "荤", "荥", "荦", "荧", "荨", "荩", "荪", "荫", "荬", "荭", "荮", "药", "莅", "莱", "莲", "莳", "莴", "莶", "获", "莸", "莹", "莺", "莼", "萚", "萝", "萤", "营", "萦", "萧", "萨", "葱", "蒀", "蒇", "蒉", "蒋", "蒌", "蒏", "蒙", "蓝", "蓟", "蓠", "蓣", "蓥", "蓦", "蔂", "蔑", "蔷", "蔹", "蔺", "蔼", "蕰", "蕲", "蕴", "薮", "藓", "藴", "蘖", "虏", "虑", "虚", "虫", "虬", "虮", "虱", "虽", "虾", "虿", "蚀", "蚁", "蚂", "蚃", "蚕", "蚝", "蚬", "蛊", "蛎", "蛏", "蛮", "蛰", "蛱", "蛲", "蛳", "蛴", "蜕", "蜗", "蜡", "蝇", "蝈", "蝉", "蝎", "蝼", "蝾", "螀", "螨", "蟏", "衅", "衔", "补", "表", "衬", "衮", "袄", "袅", "袆", "袜", "袭", "袯", "装", "裆", "裈", "裢", "裣", "裤", "裥", "褛", "褴", "襕", "见", "观", "觃", "规", "觅", "视", "觇", "览", "觉", "觊", "觋", "觌", "觍", "觎", "觏", "觐", "觑", "觞", "触", "觯", "訚", "詟", "誉", "誊", "讠", "计", "订", "讣", "认", "讥", "讦", "讧", "讨", "让", "讪", "讫", "讬", "训", "议", "讯", "记", "讱", "讲", "讳", "讴", "讵", "讶", "讷", "许", "讹", "论", "讻", "讼", "讽", "设", "访", "诀", "证", "诂", "诃", "评", "诅", "识", "诇", "诈", "诉", "诊", "诋", "诌", "词", "诎", "诏", "诐", "译", "诒", "诓", "诔", "试", "诖", "诗", "诘", "诙", "诚", "诛", "诜", "话", "诞", "诟", "诠", "诡", "询", "诣", "诤", "该", "详", "诧", "诨", "诩", "诪", "诫", "诬", "语", "诮", "误", "诰", "诱", "诲", "诳", "说", "诵", "诶", "请", "诸", "诹", "诺", "读", "诼", "诽", "课", "诿", "谀", "谁", "谂", "调", "谄", "谅", "谆", "谇", "谈", "谉", "谊", "谋", "谌", "谍", "谎", "谏", "谐", "谑", "谒", "谓", "谔", "谕", "谖", "谗", "谘", "谙", "谚", "谛", "谜", "谝", "谞", "谟", "谠", "谡", "谢", "谣", "谤", "谥", "谦", "谧", "谨", "谩", "谪", "谫", "谬", "谭", "谮", "谯", "谰", "谱", "谲", "谳", "谴", "谵", "谶", "谷", "豮", "贝", "贞", "负", "贠", "贡", "财", "责", "贤", "败", "账", "货", "质", "贩", "贪", "贫", "贬", "购", "贮", "贯", "贰", "贱", "贲", "贳", "贴", "贵", "贶", "贷", "贸", "费", "贺", "贻", "贼", "贽", "贾", "贿", "赀", "赁", "赂", "赃", "资", "赅", "赆", "赇", "赈", "赉", "赊", "赋", "赌", "赍", "赎", "赏", "赐", "赑", "赒", "赓", "赔", "赕", "赖", "赗", "赘", "赙", "赚", "赛", "赜", "赝", "赞", "赟", "赠", "赡", "赢", "赣", "赪", "赵", "赶", "趋", "趱", "趸", "跃", "跄", "跞", "践", "跶", "跷", "跸", "跹", "跻", "踌", "踪", "踬", "踯", "蹑", "蹒", "蹰", "蹿", "躏", "躜", "躯", "輼", "车", "轧", "轨", "轩", "轪", "轫", "转", "轭", "轮", "软", "轰", "轱", "轲", "轳", "轴", "轵", "轶", "轷", "轸", "轹", "轺", "轻", "轼", "载", "轾", "轿", "辀", "辁", "辂", "较", "辄", "辅", "辆", "辇", "辈", "辉", "辊", "辋", "辌", "辍", "辎", "辏", "辐", "辑", "辒", "输", "辔", "辕", "辖", "辗", "辘", "辙", "辚", "辞", "辟", "辩", "辫", "边", "辽", "达", "迁", "过", "迈", "运", "还", "这", "进", "远", "违", "连", "迟", "迩", "迳", "迹", "适", "选", "逊", "递", "逦", "逻", "遗", "遥", "邓", "邝", "邬", "邮", "邹", "邺", "邻", "郁", "郏", "郐", "郑", "郓", "郦", "郧", "郸", "酂", "酝", "酦", "酱", "酸", "酽", "酾", "酿", "醖", "采", "释", "里", "鉴", "銮", "錾", "钅", "钆", "钇", "针", "钉", "钊", "钋", "钌", "钍", "钎", "钏", "钐", "钑", "钒", "钓", "钔", "钕", "钖", "钗", "钘", "钙", "钚", "钛", "钜", "钝", "钞", "钟", "钠", "钡", "钢", "钣", "钤", "钥", "钦", "钧", "钨", "钩", "钪", "钫", "钬", "钭", "钮", "钯", "钰", "钱", "钲", "钳", "钴", "钵", "钶", "钷", "钸", "钹", "钺", "钻", "钼", "钽", "钾", "钿", "铀", "铁", "铂", "铃", "铄", "铅", "铆", "铇", "铈", "铉", "铊", "铋", "铌", "铍", "铎", "铏", "铐", "铑", "铒", "铓", "铔", "铕", "铖", "铗", "铘", "铙", "铚", "铛", "铜", "铝", "铞", "铟", "铠", "铡", "铢", "铣", "铤", "铥", "铦", "铧", "铨", "铩", "铪", "铫", "铬", "铭", "铮", "铯", "铰", "铱", "铲", "铳", "铴", "铵", "银", "铷", "铸", "铹", "铺", "铻", "铼", "铽", "链", "铿", "销", "锁", "锂", "锃", "锄", "锅", "锆", "锇", "锈", "锉", "锊", "锋", "锌", "锍", "锎", "锏", "锐", "锑", "锒", "锓", "锔", "锕", "锖", "锗", "锘", "错", "锚", "锛", "锜", "锝", "锞", "锟", "锠", "锡", "锢", "锣", "锤", "锥", "锦", "锧", "锨", "锩", "锪", "锫", "锬", "锭", "键", "锯", "锰", "锱", "锲", "锳", "锴", "锵", "锶", "锷", "锸", "锹", "锺", "锻", "锼", "锽", "锾", "锿", "镀", "镁", "镂", "镃", "镄", "镅", "镆", "镇", "镈", "镉", "镊", "镋", "镌", "镍", "镎", "镏", "镐", "镑", "镒", "镓", "镔", "镕", "镖", "镗", "镘", "镙", "镚", "镛", "镜", "镝", "镞", "镟", "镠", "镡", "镢", "镣", "镤", "镥", "镦", "镧", "镨", "镩", "镪", "镫", "镬", "镭", "镮", "镯", "镰", "镱", "镲", "镳", "镴", "镵", "镶", "长", "门", "闩", "闪", "闫", "闬", "闭", "问", "闯", "闰", "闱", "闲", "闳", "间", "闵", "闶", "闷", "闸", "闹", "闺", "闻", "闼", "闽", "闾", "闿", "阀", "阁", "阂", "阃", "阄", "阅", "阆", "阇", "阈", "阉", "阊", "阋", "阌", "阍", "阎", "阏", "阐", "阑", "阒", "阓", "阔", "阕", "阖", "阗", "阘", "阙", "阚", "阛", "队", "阳", "阴", "阵", "阶", "际", "陆", "陇", "陈", "陉", "陕", "陦", "陧", "陨", "险", "随", "隐", "隶", "隽", "难", "雇", "雏", "雕", "雠", "雳", "雾", "霁", "霉", "霡", "霭", "靓", "靔", "静", "面", "靥", "鞑", "鞒", "鞯", "鞲", "韦", "韧", "韨", "韩", "韪", "韫", "韬", "韵", "页", "顶", "顷", "顸", "项", "顺", "须", "顼", "顽", "顾", "顿", "颀", "颁", "颂", "颃", "预", "颅", "领", "颇", "颈", "颉", "颊", "颋", "颌", "颍", "颎", "颏", "颐", "频", "颒", "颓", "颔", "颕", "颖", "颗", "题", "颙", "颚", "颛", "颜", "额", "颞", "颟", "颠", "颡", "颢", "颣", "颤", "颥", "颦", "颧", "风", "飏", "飐", "飑", "飒", "飓", "飔", "飕", "飖", "飗", "飘", "飙", "飚", "飞", "飨", "餍", "饣", "饤", "饥", "饦", "饧", "饨", "饩", "饪", "饫", "饬", "饭", "饮", "饯", "饰", "饱", "饲", "饳", "饴", "饵", "饶", "饷", "饸", "饹", "饺", "饻", "饼", "饽", "饾", "饿", "馀", "馁", "馂", "馃", "馄", "馅", "馆", "馇", "馈", "馉", "馊", "馋", "馌", "馍", "馎", "馏", "馐", "馑", "馒", "馓", "馔", "馕", "马", "驭", "驮", "驯", "驰", "驱", "驲", "驳", "驴", "驵", "驶", "驷", "驸", "驹", "驺", "驻", "驼", "驽", "驾", "驿", "骀", "骁", "骂", "骃", "骄", "骅", "骆", "骇", "骈", "骉", "骊", "骋", "验", "骍", "骎", "骏", "骐", "骑", "骒", "骓", "骔", "骕", "骖", "骗", "骘", "骙", "骚", "骛", "骜", "骝", "骞", "骟", "骠", "骡", "骢", "骣", "骤", "骥", "骦", "骧", "髅", "髋", "髌", "鬓", "鬶", "魇", "魉", "鱼", "鱽", "鱾", "鱿", "鲀", "鲁", "鲂", "鲃", "鲄", "鲅", "鲆", "鲇", "鲈", "鲉", "鲊", "鲋", "鲌", "鲍", "鲎", "鲏", "鲐", "鲑", "鲒", "鲓", "鲔", "鲕", "鲖", "鲗", "鲘", "鲙", "鲚", "鲛", "鲜", "鲝", "鲞", "鲟", "鲠", "鲡", "鲢", "鲣", "鲤", "鲥", "鲦", "鲧", "鲨", "鲩", "鲪", "鲫", "鲬", "鲭", "鲮", "鲯", "鲰", "鲱", "鲲", "鲳", "鲴", "鲵", "鲶", "鲷", "鲸", "鲹", "鲺", "鲻", "鲼", "鲽", "鲾", "鲿", "鳀", "鳁", "鳂", "鳃", "鳄", "鳅", "鳆", "鳇", "鳈", "鳉", "鳊", "鳋", "鳌", "鳍", "鳎", "鳏", "鳐", "鳑", "鳒", "鳓", "鳔", "鳕", "鳖", "鳗", "鳘", "鳙", "鳚", "鳛", "鳜", "鳝", "鳞", "鳟", "鳠", "鳡", "鳢", "鳣", "鳤", "鸟", "鸠", "鸡", "鸢", "鸣", "鸤", "鸥", "鸦", "鸧", "鸨", "鸩", "鸪", "鸫", "鸬", "鸭", "鸮", "鸯", "鸰", "鸱", "鸲", "鸳", "鸴", "鸵", "鸶", "鸷", "鸸", "鸹", "鸺", "鸻", "鸼", "鸽", "鸾", "鸿", "鹀", "鹁", "鹂", "鹃", "鹄", "鹅", "鹆", "鹇", "鹈", "鹉", "鹊", "鹋", "鹌", "鹍", "鹎", "鹏", "鹐", "鹑", "鹒", "鹓", "鹔", "鹕", "鹖", "鹗", "鹘", "鹙", "鹚", "鹛", "鹜", "鹝", "鹞", "鹟", "鹠", "鹡", "鹢", "鹣", "鹤", "鹥", "鹦", "鹧", "鹨", "鹩", "鹪", "鹫", "鹬", "鹭", "鹮", "鹯", "鹰", "鹱", "鹲", "鹳", "鹴", "鹾", "麦", "麸", "麹", "麺", "麽", "黄", "黉", "黡", "黩", "黪", "黾", "鼋", "鼌", "鼍", "鼗", "鼹", "齄", "齐", "齑", "齿", "龀", "龁", "龂", "龃", "龄", "龅", "龆", "龇", "龈", "龉", "龊", "龋", "龌", "龙", "龚", "龛", "龟", "鿎", "鿏", "鿒", "鿔", "𠀾", "𠆲", "𠆿", "𠇹", "𠉂", "𠉗", "𠊉", "𠋆", "𠚳", "𠛅", "𠛆", "𠛾", "𠡠", "𠮶", "𠯟", "𠯠", "𠰱", "𠰷", "𠱞", "𠲥", "𠴛", "𠴢", "𠵸", "𠵾", "𡋀", "𡋗", "𡋤", "𡍣", "𡒄", "𡝠", "𡞋", "𡞱", "𡠟", "𡥧", "𡭜", "𡭬", "𡳃", "𡳒", "𡶴", "𡸃", "𡺃", "𡺄", "𢋈", "𢗓", "𢘙", "𢘝", "𢘞", "𢙏", "𢙐", "𢙑", "𢙒", "𢙓", "𢛯", "𢠁", "𢢐", "𢧐", "𢫊", "𢫞", "𢫬", "𢬍", "𢬦", "𢭏", "𢽾", "𣃁", "𣆐", "𣈣", "𣍨", "𣍯", "𣍰", "𣎑", "𣏢", "𣐕", "𣐤", "𣑶", "𣒌", "𣓿", "𣔌", "𣗊", "𣗋", "𣗙", "𣘐", "𣘓", "𣘴", "𣘷", "𣚚", "𣞎", "𣨼", "𣭤", "𣯣", "𣱝", "𣲗", "𣲘", "𣳆", "𣶩", "𣶫", "𣶭", "𣷷", "𣸣", "𣺼", "𣺽", "𣽷", "𤆡", "𤆢", "𤇃", "𤇄", "𤇭", "𤇹", "𤈶", "𤈷", "𤊀", "𤊰", "𤋏", "𤎺", "𤎻", "𤙯", "𤝢", "𤞃", "𤞤", "𤠋", "𤦀", "𤩽", "𤳄", "𤶊", "𤶧", "𤻊", "𤽯", "𤾀", "𤿲", "𥁢", "𥅘", "𥅴", "𥅿", "𥆧", "𥇢", "𥎝", "𥐟", "𥐯", "𥐰", "𥐻", "𥞦", "𥧂", "𥩟", "𥩺", "𥫣", "𥬀", "𥬞", "𥬠", "𥭉", "𥮋", "𥮜", "𥮾", "𥱔", "𥹥", "𥺅", "𥺇", "𦈈", "𦈉", "𦈋", "𦈌", "𦈎", "𦈏", "𦈐", "𦈑", "𦈒", "𦈓", "𦈔", "𦈕", "𦈖", "𦈗", "𦈘", "𦈙", "𦈚", "𦈛", "𦈜", "𦈝", "𦈞", "𦈟", "𦈠", "𦈡", "𦍠", "𦛨", "𦝼", "𦟗", "𦨩", "𦰏", "𦰴", "𦶟", "𦶻", "𦻕", "𦼖", "𧉐", "𧉞", "𧌥", "𧏖", "𧏗", "𧑏", "𧒭", "𧜭", "𧝝", "𧝧", "𧮪", "𧳕", "𧹑", "𧹒", "𧹓", "𧹔", "𧹕", "𧹖", "𧹗", "𧿈", "𨀁", "𨀱", "𨁴", "𨂺", "𨄄", "𨅛", "𨅫", "𨅬", "𨉗", "𨐅", "𨐆", "𨐇", "𨐈", "𨐉", "𨐊", "𨑹", "𨟳", "𨠨", "𨡙", "𨡺", "𨤰", "𨰾", "𨰿", "𨱀", "𨱁", "𨱂", "𨱃", "𨱄", "𨱅", "𨱆", "𨱇", "𨱈", "𨱉", "𨱊", "𨱋", "𨱌", "𨱍", "𨱎", "𨱏", "𨱐", "𨱑", "𨱒", "𨱓", "𨱔", "𨱕", "𨱖", "𨷿", "𨸀", "𨸁", "𨸂", "𨸃", "𨸄", "𨸅", "𨸆", "𨸇", "𨸉", "𨸊", "𨸋", "𨸌", "𨸎", "𨸘", "𨸟", "𩏼", "𩏽", "𩏾", "𩏿", "𩐀", "𩓋", "𩖕", "𩖖", "𩖗", "𩙥", "𩙦", "𩙧", "𩙨", "𩙩", "𩙪", "𩙫", "𩙬", "𩙭", "𩙮", "𩙯", "𩙰", "𩟿", "𩠀", "𩠁", "𩠂", "𩠃", "𩠅", "𩠆", "𩠇", "𩠈", "𩠉", "𩠊", "𩠋", "𩠌", "𩠎", "𩠏", "𩠠", "𩡖", "𩧦", "𩧨", "𩧩", "𩧪", "𩧫", "𩧬", "𩧭", "𩧮", "𩧯", "𩧰", "𩧱", "𩧲", "𩧳", "𩧴", "𩧵", "𩧶", "𩧸", "𩧺", "𩧻", "𩧼", "𩧿", "𩨀", "𩨁", "𩨂", "𩨃", "𩨄", "𩨅", "𩨆", "𩨇", "𩨈", "𩨉", "𩨊", "𩨋", "𩨌", "𩨍", "𩨎", "𩨏", "𩨐", "𩩈", "𩬣", "𩬤", "𩭹", "𩯒", "𩰰", "𩲒", "𩴌", "𩽹", "𩽺", "𩽻", "𩽼", "𩽽", "𩽾", "𩽿", "𩾀", "𩾁", "𩾂", "𩾃", "𩾄", "𩾅", "𩾆", "𩾇", "𩾈", "𩾊", "𩾋", "𩾌", "𩾎", "𪉂", "𪉃", "𪉄", "𪉅", "𪉆", "𪉈", "𪉉", "𪉊", "𪉋", "𪉌", "𪉍", "𪉎", "𪉏", "𪉐", "𪉑", "𪉒", "𪉓", "𪉔", "𪉕", "𪎈", "𪎉", "𪎊", "𪎋", "𪎌", "𪎍", "𪑅", "𪔭", "𪚏", "𪚐", "𪜎", "𪞝", "𪟎", "𪟝", "𪠀", "𪠟", "𪠡", "𪠳", "𪠵", "𪠸", "𪠺", "𪠽", "𪡀", "𪡃", "𪡋", "𪡏", "𪡛", "𪡞", "𪡺", "𪢌", "𪢐", "𪢒", "𪢕", "𪢖", "𪢠", "𪢮", "𪢸", "𪣆", "𪣒", "𪣻", "𪤄", "𪤚", "𪥠", "𪥫", "𪥰", "𪥿", "𪧀", "𪧘", "𪨊", "𪨗", "𪨧", "𪨩", "𪨶", "𪨷", "𪨹", "𪩇", "𪩎", "𪩘", "𪩛", "𪩷", "𪩸", "𪪏", "𪪑", "𪪞", "𪪴", "𪪼", "𪫌", "𪫡", "𪫷", "𪫺", "𪬚", "𪬯", "𪭝", "𪭢", "𪭧", "𪭯", "𪭵", "𪭾", "𪮃", "𪮋", "𪮖", "𪮳", "𪮶", "𪯋", "𪰶", "𪱥", "𪱷", "𪲎", "𪲔", "𪲛", "𪲮", "𪳍", "𪳗", "𪴙", "𪵑", "𪵣", "𪵱", "𪶄", "𪶒", "𪶮", "𪷍", "𪷽", "𪸕", "𪸩", "𪹀", "𪹠", "𪹳", "𪹹", "𪺣", "𪺪", "𪺭", "𪺷", "𪺸", "𪺻", "𪺽", "𪻐", "𪻨", "𪻲", "𪻺", "𪼋", "𪼴", "𪽈", "𪽝", "𪽪", "𪽭", "𪽮", "𪽴", "𪽷", "𪾔", "𪾢", "𪾣", "𪾦", "𪾸", "𪿊", "𪿞", "𪿫", "𪿵", "𫀌", "𫀓", "𫀨", "𫀬", "𫀮", "𫁂", "𫁟", "𫁡", "𫁱", "𫁲", "𫁳", "𫁷", "𫁺", "𫂃", "𫂆", "𫂈", "𫂖", "𫂿", "𫃗", "𫄙", "𫄚", "𫄛", "𫄜", "𫄝", "𫄞", "𫄟", "𫄠", "𫄡", "𫄢", "𫄣", "𫄤", "𫄥", "𫄦", "𫄧", "𫄨", "𫄩", "𫄪", "𫄫", "𫄬", "𫄭", "𫄮", "𫄯", "𫄰", "𫄱", "𫄲", "𫄳", "𫄴", "𫄵", "𫄶", "𫄷", "𫄸", "𫄹", "𫅅", "𫅗", "𫅥", "𫅭", "𫅼", "𫆏", "𫆝", "𫆫", "𫇘", "𫇛", "𫇪", "𫇭", "𫇴", "𫇽", "𫈉", "𫈎", "𫈟", "𫈵", "𫉁", "𫉄", "𫊪", "𫊮", "𫊸", "𫊹", "𫊻", "𫋇", "𫋌", "𫋲", "𫋷", "𫋹", "𫋻", "𫌀", "𫌇", "𫌋", "𫌨", "𫌪", "𫌫", "𫌬", "𫌭", "𫌯", "𫍐", "𫍙", "𫍚", "𫍛", "𫍜", "𫍝", "𫍞", "𫍟", "𫍠", "𫍡", "𫍢", "𫍣", "𫍤", "𫍥", "𫍦", "𫍧", "𫍨", "𫍩", "𫍪", "𫍫", "𫍬", "𫍭", "𫍮", "𫍯", "𫍰", "𫍱", "𫍲", "𫍳", "𫍴", "𫍵", "𫍶", "𫍷", "𫍸", "𫍹", "𫍺", "𫍻", "𫍼", "𫍽", "𫍾", "𫍿", "𫎆", "𫎌", "𫎦", "𫎧", "𫎨", "𫎩", "𫎪", "𫎫", "𫎬", "𫎭", "𫎱", "𫎳", "𫎸", "𫎺", "𫏃", "𫏆", "𫏋", "𫏌", "𫏐", "𫏑", "𫏕", "𫏞", "𫏨", "𫐄", "𫐅", "𫐆", "𫐇", "𫐈", "𫐉", "𫐊", "𫐋", "𫐌", "𫐍", "𫐎", "𫐏", "𫐐", "𫐑", "𫐒", "𫐓", "𫐔", "𫐕", "𫐖", "𫐗", "𫐘", "𫐙", "𫐷", "𫑘", "𫑡", "𫑷", "𫓥", "𫓦", "𫓧", "𫓨", "𫓩", "𫓪", "𫓫", "𫓬", "𫓭", "𫓮", "𫓯", "𫓰", "𫓱", "𫓲", "𫓳", "𫓴", "𫓵", "𫓶", "𫓷", "𫓸", "𫓹", "𫓺", "𫓻", "𫓼", "𫓽", "𫓾", "𫓿", "𫔀", "𫔁", "𫔂", "𫔃", "𫔄", "𫔅", "𫔆", "𫔇", "𫔈", "𫔉", "𫔊", "𫔋", "𫔌", "𫔍", "𫔎", "𫔏", "𫔐", "𫔑", "𫔒", "𫔓", "𫔔", "𫔕", "𫔖", "𫔭", "𫔮", "𫔯", "𫔰", "𫔲", "𫔴", "𫔵", "𫔶", "𫔽", "𫕚", "𫕥", "𫕨", "𫖃", "𫖅", "𫖇", "𫖑", "𫖒", "𫖓", "𫖔", "𫖕", "𫖖", "𫖪", "𫖫", "𫖬", "𫖭", "𫖮", "𫖯", "𫖰", "𫖱", "𫖲", "𫖳", "𫖴", "𫖵", "𫖶", "𫖷", "𫖸", "𫖹", "𫖺", "𫗇", "𫗈", "𫗉", "𫗊", "𫗋", "𫗚", "𫗞", "𫗟", "𫗠", "𫗡", "𫗢", "𫗣", "𫗤", "𫗥", "𫗦", "𫗧", "𫗨", "𫗩", "𫗪", "𫗫", "𫗬", "𫗭", "𫗮", "𫗯", "𫗰", "𫗱", "𫗳", "𫗴", "𫗵", "𫘛", "𫘜", "𫘝", "𫘞", "𫘟", "𫘠", "𫘡", "𫘣", "𫘤", "𫘥", "𫘦", "𫘧", "𫘨", "𫘩", "𫘪", "𫘫", "𫘬", "𫘭", "𫘮", "𫘯", "𫘰", "𫘱", "𫘽", "𫙂", "𫚈", "𫚉", "𫚊", "𫚋", "𫚌", "𫚍", "𫚎", "𫚏", "𫚐", "𫚑", "𫚒", "𫚓", "𫚔", "𫚕", "𫚖", "𫚗", "𫚘", "𫚙", "𫚚", "𫚛", "𫚜", "𫚝", "𫚞", "𫚟", "𫚠", "𫚡", "𫚢", "𫚣", "𫚤", "𫚥", "𫚦", "𫚧", "𫚨", "𫚩", "𫚪", "𫚫", "𫚬", "𫚭", "𫛚", "𫛛", "𫛜", "𫛝", "𫛞", "𫛟", "𫛠", "𫛡", "𫛢", "𫛣", "𫛤", "𫛥", "𫛦", "𫛧", "𫛨", "𫛩", "𫛪", "𫛫", "𫛬", "𫛭", "𫛮", "𫛯", "𫛰", "𫛱", "𫛲", "𫛳", "𫛴", "𫛵", "𫛶", "𫛷", "𫛸", "𫛹", "𫛺", "𫛻", "𫛼", "𫛽", "𫛾", "𫜀", "𫜁", "𫜂", "𫜃", "𫜄", "𫜅", "𫜊", "𫜑", "𫜒", "𫜓", "𫜔", "𫜕", "𫜙", "𫜟", "𫜨", "𫜩", "𫜪", "𫜫", "𫜬", "𫜭", "𫜮", "𫜯", "𫜰", "𫜲", "𫜳", "𫝈", "𫝋", "𫝦", "𫝧", "𫝨", "𫝩", "𫝪", "𫝫", "𫝬", "𫝭", "𫝮", "𫝵", "𫞅", "𫞗", "𫞚", "𫞛", "𫞝", "𫞠", "𫞡", "𫞢", "𫞣", "𫞥", "𫞦", "𫞧", "𫞨", "𫞩", "𫞷", "𫟃", "𫟄", "𫟅", "𫟆", "𫟇", "𫟑", "𫟕", "𫟞", "𫟟", "𫟠", "𫟡", "𫟢", "𫟤", "𫟥", "𫟦", "𫟫", "𫟬", "𫟲", "𫟳", "𫟴", "𫟵", "𫟶", "𫟷", "𫟸", "𫟹", "𫟺", "𫟻", "𫟼", "𫟽", "𫟾", "𫟿", "𫠀", "𫠁", "𫠂", "𫠅", "𫠆", "𫠇", "𫠈", "𫠊", "𫠋", "𫠌", "𫠏", "𫠐", "𫠑", "𫠒", "𫠖", "𫠜", "𫢸", "𫮃", "𫰛", "𫶇", "𫷷", "𫸩", "𬀩", "𬬭", "𬬻", "𬭊", "𬭛", "𬭳", "𬭶", "𬶋", "𬶍", "𬶏", "𬶟", "𬸪", "阿斗", "阿杰", "阿卡提里", "阿拉干山脉", "阿拉克", "阿里", "阿里斯托芬", "阿娘", "阿森松岛", "阿松森岛", "阿托品", "阿咸", "阿扎伦卡", "阿芝特克人", "阿芝特克语", "哀吊", "哀戚", "哀挽", "埃夫伯里", "埃格尔松", "埃及历", "埃克托", "埃拉托塞尼斯", "挨打", "挨到", "挨得", "挨饿", "挨过", "挨饥抵饿", "挨苦", "挨了", "挨骂", "挨满", "挨磨", "挨日子", "挨三顶四", "挨上", "挨时间", "挨整", "挨揍", "矮杆品种", "矮几", "艾回", "艾里赛宫", "艾里森", "艾瑞里", "爱彼表", "爱困", "爱丽舍宫", "爱欲", "碍难照准", "安纳托利亚", "安沈铁路", "庵蔼", "庵庵", "庵婪", "庵庐", "庵罗树园", "庵舍", "案几", "案准", "暗吃一惊", "暗火", "暗扣", "暗蓝发", "暗劣", "暗乱", "暗伦", "暗昧", "暗冥", "暗莫", "暗浅", "暗然", "暗弱", "暗诵", "暗叹", "暗跳", "暗箱操作", "暗中行事", "盎盂相系", "敖荡", "遨游四海", "熬姜呷醋", "熬制", "翱游四海", "鼇头独占", "奥里里亚", "奥里萨", "奥特朗托", "奥托", "八斗", "八蜡", "八里", "八辟", "八扎", "八只", "八周", "八字胡", "巴斗", "巴拉克", "巴拉松", "巴厘岛", "巴里", "巴松管", "巴托丽", "巴托莉", "巴游", "芭托莉", "笆斗", "拔发", "拔须", "拔宅上升", "把饭叫饥", "吧台", "吧托女", "白发", "白发银须", "白粉面", "白杆兵", "白干", "白骨松", "白果松", "白胡", "白里安", "白面", "白面无须", "白皮松", "白千层", "白术", "白松", "白须", "白云", "白云岩", "百不当一", "百谷", "百花历", "百里", "百炼", "百辟", "百叶", "百叶卷", "百扎", "百只", "柏节松操", "摆布", "摆荡", "摆乌龙", "摆钟", "拜斗", "拜复", "拜岳", "拜占庭", "颁布", "斑岩", "板板", "板栗", "板岩", "办公台", "办伙", "半干", "半个世纪", "半里", "半托", "半只", "拌面", "绑扎", "棒子面", "包干", "包谷", "包伙", "包扎", "包占", "褒采一介", "褒赞", "薄幸", "宝丰", "宝里宝气", "宝历", "宝志", "保险杆", "抱大足杆", "抱朴", "杯布", "杯干", "杯面", "杯赛", "悲戚", "悲筑", "碑志", "北斗", "北回", "北里", "北征", "贝尔托內", "贝里", "贝那芬托", "贝胄", "备尝艰苦", "备御", "备注", "背榜", "背包", "背出", "背出去", "背带", "背负", "背回", "背饥荒", "背筐", "背來", "背篓", "背你", "背人", "背酸", "背他", "背她", "背我", "背物", "背小孩", "背债", "背着", "背走", "被发", "被复", "被人背", "被头散发", "焙干", "呗赞", "本里", "本周", "畚斗", "逼并", "鼻梁", "鼻烟", "比杆赛", "比干", "彼得里皿", "秕谷", "笔杆", "笔管面", "笔卷", "笔秃墨干", "毕升", "庇里牛斯", "庇荫", "碧眼紫须", "弊幸", "壁志", "壁钟", "避凶就吉", "避凶趋吉", "嬖幸", "臂一卷", "砭针", "编发", "编余", "编钟", "鞭辟近里", "鞭辟入里", "扁拟谷盗虫", "变速杆", "变质岩", "便吃干", "便辟", "遍布", "辨奸论", "辫发", "标占", "标志", "标致", "标注", "标准", "标准尺寸", "标准单位", "标准杆", "表板", "表厂", "表带", "表的嘀嗒", "表的历史", "表店", "表冠", "表行", "表壳", "表快", "表款", "表链", "表慢", "表蒙子", "表盘", "表速", "表停", "表王", "表演", "表演欲", "表针", "表转", "别别扭扭", "别口气", "别扭", "别拗", "别气", "别强", "别日南鸿才北去", "别着", "别只", "别致", "别嘴", "玢岩", "滨松市", "鬓发", "冰斗", "冰碛岩", "冰岩", "饼干", "禀复", "并案", "并包", "并不", "并不并", "并产", "并成", "并除", "并到", "并叠", "并发", "并购", "并骨", "并合", "并火", "并肩", "并肩子", "并兼", "并卷机", "并科", "并力", "并拢", "并名", "并入", "并纱", "并吞", "并网", "并为", "并线", "并一不二", "并赃拿贼", "并赃治罪", "并州", "病愈", "拨谷", "拨弦", "波荡", "波尔干", "波发藻", "波拉克", "波里", "剥制", "菠萝干", "伯里克利", "伯娘", "伯余", "泊松", "泊松分布", "博汇", "簸荡", "卜征", "逋发", "补扣", "补注", "捕虫植物", "捕风系影", "捕虏岩", "捕影系风", "哺喂", "不并", "不惮强御", "不吊", "不负所托", "不干", "不干不净", "不干不淨吃了没病", "不干己事", "不干胶", "不干你", "不干他", "不干它", "不干她", "不干我", "不谷", "不好干預", "不气干", "不舍", "不食干腊", "不讨采", "不通吊庆", "不畏强御", "不系", "不赞", "不占", "不占吉凶", "不占算", "不占凶吉", "不准", "不准翻印", "不准没", "不准你", "不准谁", "不准他", "不准它", "不准她", "不准问", "不准我", "布摆", "布班尼斯瓦", "布菜", "布慈", "布道", "布德", "布防", "布复", "布干维尔", "布岗", "布告", "布谷", "布谷鸟钟", "布划", "布会", "布教", "布景", "布局", "布扣", "布拉克", "布雷", "布里", "布里斯托", "布列", "布鲁", "布鲁托", "布伦", "布伦托海", "布满", "布囊", "布囊其口", "布让", "布散", "布哨", "布设", "布施", "布势", "布署", "布托", "布网", "布希总统", "布下", "布线", "布雪", "布一个", "布疑阵", "布于", "布阵", "布政", "布置", "步步高升", "步斗踏罡", "步罡踏斗", "擦干", "猜三划五", "才不", "才出", "才此", "才打", "才当曹斗", "才到", "才得到", "才得两年", "才等", "才读", "才对", "才多", "才敢", "才干", "才干杯", "才干旱", "才干淨", "才干透", "才刚", "才给", "才跟", "才够", "才怪", "才过来", "才过去", "才行", "才好", "才回", "才会", "才将", "才讲", "才开", "才看", "才可", "才来", "才料", "才买", "才没", "才拿", "才能", "才能夠", "才能勇敢追", "才能有", "才派", "才起来", "才气纵横", "才去", "才上到", "才上来", "才上去", "才始", "才是", "才松下", "才算", "才爲", "才下来", "才下去", "才想", "才像", "才信", "才要", "才用", "才用到", "才有", "才再", "才在", "才则", "裁并", "裁制", "采采", "采椽不斲", "采地", "采蘩", "采风", "采风录", "采葛", "采光", "采光剖璞", "采缉", "采及葑菲", "采兰赠芍", "采烈", "采苓", "采绿", "采女", "采芑", "采芹", "采色", "采声", "采诗", "采石", "采石之役", "采石之战", "采石之戰", "采菽", "采头", "采薇", "采薪", "采薪之疾", "采薪之忧", "采衣", "采邑", "采制", "彩笔", "彩笔生", "彩笔生花", "彩绸", "彩船", "彩带", "彩缎", "彩凤", "彩楼", "彩鸾", "彩女", "彩牌楼", "彩棚", "彩球", "彩色世界", "彩胜", "彩线", "彩衣", "彩缯", "菜干", "菜苔", "蔡松坡", "参茸", "参绥", "参汤", "餐松啖柏", "餐松食柏", "餐松饮涧", "餐台", "惨戚", "苍发", "苍术", "苍松", "藏历", "藏蒙歌儿", "操纵杆", "操纵台", "操作台", "操作钟", "曹郁芬", "漕挽", "艸木丰丰", "草庵", "草荐", "草签", "草席", "测量杆", "杈杆儿", "茶几", "茶面", "茶托", "茶已干", "查核", "査报表", "査表", "査不出", "査出", "査兑克", "査号台", "査回", "査获", "査价", "査卷", "査克拉", "査问出", "査无实据", "査修", "査询台", "査找周期", "搽穰卷儿", "察核", "产后", "产后检査", "产制", "铲板", "铲草", "铲铲", "铲车", "铲出", "铲除", "铲刀", "铲倒", "铲掉", "铲斗", "铲断", "铲煤", "铲平", "铲起", "铲抢", "铲球", "铲伤", "铲射", "铲土", "铲下", "铲雪", "铲刈", "铲凿", "肠系膜", "肠脏", "尝遍", "尝尝", "尝出", "尝到", "尝点", "尝个", "尝尽", "尝来尝去", "尝了", "尝了尝", "尝起来", "尝鲜", "厂部", "唱念", "超基性岩", "超级杯", "超赞", "朝钟", "潮烟", "炒面", "车斗", "车夫", "车里", "车里雅宾斯克", "车仔面", "扯面", "扯纤", "撤并", "沉积岩", "陈冲", "陈汉升", "陈杰", "陈炼", "陈升", "陈世杰", "陈万松", "陈幸", "陈幸嫚", "陈尹杰", "陈瀛钟", "陈郁秀", "晨钟", "衬托", "称叹", "称赞", "蛏干", "成岩作用", "呈准", "承制", "乘凶完配", "秤杆", "秤平斗满", "吃板刀面", "吃饱了饭撑的", "吃饱没事干", "吃不出", "吃不了", "吃出", "吃错药", "吃得出", "吃得了", "吃地面", "吃钉板", "吃豆干", "吃饭别忘了种谷人", "吃饭家伙", "吃饭傢伙", "吃干醋", "吃干了", "吃挂络儿", "吃过面", "吃合家欢", "吃后悔药", "吃回头草", "吃几碗干饭", "吃姜", "吃尽", "吃亏就是占便宜", "吃亏上当", "吃辣面", "吃了", "吃里扒外", "吃里爬外", "吃面", "吃枪药", "吃敲才", "吃人虫", "吃人一个蛋恩情无法断", "吃伤了", "吃完面", "吃闲饭", "吃闲话", "吃烟", "吃药", "吃一顿挨一顿", "喫亏的是乖占便宜的是呆", "迟回", "尺布斗粟", "尺寸斗粟", "齿发", "齿危发秀", "斥卤", "赤绳系足", "赤术", "赤松", "冲鼻", "冲厕所", "冲茶", "冲冲", "冲牀工", "冲淡", "冲涤", "冲掉", "冲断", "冲服", "冲沟", "冲冠", "冲冠发怒", "冲和", "冲怀", "冲坏", "冲毁", "冲积", "冲剂", "冲襟", "冲决", "冲克", "冲空机", "冲垮", "冲扩", "冲凉", "冲淋浴", "冲龄", "冲流", "冲昧", "冲模", "冲末", "冲默", "冲年", "冲泡", "冲人", "冲弱", "冲绳", "冲蚀", "冲刷", "冲水", "冲税", "冲塌", "冲天", "冲田", "冲洗", "冲喜", "冲销", "冲霄", "冲泻", "冲虚", "冲印", "冲澡", "冲帐", "冲州撞府", "冲走", "舂谷", "虫部", "虫吃牙", "抽斗", "抽干", "抽烟", "仇仇", "丑旦", "丑角", "丑年", "丑牛", "丑奴儿", "丑日", "丑三", "丑时", "丑月", "偢采", "瞅下表", "瞅下钟", "臭局", "臭熏熏", "出陈布新", "出丑", "出丑狼借", "出儿", "出分子", "出家", "出家人", "出家人吃八方", "出口贸易", "出事情", "出水管", "出太阳", "出戏", "出汙泥而不染", "出征", "初征", "除旧布新", "触须", "搋面", "川谷", "川后", "传布", "传位于四太子", "船夫", "船娘", "船只", "船钟", "遄征", "串游", "窗明几净", "窗明几亮", "床席", "闯炼", "创获", "创巨", "创伤后压力", "创意杯", "吹发", "吹干", "吹胡", "炊臼之戚", "炊烟", "炊烟袅袅", "垂发", "捶炼", "捶台拍凳", "锤炼", "春卷", "春秋大一统", "春笋怒发", "春武里府", "唇干", "唇燥舌干", "脣若抹朱", "脣若涂朱", "脣似抹朱", "淳于", "醇郁", "词采", "词汇", "辞采", "辞汇", "慈悲喜舍", "磁制", "此仆彼起", "此系", "刺参", "刺干", "赐恤", "葱葱郁郁", "葱姜蒜", "粗管面", "粗卤", "粗面", "粗面岩", "粗制", "醋栗", "醋坛", "催并", "脆谷乐", "淬炼", "存托股", "存折", "寸发千金", "错彩镂金", "搭干铺", "达欣杯", "答复", "打挨", "打并", "打出", "打出吊入", "打出头棒子", "打谷", "打哄", "打簧表", "打桨杆", "打卡钟", "打游飞", "打游击", "打制", "打中伙", "打钟", "大本钟", "大笨钟", "大病初愈", "大不里士", "大采", "大冲", "大虫", "大虫不吃伏肉", "大虫吃小虫", "大丑", "大斗", "大发慈悲", "大夫松", "大鉴", "大金发苔", "大蜡", "大里", "大理岩", "大历", "大利面", "大麻里", "大明历", "大目干连", "大辟", "大千世界", "大曲", "大水冲倒龙王殿", "大水冲倒龙王庙", "大水冲溺", "大同世界", "大武仑", "大咸", "大型钟", "大凶", "大言非夸", "大衍历", "大一统志", "大赞", "大折儿", "大只", "大钟", "大周后", "大专杯", "呆致致", "逮系", "代表人物", "代签", "带发修行", "带凶", "待核", "袋表", "戴表", "戴发含齿", "戴维斯杯", "丹参", "丹干", "担担面", "担干纪", "担仔面", "单夫只妇", "单位信托", "单弦", "单于", "单只", "单周", "单子叶植物", "胆大如斗", "胆大于天", "但得一片橘皮吃且莫忘了洞庭湖", "但云", "弹珠台", "弹子台", "澹荡", "当当", "当当当", "当的一声", "当的一响", "当家", "当家才知柴米价", "当啷", "当一声", "当周", "挡御", "党参", "党怀英", "党进", "党太尉", "党太尉吃匾食", "党项", "党項", "党支书", "荡出", "荡船", "荡荡", "荡荡悠悠", "荡到", "荡涤", "荡风", "荡复", "荡垢涤汙", "荡寒", "荡检逾闲", "荡酒", "荡开", "荡口", "荡来荡去", "荡气回阳", "荡秋千", "荡瑕涤秽", "荡漾", "荡悠悠", "荡舟", "叨念", "捣鬼吊白", "倒八字须", "倒念", "倒悬挨命", "祷念", "盗钟", "道里", "稻谷", "得采", "德干", "德高而毁来", "德里", "德胜头回", "的钟", "灯彩", "等闲人物", "低荡", "低回", "低价买进", "嘀嗒的表", "滴干", "滴里耷拉", "滴里搭拉", "滴里嘟噜", "狄拉克", "狄志杰", "涤荡", "涤秽荡瑕", "涤瑕荡垢", "涤瑕荡秽", "诋毁", "抵触", "抵牾", "抵御", "地复天翻", "地塞米松", "地无三里平", "地一卷", "地志", "帝后", "递回", "第八出", "第二出", "第九出", "第六出", "第七出", "第三出", "第十出", "第四出", "第五出", "第一出", "颠颠仆仆", "颠复", "颠仆", "攧攧仆仆", "点半钟", "点多钟", "点烟", "点钟", "电表", "电复", "电胡刀", "电脑台", "电须刀", "电钟", "电子表", "电子钟", "殿钟自鸣", "刁斗", "刁奸", "貂复额", "雕鹗", "雕悍", "雕具座", "雕梁", "雕翎", "雕心雁爪", "吊词", "吊奠", "吊斗", "吊古", "吊诡", "吊贺迎送", "吊鹤", "吊喉", "吊谎", "吊祭", "吊脚", "吊脚儿", "吊脚儿事", "吊拷", "吊客", "吊民", "吊旗", "吊撒", "吊丧", "吊书", "吊死", "吊死问孤", "吊死问疾", "吊头", "吊慰", "吊文", "吊问", "吊孝", "吊唁", "吊宴", "吊喭", "吊腰撒跨", "吊影", "吊誉沽名", "吊者大悦", "吊纸", "吊钟", "掉发", "爹娘", "叠层岩", "蹀里蹀斜", "丁丑", "丁固生松", "叮当", "顶针", "顶针挨住", "订制", "钉扣", "定制", "东仓里", "东干", "东里", "东山里", "东升", "东征", "东芝医疗系", "东周", "东周钟", "冬冬", "冬季世界", "董里府", "董氏封发", "动荡", "栋梁", "都舍下", "斗柄", "斗车", "斗城", "斗储", "斗大", "斗胆", "斗的", "斗灯", "斗店", "斗顿", "斗方", "斗分子", "斗府", "斗概", "斗拱", "斗栱", "斗哄", "斗斛之禄", "斗箕", "斗极", "斗酒", "斗酒只鸡", "斗居", "斗绝", "斗魁", "斗笠", "斗量", "斗六", "斗罗大陆", "斗门", "斗南", "斗牛", "斗牛之间", "斗蓬装", "斗篷", "斗渠", "斗然", "斗山", "斗筲", "斗杓", "斗升", "斗食", "斗室", "斗数", "斗薮", "斗粟尺布", "斗粟囊金", "斗尾港", "斗纹", "斗香", "斗小马", "斗烟丝", "斗斋", "斗帐", "斗折蛇行", "斗真", "斗重山齐", "斗转参横", "斗转星移", "斗子", "豆干", "豆干肉丝", "豆面", "豆签", "毒僵指", "独占鼇头", "读书三余", "赌台", "杜老志道", "杜雅里克", "短发", "短几", "短须", "断发", "断弦", "断纸余墨", "锻炼", "堆案盈几", "对表", "对折", "对准", "对准表", "对准钟", "多采", "多层复", "多吃多占", "多个", "多少只", "多凶少吉", "多只", "夺杯", "俄制", "额我略历", "恶心", "恶意", "恶意毁谤", "恶唑啉", "恩同父母", "恩准", "摁扣", "而云", "二斗", "二恶英", "二缶钟惑", "二杆子", "二里", "二仑", "二娘", "二撇胡", "二手烟", "二弦", "二叶松", "二只", "二周", "发辫", "发表", "发表欲", "发鬓", "发布", "发采扬明", "发菜", "发钗", "发带", "发雕", "发短心长", "发匪", "发肤", "发复", "发干", "发根", "发箍", "发光", "发光可鉴", "发号", "发号布令", "发皓齿", "发际", "发髻", "发夹", "发笺", "发胶", "发脚", "发结", "发姐", "发禁", "发卷", "发卡", "发困", "发腊", "发蜡", "发廊", "发量", "发乱钗横", "发蒙", "发面", "发漂", "发妻", "发圈", "发如飞蓬", "发乳", "发色", "发纱", "发上", "发上冲冠", "发上指冠", "发梢", "发式", "发饰", "发梳", "发束", "发霜", "发丝", "发套", "发挽双髻", "发网", "发为血之本", "发尾", "发屋", "发香", "发型", "发须斑", "发须都", "发须皆", "发须俱", "发须已", "发癣", "发已霜白", "发引", "发引千钧", "发缨", "发踊冲冠", "发油", "发簪", "发展中国", "发长", "发针", "发政施仁", "发指", "发质", "发状", "廢后", "法拉托", "翻复", "翻来复去", "翻手作云复手雨", "翻台", "翻云复雨", "烦复", "繁复", "繁钟", "反铲", "反斗", "反反复复", "反复", "反卷", "反乱并", "返还占有", "返里", "返照回光", "饭后", "饭后钟", "饭团", "泛滥", "泛水", "泛水凌山", "范冰冰", "范陈柏", "范成大", "范德格拉夫", "范德林特", "范德萨", "范德瓦耳斯", "范德维德", "范登堡", "范范之辈", "范甘迪", "范纲武", "范戈德", "范公偁", "范公堤", "范光群", "范国铨", "范哈能", "范皓阗", "范洪森", "范家谦", "范嘉骅", "范姜", "范进", "范靖瑶", "范雎", "范可钦", "范宽", "范蠡", "范伦铁诺", "范履霜", "范尼斯特鲁伊", "范佩西", "范琪斐", "范绮馨", "范士丹", "范斯坦", "范特尔", "范特西", "范玮琪", "范文", "范文程", "范文芳", "范文虎", "范文澜", "范文藤", "范文同", "范文照", "范文正公", "范湘暄", "范小姐", "范晓萱", "范筱梵", "范欣妤", "范阳", "范晔", "范逸臣", "范增", "范张鸡黍", "范正祥", "范织钦", "范植谷", "范植伟", "范志毅", "范仲淹", "方便面", "方才", "方几", "方里", "方岳", "方志", "防台", "防御", "仿佛", "仿制", "放蒙挣", "放松", "放松管制", "飞刍挽粒", "飞刍挽粮", "飞刍挽粟", "飞行钟", "飞梁", "飞粮挽秣", "飞升", "飞扎", "飞征", "非签不可", "非游离辐射伤害", "菲舍尔", "肥皂", "肥皂莢", "肥皂劇", "肥皂絲", "肥筑方言", "肺脏", "废后", "费尔干纳", "费里克斯", "分半钟", "分布", "分多钟", "分钟", "分子钟", "焚毁", "粉面", "粉面朱脣", "粪秽蔑面", "丰标", "丰采", "丰城", "丰城贯斗", "丰度", "丰情", "丰茸", "丰容", "丰若有肌柔若无骨", "丰神", "丰溪里", "丰仪", "丰儀", "丰韵", "丰韻", "丰姿", "风采", "风斗", "风干", "风刮", "风后", "风卷", "风入松", "风土志", "风物志", "风云人物", "封后", "封妻荫子", "風采", "峰回", "葑菲之采", "蜂后", "逢凶化吉", "缝制", "凤皇于蜚", "凤凰于蜚", "凤台", "凤尾松", "凤占", "奉干", "奉公克己", "佛里特", "佛历", "佛罗里达", "佛钟", "夫力", "夫役", "肤发", "弗里得里希", "弗里德里希", "弗里敦", "弗里曼", "伏几", "伏尸流血", "扶幼周", "扶余", "拂荡", "拂须", "拂钟无声", "服饰周", "服务台", "服装周", "浮梁", "浮签", "浮托", "浮游", "符采", "福生于微", "福荫", "抚尸", "抚尸恸哭", "抚松", "抚恤", "腑脏", "腐干", "父母在不远游", "负图之托", "妇人生须", "附膻逐臭", "附膻逐秽", "附膻逐腥", "附注", "复按", "复败", "复杯", "复被", "复本", "复比", "复庇之恩", "复蔽", "复壁", "复变函数", "复瓿", "复测", "复查", "复査", "复车", "复称", "复成", "复呈", "复帱", "复词", "复次", "复道", "复电", "复鼎", "复对数", "复发", "复发率", "复发性", "复方", "复肥", "复分解", "复分数", "复分析", "复辅音", "复复", "复盖", "复阁", "复共轭", "复果", "复海移山", "复醢", "复函", "复函数", "复合", "复核", "复婚", "复婚制", "复基因", "复检", "复酱瓿", "复蕉寻鹿", "复句", "复决", "复军", "复利", "复列", "复流", "复鹿寻蕉", "复鹿遗蕉", "复露", "复冒", "复没", "复灭", "复名", "复命", "复亩珍", "复目", "复墓", "复逆", "复拍子", "复盘", "复盆", "复辟", "复频", "复品牌", "复平面", "复评", "复钱", "复去翻来", "复仞年如", "复襦", "复赛", "复色", "复上", "复审", "复式", "复试", "复视", "复述", "复数", "复水", "复诵", "复苏", "复文", "复习", "复线", "复相关", "复校", "复写", "复信", "复姓", "复选", "复穴", "复循环发电", "复训", "复盐", "复眼", "复验", "复叶", "复以百万", "复议", "复意", "复音", "复印", "复用", "复盂", "复雨翻云", "复育", "复元", "复元音", "复员", "复员军人", "复阅", "复韵", "复杂", "复载", "复在", "复帐", "复辙", "复诊", "复殖目", "复殖吸虫", "复制", "复种", "复舟", "复住", "复子明辟", "复字键", "复宗", "复综语", "复𫗧", "傅里叶", "傅匀余", "富里", "覆雨翻云", "馥郁", "该钟", "改念", "改签", "肝脏", "扞御", "杆秤", "杆刀", "杆菌", "杆茵", "杆直", "杆状", "赶面棍", "赶制", "敢斗了胆", "橄榄岩", "擀面", "干阿奶", "干碍", "干熬", "干巴", "干爸", "干白", "干杯", "干贝", "干绷", "干逼", "干煸", "干扁豆角", "干瘪", "干冰", "干剥剥", "干不", "干不干杯", "干不干净", "干布", "干擦", "干材", "干菜", "干草", "干茶钱", "干柴", "干产", "干唱", "干炒牛河", "干城", "干池", "干刍", "干茨腊", "干此杯", "干此坛", "干醋", "干脆", "干村沙", "干打雷", "干打垒", "干旦", "干得", "干得很", "干得两杯", "干得三杯", "干得一杯", "干的", "干灯盏", "干等", "干瞪眼", "干地", "干弟", "干点", "干电", "干吊着下巴", "干掉", "干掉那杯", "干掉那碗", "干掉一杯", "干掉一瓶", "干掉一碗", "干掉这杯", "干掉这碗", "干爹", "干断", "干儿", "干犯", "干饭", "干肥", "干粉", "干干", "干纲", "干稿", "干戈", "干哥", "干沟", "干股", "干卦", "干馆", "干果", "干过", "干过杯", "干过一杯", "干过瘾", "干旱", "干嚎", "干号", "干耗", "干和", "干涸", "干红", "干糇", "干花", "干回付", "干哕", "干货", "干霍乱", "干姬松茸", "干急", "干几杯", "干几手", "干几碗", "干季", "干将", "干姜", "干焦", "干酵母", "干结", "干姐", "干疥", "干尽", "干尽一杯", "干尽一壺", "干尽一坛", "干尽一碗", "干井", "干净", "干淨", "干凯文", "干咳", "干渴", "干刻版", "干枯", "干哭", "干酪", "干了", "干了杯", "干了这杯", "干了这碗", "干了这一杯", "干了这一瓶", "干雷", "干冷", "干礼", "干李子", "干连", "干凉", "干粮", "干两杯", "干量", "干料", "干撂台", "干裂", "干馏", "干落", "干妈", "干毛巾", "干冒烟", "干没", "干梅", "干妹", "干面", "干篾片", "干那", "干那杯", "干那一杯", "干挠", "干你", "干你娘", "干娘", "干奴才", "干暖", "干女", "干呕", "干皮症", "干啤", "干片", "干撇下", "干戚", "干漆", "干虔", "干乔", "干亲", "干卿底事", "干卿何事", "干球温度", "干渠", "干扰", "干热", "干熱", "干肉", "干颡", "干涩", "干烧", "干涉", "干伸舌", "干生气", "干生受", "干生子", "干尸", "干湿", "干湿发", "干时", "干屎橛", "干式", "干手净脚", "干瘦", "干数杯", "干爽", "干丝", "干死", "干闼婆", "干台", "干苔", "干坛子", "干堂婶", "干啼", "干田", "干透", "干图", "干土", "干坞", "干洗", "干系", "干鲜", "干象", "干笑", "干薪", "干性", "干休", "干癣", "干血浆", "干哑", "干咽", "干眼", "干曜", "干谒", "干一", "干一杯", "干一坛", "干一碗", "干衣", "干邑", "干瘾", "干硬", "干又热", "干鱼", "干与", "干预", "干圆洁净", "干云蔽日", "干造", "干燥", "干躁", "干宅", "干站着", "干折", "干这", "干这杯", "干这一杯", "干着", "干着急", "干政", "干支", "干支剌", "干支支", "干枝", "干制", "干重", "干子", "干姊", "干坐着", "冈田准", "刚才", "刚干", "钢扣", "钢梁", "钢制", "港制", "杠杆", "槓杆", "高柏松", "高几", "高丽参", "高良姜", "高清愿", "高郁淨", "高政升", "高周波", "糕干", "咯当", "哥里", "割舍", "歌后", "歌钟", "格里", "格里高利历", "格里历", "格列高利历", "隔周", "葛里芬", "葛罗托斯基", "葛斯范桑", "个别教学", "个旧", "个人电脑", "个人消费", "个中", "个钟", "各类钟", "各里", "各签", "根须", "根烟", "跟斗", "耕获", "骾朴", "更待干罢", "更钟", "工欲善其事", "工致", "公布", "公斗", "公里", "公历", "公认会计准", "公孙丑", "公仔面", "公制单位", "功致", "宫里", "宫里蓝", "拱托", "共铲", "共和历", "共同管道", "共御外侮", "贡烟", "供制", "狗扣", "狗娘养的", "狗占马坑", "购并", "购买欲", "咕咕钟", "沽名干誉", "孤征", "辜濂松", "古迹", "古里古怪", "古切里", "古书云", "古書云", "古弦", "古語云", "古云", "古钟", "谷氨酸", "谷氨酰胺", "谷保家商", "谷仓", "谷舱", "谷草", "谷场", "谷旦", "谷道", "谷粉", "谷风", "谷圭", "谷贵饿农", "谷贵饿农谷贱伤农", "谷贱伤农", "谷糠", "谷壳", "谷口", "谷口耕岩", "谷类", "谷粒", "谷梁", "谷米", "谷苗", "谷皮", "谷人", "谷日", "谷神", "谷食", "谷穗", "谷物", "谷雨", "谷子", "股栗", "股栗肤粟", "骨干分子", "骨灰坛", "骨坛", "骨头里挣出来的钱才做得肉", "鼓荡", "鼓噪", "鹘仑吞枣", "故里", "故云", "顾借", "刮大风", "刮倒", "刮得", "刮风", "刮胡", "刮了", "刮起", "刮去", "刮须", "刮雪", "刮着", "刮走", "寡欲", "挂表", "挂斗", "挂冠", "挂冠归里", "挂角读书", "挂历", "挂面", "挂钟", "拐棒", "拐棍", "拐杖", "拐子", "怪里怪气", "关弓与我确", "关系", "关岳", "关征", "观光周", "观叶植物", "官地为采", "官历", "官准", "馆谷", "管道升", "管弦", "冠军杯", "冠胄", "灌制", "光采", "光杆", "光卤石", "光致致", "咣当", "广布", "广部", "广舍", "归并", "硅质岩", "癸丑", "柜柳", "柜台", "贵价", "桂圆干", "桂仔云", "郭采洁", "郭松焘", "郭子干", "锅伙", "国际漫游拨接服务", "国历", "国立台湾图书馆", "国梁", "国之桢干", "果干", "果子干", "裹扎", "过冲", "过梁", "过水面", "哈里", "哈里里", "还辟", "海参", "海淀", "海干", "海里", "海马回", "海松", "海鲜面", "海于格松", "海岳名言", "含齿戴发", "含油岩", "函复", "韩复矩", "韩国制", "韩升洙", "韩侂胄", "韩制", "汉弥登钟", "旱干", "旱烟", "捍御", "行个方便", "行李卷", "行事历", "行万里路胜读万卷书", "行佣", "行针", "行针布线", "航海历", "蒿里", "毫发", "豪气干云", "好不容易才", "好干", "好困", "好凶", "好一出", "号志", "皓齿朱唇", "皓发", "喝采", "喝倒采", "喝干", "禾谷", "合八字", "合并", "合成乐器", "合吃", "合儿", "合府", "合家", "合历", "合签", "合眼", "合于", "合于时宜", "合中", "合作", "合作伙伴", "何干", "何杰金氏病", "何小升", "和奸", "和面", "和平里", "和弦", "劾系", "河干", "河里", "河里孩儿岸上娘", "河升镇", "荷里活", "核保", "核报", "核编", "核拨", "核查", "核定", "核对", "核复", "核计", "核价", "核减", "核批", "核审", "核实", "核示", "核收", "核算", "核销", "核验", "核准", "核资", "核字", "涸干", "贺后骂殿", "赫弗里希", "鹤吊", "鹤发", "鹤骨松姿", "黑暗世界", "黑发", "黑亮发", "黑面", "黑松", "黑须", "黑曜岩", "很干", "横梁", "烘干", "烘托", "烘熏", "烘云托月", "烘制", "弘历", "红发", "红绳系足", "红丝暗系", "红松", "红须绿眼", "红叶", "红叶杯", "红钟", "洪复", "洪升", "洪士杰", "洪适", "洪钟", "哄动", "哄伙", "哄闹", "哄然", "哄堂", "哄笑", "澒蒙", "喉干舌燥", "后安路", "后摆", "后北街", "后帝", "后发座", "后妃", "后丰", "后豐", "后冠", "后海湾", "后海灣", "后皇", "后稷", "后角", "后街", "后里", "后娘", "后辟", "后平路", "后土", "后王", "后辛", "后羿", "后制", "后座", "后座系", "厚朴", "忽舍下", "狐借虎威", "胡碴子", "胡吃海喝", "胡吃闷睡", "胡匪", "胡椒", "胡椒面", "胡杰", "胡里胡涂", "胡髯", "胡梢", "胡同", "胡托莫", "胡须", "胡云", "胡渣", "胡髭", "胡子", "鹄发", "糊口", "糊里糊涂", "虎皮松", "虎须", "护发", "戽斗", "花庵词选", "花采", "花发老", "花岗岩", "花哄", "花椒面", "花卷", "花马吊嘴", "花托", "花药", "花钟", "划不来", "划船", "划单人艇", "划到", "划到岸", "划到江心", "划得", "划得来", "划动", "划过", "划过来", "划过去", "划行", "划桨", "划进", "划具", "划来", "划了", "划了一会", "划龙舟", "划起", "划拳", "划双人", "划水", "划算", "划艇", "划向", "划一", "划一桨", "划着", "划子", "划走", "华发", "华核", "华里", "华严钟", "哗的", "哗地", "哗哗", "哗啦", "滑杆", "滑借", "画梁雕栋", "怀表", "怀钟", "环游世界", "换挡杆", "换档杆", "换发", "换只", "擐系", "慌里慌张", "皇后", "皇极", "皇极历", "皇历", "皇辟", "黄东梁", "黄发", "黄干黑瘦", "黄金周", "黄俊杰", "黄历", "黄梁", "黄梁美梦", "黄珮筑", "黄曲毒素", "黄须", "黄旭升", "黄岩", "黄郁涵", "黄郁茹", "黄育杰", "黄钰筑", "黄钟", "黄钟毁弃", "黃鈺筑", "湟潦生苹", "晃荡", "灰发", "灰胡", "灰蒙", "辉绿岩", "辉长岩", "回避", "回飙", "回肠", "回冲", "回带", "回荡", "回递性", "回风", "回复", "回光返照", "回归", "回护", "回环", "回黄倒皂", "回魂仙梦", "回敎会议组织", "回教世界", "回卷", "回廊", "回历", "回流", "回路", "回銮", "回盲瓣", "回梦", "回清倒影", "回圈", "回绕", "回声", "回声探测", "回诵", "回天", "回腕", "回文", "回纹针", "回斡", "回翔", "回响", "回向", "回心", "回形夹", "回旋", "回穴", "回雪", "回阳荡气", "回音", "回应", "回佣", "回游", "回匝", "回转", "洄暗", "洄游", "毁弃", "毁诬", "毁犀", "毁炎", "毁誉", "毁钟为铎", "汇报", "汇编", "汇集", "汇辑", "汇刊", "汇算", "汇映", "汇整", "汇总", "汇纂", "会吊", "会签制度", "绘里", "绘制", "烩面", "惠里香", "會干擾", "浑朴自然", "浑仪注", "馄饨面", "活扣", "活塞杆", "火杯", "火并", "火柴杆", "火成岩", "火斗", "火炬松", "火绳杆", "火中取栗", "伙房", "伙夫", "伙食", "伙头", "或系之牛", "获准", "霍克松", "霍里", "击钟", "饥荒", "饥馑", "饥民", "饥年", "机辟", "机器压制", "机械表", "机械钟", "鸡蛋里挑骨头", "鸡蛋面", "鸡奸", "鸡尸牛从", "鸡丝", "鸡丝面", "鸡腿面", "鸡只", "积谷", "积谷防饥", "积金至斗", "基里巴斯", "基里兰柯", "基岩", "赍志没地", "箕斗", "稽核", "激荡", "羁系", "吉里", "吉普斯夸", "吉田松阴", "吉凶", "吉凶庆吊", "吉占", "极乐世界", "急征重敛", "集中托运", "集注", "几案", "几出", "几凳", "几杆", "几乎完全", "几几", "几净窗明", "几率", "几面上", "几旁", "几匹马", "几上", "几世纪", "几榻", "几席", "几筵", "几椅", "几杖", "几只", "几子", "己丑", "虮蝨相吊", "脊梁", "计时表", "纪里谷", "纪历", "纪念", "纪念周", "技术单位", "技术潜水", "忌烟", "季咸", "迹蹈", "觊幸", "祭吊", "祭尸", "加卷", "加拉干达", "加里", "加签", "加注", "佳里", "家伙", "家具", "家俱", "家什", "家私", "傢伙", "嘉柏隆里", "嘉谷", "夹注", "颊须", "甲胄", "贾后", "假发", "賈后", "价电子", "架梁", "架钟", "尖管面", "奸盗邪淫", "奸非", "奸夫", "奸伏", "奸妇", "奸情", "奸杀", "奸尸", "奸通", "奸污", "奸凶", "奸淫", "坚致", "间不容发", "艰巨", "艰苦备尝", "监系", "监制", "兼并", "兼筹并顾", "兼容并蓄", "兼收并蓄", "笺注", "煎面", "茧栗", "俭确之教", "检复", "剪彩", "剪发", "剪其发", "简并", "简朝仑", "简余晏", "戬谷", "碱性岩", "翦彩", "见复", "见鉴", "件钟", "建筑机械", "荐饥", "荐居", "荐臻", "剑杆", "舰只", "鉴核备査", "箭杆", "江采苹", "江干", "江干区", "姜饼", "姜茶", "姜桂", "姜还是老", "姜黄", "姜就是老", "姜辣", "姜老辣", "姜麻园", "姜末", "姜母", "姜片", "姜切片", "姜蓉", "姜石年", "姜是老", "姜丝", "姜汤", "姜糖", "姜文杰", "姜郁美", "姜愈老愈辣", "姜越老越辣", "姜汁", "僵蚕", "僵仆", "僵尸", "彊御", "奖杯", "交并", "交哄", "浇制", "胶卷", "焦干", "焦获", "礁岩", "嚼谷", "角里", "绞干", "矫情干誉", "脚夫", "脚扣", "脚炼", "脚酸", "脚注", "儌幸", "徼幸", "轿夫", "教学钟", "嗟吁", "街里街坊", "节欲", "杰弗里乔叟", "杰里米", "杰里森", "杰伦", "杰特", "拮据", "结采", "结彩", "结发", "结晶岩", "结扣", "结托", "结扎", "截发", "解发", "解发佯狂", "解扣", "解铃系铃", "介系词", "介胄", "戒烟", "借草枕块", "借词", "借此", "借端", "借槁", "借故", "借卉", "借机", "借借", "借口", "借寇兵", "借甚", "借手", "借托", "借以", "借由", "借着", "借助", "借箸", "借箸代筹", "借资", "斤斗", "金杯", "金表", "金伯利岩", "金斗", "金发", "金里奇", "金链", "金仑溪", "金马仑道", "金钱松", "金升圭", "金属杆", "金属制", "金印如斗", "金钟", "觔斗", "津梁", "筋斗", "紧系", "紧致", "锦囊佳制", "尽底下", "尽够", "尽管", "尽教", "尽尽", "尽可", "尽快", "尽里", "尽量", "尽落尾", "尽让", "尽速", "尽先", "尽想", "尽性", "尽意", "尽意随心", "尽有可能", "尽早", "尽子", "尽自", "近日無仇", "浸制", "禁毁", "禁烟", "禁欲", "经折", "荆尸", "菁英杯", "旌恤", "惊叹", "惊赞", "惊钟", "精采", "精奇里江", "精制", "精致", "鲸须", "井干", "颈链", "景致", "警报钟", "警示钟", "警世钟", "警钟", "净发", "淨发", "敬鉴", "敬挽", "敬烟", "迥然回异", "揪发", "揪须", "九谷", "九里", "九炼成钢", "九扎", "九只", "酒干掉", "酒干尽", "酒干了", "酒帘", "酒气熏人", "酒曲", "酒坛", "酒已干", "酒游花", "旧表", "旧历", "旧钟", "就吃干", "就系", "拘系", "居里", "局促", "局限", "咀咽", "莒光周", "举荐征辟", "举手可采", "巨变", "巨额", "巨富", "巨公", "巨奸", "巨舰", "巨奖", "巨款", "巨亏", "巨鹿", "巨商", "巨黍", "巨贪", "巨万", "巨细", "巨献", "巨祥", "巨野", "巨业", "巨债", "巨制", "巨著", "巨子", "巨作", "拒烟", "据干而窥井底", "据云", "聚药雄蕊", "卷包", "卷饼", "卷不起", "卷層云", "卷缠", "卷成", "卷尺", "卷到", "卷动", "卷发", "卷风", "卷钢", "卷过", "卷回", "卷甲", "卷甲重来", "卷进", "卷开", "卷款", "卷来", "卷来卷去", "卷浪", "卷了", "卷帘", "卷拢", "卷落叶", "卷毛", "卷棚", "卷铺盖", "卷舖盖", "卷起", "卷翘", "卷曲", "卷去", "卷刃", "卷入", "卷上", "卷舌", "卷缩", "卷逃", "卷筒", "卷图", "卷土", "卷腿裤", "卷尾猴", "卷吸作用", "卷线器", "卷心", "卷袖", "卷须", "卷旋", "卷烟", "卷烟画片", "卷菸", "卷扬", "卷一卷", "卷衣袖", "卷云", "卷纸", "卷住", "卷走", "绝缘台", "军舰岩", "君子于役", "钧复", "钧鉴", "菌托", "喀拉喀托火山", "卡迪拉克", "卡里", "卡里扎德", "卡洛里", "卡耶里", "开诚布公", "开吊", "开发为", "开发中国", "开哄", "开伙", "揩干", "揩台抹凳", "凯迪拉克", "凯里", "铠胄", "刊布", "看表", "看下表", "看下钟", "看钟", "康采恩", "扛大梁", "抗御", "炕席", "考核", "烤干", "珂里", "柯里", "科斗", "科托努", "科学面", "蚵仔面线", "可干拭", "可干饮", "可可托海", "可周", "克薄", "克剥", "克分子", "克夫", "克核", "克己奉公", "克克", "克扣", "克拉克", "克莱", "克莱査克", "克里", "克里斯蒂安松", "克里斯托", "克里特克里岛", "克落", "克期", "克日", "克啬", "克死", "克星", "克意", "克原子", "克制", "刻半钟", "刻多钟", "刻钟", "峇峇娘惹", "峇里岛", "客制化", "垦丁杯", "垦复", "空蒙", "空前绝后", "空前绝后后", "空钟", "涳蒙", "孔章望斗", "控卷", "控制", "控制杆", "控制台", "控制欲", "口干", "口念", "口燥脣干", "口占", "口钟", "叩钟", "扣环", "扣克", "扣眼", "扣针", "扣子", "枯干", "苦参", "苦瓜干", "苦卤", "裤扣", "夸诞", "夸尔", "夸父", "夸姣", "夸克", "夸丽", "夸毗", "夸人", "夸容", "夸特", "夸脱", "夸赞", "快吃干", "快干", "快舍下", "快速面", "狂并潮", "奎松市", "坤表", "昆剧", "昆仑", "昆腔", "昆曲", "昆山", "昆苏", "昆调", "崑仑", "髡发", "捆扎", "綑扎", "困乏", "困觉", "困倦", "困意", "括发", "拉杆", "拉克施尔德钟", "拉里", "拉链", "拉面", "拉涅里", "拉升", "拉纤", "拉祖里", "邋里邋遢", "腊之以为饵", "蜡査", "蜡祭", "蜡月", "辣椒面", "来复", "莱彩", "莱彩北堂", "莱里达", "栏干", "阑干", "蓝采和", "蓝发", "蓝托斯", "篮虹杯", "啷当", "狼餐虎咽", "狼飧虎咽", "狼吞虎咽", "浪蝶游蜂", "浪琴表", "捞干", "捞面", "劳力士表", "老板", "老雕", "老斗", "老鸹窝里出凤凰", "老姜", "老蒙", "老娘", "老爷钟", "乐器钟", "泐复", "勒里勒得", "了解", "了然", "了如", "了若指掌", "了望", "雷夫范恩斯", "诔赞", "泪干", "冷地里", "冷面", "厘金", "梨干", "漓江", "漓然", "漓水", "漓湘", "礼斗", "礼赞", "李干龙", "李连杰", "李連杰", "李链福", "李盟干", "李三娘", "李钟奭", "李钟郁", "李锺郁", "里昂", "里奥", "里包恩", "里贝利", "里布", "里程", "里程表", "里党", "里尔", "里耳", "里弗赛德", "里港", "里根", "里闬", "里豪", "里肌", "里加", "里贾纳", "里居", "里君", "里科", "里克特", "里扣", "里拉", "里老", "里邻长", "里路", "里闾", "里美", "里门", "里蒙诺夫", "里民", "里名", "里纳", "里尼", "里弄", "里欧", "里奇蒙", "里契蒙", "里人", "里仁", "里舍", "里社", "里士满", "里氏", "里斯", "里谈巷议", "里特维宁科", "里瓦几亚条约", "里维拉", "里希特霍芬", "里巷", "里胥", "里亚", "里谚", "里语", "里约", "里宰", "里长", "里正", "里兹", "理次发", "理发", "理个", "理个发", "理事长杯", "理完发", "理一次发", "理一个发", "力征", "历本", "历法", "历纪", "历命", "历史人物", "历始", "历室", "历书", "历头", "历尾", "历象", "历狱", "历元", "立方厘米", "利得汇", "利默里克", "利托", "利欲", "沥干", "俪采", "栗苞", "栗暴", "栗爆", "栗喇", "栗烈", "栗碌", "栗色", "栗鼠", "栗田雄介", "栗尾", "栗薪", "栗凿", "栗子", "砾岩", "粒变岩", "连杆", "连三并四", "连系", "莲须", "联赛杯", "联系", "镰仓", "炼度", "炼钢", "炼汞", "炼金", "炼铝", "炼贫", "炼师", "炼铁", "炼铜", "炼冶", "炼制", "链甲", "链扣", "链形", "链坠", "良价", "凉面", "凉席", "梁栋", "梁架", "梁龙", "梁木", "梁木其坏", "梁上", "梁文冲", "梁柱", "梁子", "两当一", "两撇胡", "两扎", "两只", "两周", "亮钟", "晾干", "辽沈", "寮采", "燎发", "料斗", "廖于诚", "邻里", "林冲", "林芳郁", "林干闵", "林杰梁", "林俊杰", "林荣松", "林松", "林郁方", "林占梅", "林正杰", "林钟", "临海水土志", "淋冲", "磷酸盐岩", "鳞游", "躏借", "灵迹", "灵修", "灵欲", "凌策", "凌昌焕", "凌驰", "凌河", "凌惠平", "凌借", "凌蒙初", "凌如焕", "凌十八", "凌氏", "凌水", "凌统", "凌退思", "凌小姐", "凌姓", "凌云", "凌云翰", "凌志", "凌志美", "零只", "领扣", "领台", "领袖欲", "令狐冲", "令岳", "溜须", "刘峰松", "刘松藩", "刘松年", "刘伟杰", "刘幸如", "刘占吉", "刘志升", "留发", "流布", "流干", "流纹岩", "流血浮尸", "流血漂卤", "柳斌杰", "柳升耀", "六冲", "六谷", "六通四辟", "六弦", "六须鲇", "六须鲶", "六欲", "六扎", "六只", "六周", "龙卷", "龙虾面", "龙须", "龙须面", "龙游", "龙游浅水", "砻谷机", "镂金错采", "漏斗", "噜噜苏苏", "噜苏", "卢贝松", "卢郁佳", "芦席", "卤簿", "卤代烃", "卤地", "卤钝", "卤化", "卤莽", "卤面", "卤牛肉", "卤人", "卤素", "卤制", "卤族", "甪里", "陆均松", "陆游", "录录", "录制", "鹿门采药", "路里", "路志", "露复", "乱发", "乱哄", "仑背", "仑丰村", "轮回", "轮奸", "罗汉松", "罗兴梁", "萝卜", "萝卜干", "螺杆", "螺旋面", "裸岩", "荦确", "洛克希德马丁", "洛钟东应", "络腮胡", "落发", "落腮胡", "落托", "落叶", "落叶松", "落叶植物", "囉囉苏苏", "囉苏", "驴蒙虎皮", "闾里", "吕后", "吕宋烟", "吕岩", "呂后", "郘钟", "捋采", "铝制", "屡仆屡起", "律历志", "绿发", "麻杆", "麻酱面", "麻栗坡", "马表", "马丁杜里荷", "马夫", "马干", "马后", "马后练服", "马拉巴栗", "马拉松", "马里安纳海沟", "马里克", "马里兰", "马里内斯科", "马里奇", "马鬣松", "马尼托巴", "马普托", "马斯垂克", "马尾松", "马扎", "马占山", "码表", "馬占山", "埋布", "埋头寻表", "埋头寻钟", "买进对冲", "买烟", "迈科里", "麦卡托", "麦科里", "麦托姆", "卖拐", "卖奸", "脉岩", "颟里颟顸", "满布", "满布疑云", "满头洋发", "满洲里", "曼尼托巴省", "慢咽", "忙并", "莽卤", "毛发", "毛姜", "毛里求斯", "毛里塔尼亚", "茂松", "贸易伙伴", "没采", "没地里", "没爹没娘", "没精打采", "没量斗", "没签", "没折至", "梅干", "梅里", "媒人口无量斗", "每只", "每周", "美发", "美国制", "美后", "美里", "美仑", "美制", "美洲杯", "门斗", "闷表", "蒙蔽", "蒙懂", "蒙古大夫", "蒙鸿", "蒙混", "蒙聩", "蒙眬", "蒙昧", "蒙昧不清", "蒙蒙", "蒙蒙懂懂", "蒙蒙黑", "蒙蒙亮", "蒙蒙眬眬", "蒙骗", "蒙事", "蒙汜", "蒙松雨", "蒙瞍", "蒙特卡洛", "蒙头", "蒙托罗拉", "蒙雾", "蒙眼", "蒙在", "蒙在鼓里", "蒙直", "蒙住", "孟德尔松", "梦回", "梦兰叶吉", "梦系", "梦有五不占", "弥漫", "弥蒙", "弥弥", "弥山遍野", "迷奸", "迷蒙", "縻系", "米德尔伯里", "米谷", "米里", "米面", "秘制", "密布", "密折", "密致", "棉制", "免胄", "面霸", "面白无须", "面包", "面饼", "面茶", "面厂", "面点", "面店", "面坊", "面肥", "面粉", "面缸", "面疙瘩", "面馆", "面糊", "面灰", "面价", "面浆", "面酱", "面饺", "面筋", "面码儿", "面坯儿", "面皮", "面票", "面人", "面食", "面塑", "面摊", "面汤", "面条", "面团", "面碗", "面向", "面向对象的技术", "面向对象语言", "面誉", "面誉背毁", "面皂", "面杖", "面子", "面子药", "苗栗", "秒表", "秒钟", "民政里", "民族志", "闵采尔", "闵凶", "抿发", "名表", "名复金瓯", "名录服务", "明査暗访", "明窗净几", "明窗淨几", "明复", "明伙画供", "明扣", "明了", "鸣钟", "冥凌", "冥蒙", "溟蒙", "谬赞", "摸钟", "摹扎特", "模范人物", "模里西斯", "模制", "摩根费里曼", "摩里西斯", "摩托", "磨变岩", "磨炼", "磨制", "魔表", "魔术数字", "抹干", "莫干山", "莫吉托", "莫里", "莫余毒也", "墨荡子", "墨斗", "墨沈", "墨沈未干", "默念", "某只", "母后", "母钟", "木梁", "木偶戏扎", "木制", "木钟", "目牛游刃", "墓志", "幕后人物", "穆罕默德历", "拿坡里", "拿破仑", "拿下表", "拿下钟", "拿贼要赃拿奸要双", "哪一出", "哪只", "那出电影", "那出好戏", "那出剧", "那个", "那个猫儿不吃腥", "那卷", "那只", "纳采", "乃里", "奶卷", "奶娘", "男用表", "南宫适", "南回", "南京钟", "南里", "南山杯", "南征", "南筑", "难挨", "难舍", "难咽", "难荫", "闹表", "闹哄", "闹钟", "内哄", "内脏", "内制", "能舍", "能征惯战", "能征善战", "尼采", "尼克", "尼克松", "泥灰岩", "泥岩", "泥质岩", "倪嗣冲", "拟制", "你才子发昏", "你斗了胆", "你系", "逆钟", "惄如调饥", "拈须", "年谷", "年历", "捻须", "捻针", "念啊", "念吧", "念白", "念错", "念叨", "念到", "念的", "念对", "念佛", "念经", "念了", "念念", "念念有词", "念诗", "念书", "念诵", "念完", "念曰", "念咒", "念作", "娘的", "娘儿", "娘家", "娘舅", "娘老子", "娘亲", "娘胎", "娘姨", "酿制", "鸟松", "袅袅", "袅袅炊烟", "袅袅上升", "袅绕", "袅窕", "尿斗", "捏制", "宁成子", "宁悼子", "宁斧成", "宁浩", "宁惠子", "宁猛力", "宁戚", "宁调元", "宁武子", "宁越", "宁中则", "宁庄子", "凝灰岩", "凝炼", "拧干", "牛骥同皂", "牛柳面", "牛肉面", "牛只", "纽扣", "钮扣", "拗别", "农历", "农民历", "浓发", "浓郁", "醲郁", "弄干", "弄鬼吊猴", "弄面吃", "奴儿干", "怒发冲冠", "怒发冲天", "怒气冲发", "暖荡撩锅", "诺里", "糯米团", "女丑", "女生外向", "欧伯托", "欧几里得", "欧几里德", "欧里庇得斯", "欧里桑", "欧洲杯", "藕复", "帕累托法则", "帕累托最优", "拍出", "拍出好戏", "拍台拍凳", "排档杆", "排骨面", "排须", "徘回", "潘威志", "潘岳", "盘回", "盘松", "蟠采", "蟠龙松", "旁注", "匏系", "跑台子", "泡面", "泡制", "炮制", "陪吊", "培尔松", "培里克里斯", "裴里诺", "裴松之", "佩脱拉克", "配膳台", "配制", "砰当", "烹制", "彭咸", "彭于晏", "蓬发", "篷盖布", "膨土岩", "碰钟", "批复", "批核", "批回", "批注", "批准", "披发", "披复", "披红挂彩", "披头散发", "披榛采兰", "劈里", "皮特拉克", "皮托管", "皮制", "毗婆尸佛", "疲困", "埤塘里", "脾脏", "匹马只轮", "辟恶", "辟咡", "辟谷", "辟举", "辟君三舍", "辟历施鞭", "辟纑", "辟逻", "辟命", "辟匿", "辟辟", "辟然", "辟人之士", "辟色", "辟世", "辟书", "辟违", "辟邪", "辟言", "辟易", "辟淫", "辟引", "辟雍", "辟廱", "辟召", "辟支佛", "辟芷", "偏信则暗", "偏幸", "片麻岩", "片言只语", "片岩", "片语只辞", "漂荡", "漂游", "飘游四海", "撇吊", "拚舍", "品尝", "品汇", "平康里", "平面测量", "评核", "评注", "苹果电脑", "苹萦", "凭吊", "凭几", "凭借", "凭闲", "凭折", "颇复", "破表", "破坏欲", "掊斗折衡", "裒克", "扑冬", "铺盖卷儿", "仆倒", "仆地", "仆夫", "仆街", "仆然", "葡萄干", "蒲松龄", "朴茨茅斯", "朴刀", "朴恩惠", "朴凤柱", "朴父", "朴吉渊", "朴槿惠", "朴京琳", "朴璐美", "朴茂", "朴诗妍", "朴世莉", "朴树", "朴泰桓", "朴硝", "朴新阳", "朴宣英", "朴永训", "朴仔树", "朴赞浩", "朴真熙", "朴正恩", "朴正熙", "朴正祥", "朴志胤", "朴智星", "朴忠", "朴周永", "朴资茅斯", "朴子", "埔里", "普勒托利亚", "普里", "普利艾托", "七里河", "七里香", "七娘", "七娘妈", "七坛", "七弦", "七扎", "七只", "七周", "凄沧", "凄寒", "凄冷", "凄厉", "凄凉", "凄雨", "戚里", "戚戚", "戚墅堰", "欺蒙", "齐后破环", "齐梁世界", "齐王舍牛", "其次辟地", "奇杯", "奇迹", "奇里安", "奇台", "棋布", "棋罗星布", "启发", "启发式敎学法", "起哄", "起扑杆", "气冲斗牛", "气冲牛斗", "气克斗牛", "气若游丝", "气势熏灼", "气吞牛斗", "气扬采飞", "气一冲", "弃舍", "碛卤", "恰才", "千层面", "千回百折", "千回百转", "千钧", "千钧一发", "千里", "千扎", "千只", "迁思回虑", "牵系", "牵一发", "铅制", "悭吝苦克", "谦冲", "签报", "签唱", "签呈", "签出", "签单", "签到", "签得", "签订", "签定", "签赌", "签发", "签过", "签好", "签结", "签了", "签名", "签派室", "签入", "签上", "签收", "签书会", "签署", "签退", "签妥", "签完", "签爲", "签下", "签些", "签写", "签押", "签印", "签有", "签约", "签在", "签章", "签帐", "签着", "签证", "签証", "签注", "签字", "前车复后车戒", "前车之复", "前后左右", "前仆后继", "前仆后起", "钱谷", "乾清宫", "乾象历", "潜水钟", "潜游", "枪杆", "枪托", "强奸", "强投松", "强咽", "强制保险", "硗确", "敲钟", "乔松", "荞麦面", "荞面", "桥梁", "巧历", "窃占", "窃钟掩耳", "侵并", "侵入岩", "亲娘", "亲幸", "亲征", "秦少游", "琴断朱弦", "琴杆", "琴弦", "琴钟", "噙齿戴发", "青帘", "青苹", "青山一发", "青松", "轻扣", "氢卤酸", "倾复重器", "清査不当党产", "清晨杯", "清杆运动", "清台", "情采", "情系", "情欲", "庆吊", "庆历", "磬钟", "穷发", "穷里", "邱富郁", "邱郁婷", "秋不干", "秋发", "秋千", "秋征", "囚系", "求知欲", "虬须", "球后", "球台", "趋吉避凶", "曲车", "曲尘", "曲道", "曲道士", "曲酒", "曲卷", "曲菌", "曲霉", "曲蘖", "曲钱", "曲生", "曲松", "曲秀才", "曲院", "取舍", "圈扣", "圈梁", "权力欲", "权欲熏心", "全彩", "全斗焕", "全干", "全球定位系统卫星测量", "诠注", "痊愈", "鬈发", "犬只", "却才", "确瘠", "确系", "阙里", "裙摆", "群后", "群谋咸同", "群辟", "然身死才数月耳", "髯胡", "染发", "染干", "绕梁", "人参", "人物志", "人欲", "人云", "人云亦云", "仁杰", "忍饥受饿", "忍饥受渴", "认制修", "日本国志", "日本制", "日复一日", "日干", "日进斗金", "日历", "日志", "日制", "荣登后座", "荣获冠军", "溶岩", "熔毁", "熔炼", "熔岩", "镕岩", "揉面", "鞣制", "肉干", "肉羹面", "肉丝面", "肉欲", "如干", "如坐针毡", "茹志鹃", "儒略改革历", "儒略历", "乳娘", "辱游", "入托", "阮咸", "軟肥皂", "瑞郎方面", "瑞签", "瑞士卷", "润发", "若干", "弱智赖于涵", "撒布", "萨巴托", "萨里", "腮斗", "腮托", "赛里木湖", "三复", "三里河", "三里屯", "三娘教子", "三辟", "三尸", "三统历", "三瓦四舍", "三弦", "三熏三沐", "三叶松", "三浴三熏", "三元里", "三扎", "三征七辟", "三只", "三周", "散布", "桑干", "桑托里尼岛", "桑托荣", "桑托斯", "丧荡游魂", "丧钟", "色欲", "铯钟", "沙参", "沙坑杆", "沙里夫", "沙仑", "沙岩", "砂锅面", "砂岩", "傻里傻气", "晒干", "晒谷", "晒烟", "山崩钟应", "山斗", "山里", "山里站", "山梁", "山岩", "山羊胡", "山羊须", "山仔后", "山中无历日", "山重水复", "膻中", "鳝鱼面", "商历", "赏赞", "上冲", "上冲下洗", "上复", "上合屋", "上课钟", "上栗县", "上链", "上梁", "上梁不正下梁歪", "上野树里", "上游", "上周", "烧干", "烧毁", "烧制", "筲斗", "韶山冲", "邵廷采", "舌干唇焦", "舌干脣焦", "舌一卷", "蛇发女妖", "蛇绿混杂岩", "蛇绿岩", "蛇皮松", "蛇纹岩", "舍安就危", "舍本", "舍不得", "舍车保帅", "舍出", "舍得", "舍短从长", "舍短录长", "舍短取长", "舍短用长", "舍堕", "舍己", "舍近即远", "舍近谋远", "舍近求远", "舍近务远", "舍旧迎新", "舍命", "舍弃", "舍去", "舍身", "舍生", "舍实", "舍死忘生", "舍我复谁", "舍我其谁", "舍下", "舍下你", "舍下他", "舍下她", "舍下我", "舍正从邪", "设言托意", "射雕", "射复", "射干", "摄制", "申复", "身敎重于言敎", "身系囹圄", "深成岩", "深山何处钟", "什锦面", "神采", "神雕", "神魂荡飏", "神迹", "神曲", "神曲茶", "神圣周", "神摇魂荡", "沈海", "沈河", "沈积", "沈积岩", "沈吉线", "沈山线", "沈水", "沈阳", "沈州", "审核", "婶娘", "肾脏", "升斗", "升汞", "升华", "升平", "升迁管道", "升天", "升仙", "升学压力", "升阳", "生旦淨末丑", "生栋复屋", "生发", "生姜", "生力面", "生态环境游", "生田斗", "生物钟", "绳扣", "省欲去奢", "圣杯", "圣哈辛托", "圣迹", "圣帕强台", "圣神降临周", "圣修伯里", "胜部冠军", "胜迹", "胜键", "胜肽", "盛价", "盛赞", "尸臣", "尸谏", "尸解", "尸鸠", "尸居龙见", "尸居余气", "尸利", "尸禄", "尸陀林", "尸位", "尸饔", "尸祝", "失之毫厘", "失之毫厘谬以千里", "师娘", "师生杯", "师云而云", "诗云", "诗赞", "诗钟", "施佳升", "施仁布恩", "施仁布泽", "施舍", "施粥舍饭", "湿地松", "詩云", "十分干", "十干", "十里", "十扎", "十只", "十周", "石拐", "石灰岩", "石几", "石枯松老", "石梁", "石松", "石屋制果", "石英表", "石英岩", "石英钟", "石钟乳", "时尚周", "时宪历", "时钟", "时装周", "拾沈", "食虫植物", "食面", "食野之苹", "食欲", "史査克", "史迹", "史托姆", "史托瑟", "史托苏儿", "史托腾柏格", "史托威", "史游", "示范单位", "示范教学", "示复", "世彩堂", "世纪钟", "世界杯", "市长杯", "事迹", "事可干", "事情干脆", "事情可干", "饰扣", "试验台", "试制", "拭干", "是只", "适当的", "視如寇仇", "嗜酸乳干菌", "嗜欲", "收获", "手表", "手工台", "手链", "手面", "手面赚吃", "手术", "手术台", "手酸", "手一卷", "手折", "手制", "手冢治虫", "守御", "首都杯", "首只", "寿面", "受命于天", "售后服务", "兽奸", "兽欲", "书台", "殊域周咨录", "梳发", "梳妆台", "淑郁", "舒卷", "输征", "鼠曲草", "术赤", "束发", "束修", "树梁", "恕乏价催", "数个", "数与虏确", "数周", "数字钟", "数罪并罚", "甩发", "双雕", "双拐", "双后前兵开局", "双胜类", "双折", "双周", "双子叶植物", "水表", "水才干", "水成岩", "水斗", "水干", "水谷", "水谷之海", "水管面", "水里", "水里乡", "水里鄉", "水一冲", "水已干", "水准测量", "睡眠欲", "顺德者吉逆天者凶", "顺钟向", "顺朱儿", "说岳", "丝恩发怨", "丝发", "丝杆", "丝托索", "丝弦", "私欲", "斯伯丁杯", "斯迪里", "斯干", "斯里", "斯里査潘", "斯托", "斯瓦希里", "死面", "死伤相借", "四分历", "四海皆准", "四里", "四面", "四面钟", "四舍六入", "四舍五入", "四凶", "四扎", "四只", "寺钟", "饲喂", "松巴哇", "松柏", "松北", "松本", "松材线虫", "松大辅", "松风", "松冈", "松高路", "松果", "松鹤", "松花", "松化石", "松篁交翠", "松鸡", "松江", "松胶", "松焦油", "松节油", "松井秀", "松菌", "松筠之操", "松筠之节", "松科", "松口", "松口蘑", "松扣", "松类", "松辽平原", "松林", "松岭", "松隆子", "松露", "松萝", "松毛", "松煤", "松明", "松蘑", "松漠", "松木", "松潘", "松皮癣", "松浦", "松乔", "松青", "松丘", "松球", "松毬", "松瓤", "松仁", "松茸", "松蕊", "松山", "松狮犬", "松石", "松鼠", "松树", "松涛", "松桃苗族自治县", "松桃县", "松田", "松尾芭蕉", "松纹", "松溪", "松下", "松香", "松雪泰子", "松蕈", "松鸦", "松烟", "松阳", "松叶", "松油", "松鱼", "松原", "松赞干布", "松藻虫", "松针", "松枝", "松脂", "松指部", "松智路", "松竹", "松子", "送报夫", "诵念", "颂系", "颂赞", "搜藏", "搜査证", "搜购", "搜集", "搜括", "搜罗", "搜证", "溲面", "苏杯", "苏俄在中国", "苏格兰", "苏格兰折耳猫", "苏哈托", "苏家屯", "苏昆", "苏里", "苏仙区", "苏醒", "苏扬托", "酥签", "素发", "素借", "塑胶制", "溯游", "酸懒", "酸麻", "酸软", "酸疼", "酸痛", "蒜发", "蒜苔", "算发", "算历", "虽复能复", "岁凶", "岁聿云暮", "脺脏", "碎发", "碎屑岩", "穗裳", "穗帏飘井干", "穗帷", "穗帐", "孙杰", "笋干", "缩影微卷", "所布的", "所布之", "所系", "所云", "索尔兹伯里平原", "索尔兹伯里石环", "索里士", "索马里", "索馬里", "索面", "索托", "锁扣", "他钟", "塔克", "塔克拉马干", "塔克拉玛干", "塔里班", "塔里木", "塔里契亚努", "塔什干", "塔什库尔干塔吉克自治县", "塔什库尔干乡", "塔什库尔干自治县", "塔钟", "胎发", "台安", "台北市长", "台笔", "台布", "台秤", "台灯", "台凳", "台风", "台鉴", "台历", "台面", "台盘", "台澎金马", "台球", "台山", "台扇", "台湾", "台湾敎育学院", "台湾台", "台制", "台中", "台中敎育大学", "台钟", "台州", "骀借", "太冲", "太初历", "太干", "太后", "太麻里", "太阳黑子周", "泰斗", "贪欲", "坛白干", "坛陈年", "坛高粱", "坛好酒", "坛騞", "坛佳酿", "坛老酒", "坛美酒", "坛女儿红", "坛烧刀子", "坛燒刀子", "坛坛罐罐", "坛子", "潭祉叶吉", "叹服", "叹号", "叹绝", "叹赏", "叹为", "叹为观止", "叹羡", "探知欲", "碳酸岩", "歎吁", "汤面", "汤团", "汤下面", "糖醋里脊", "烫次发", "烫发", "烫个", "烫个发", "烫面", "烫完发", "烫一次发", "烫一个发", "陶制", "特别公积金", "特里", "特内里费", "特松加", "特制", "特准", "腾升", "藤制", "提梁", "提制", "提子干", "题签", "剃发", "剃胡", "剃须", "天冬酰胺", "天翻地复", "天复", "天干", "天干物燥", "天后", "天克地冲", "天历", "天台", "天文学钟", "天文钟", "天心岩", "天要落雨娘要嫁人", "天要下雨娘要嫁人", "田谷", "甜面醬", "甜水面", "捵面", "挑大梁", "挑了", "挑了只", "条几", "调表", "调査团", "调弦", "调制", "跳表", "跳梁", "跳梁猖獗之小丑", "跳梁小丑", "跳只舞", "铁板", "铁板面", "铁板牛柳", "铁板牛肉", "铁杆", "铁拐", "铁扣", "铁托", "铁制", "铁钟", "听弦", "停制", "通布图", "通奸", "通历", "通心面", "同步卫星", "同侪压力", "同吃", "同人", "同人志", "同休共戚", "铜斗儿", "铜扣", "铜制", "铜钟", "瞳蒙", "偷尝禁果", "偷梁换柱", "头发", "头发胡子一把抓", "头花发", "头巾吊在水里", "投托", "秃发", "秃妃之发", "图书馆周", "涂長望", "涂尔干", "涂逢年", "涂鸿钦", "涂鴻欽", "涂惠元", "涂惠源", "涂浆台", "涂谨申", "涂謹申", "涂居贤", "涂坤", "涂美伦", "涂敏恒", "涂敏恆", "涂善妮", "涂绍煃", "涂天相", "涂文生", "涂醒哲", "涂姓", "涂序瑄", "涂永辉", "涂羽卿", "涂月", "涂泽", "涂泽民", "涂澤民", "涂长望", "涂壮勋", "涂壯勳", "土谷祠", "土托鱼", "土制", "团粉", "团伙", "团体冠军", "团子", "团𪢮", "推陈布新", "推诚布公", "推诚布信", "推辇归里", "推托", "推托之词", "推挽", "推弦", "腿酸", "吞并", "吞咽", "屯扎", "托杯", "托比麦奎尔", "托比亚斯", "托钵人", "托钵僧", "托钵修会", "托鉢", "托出", "托胆", "托登汉队", "托地", "托蒂", "托儿", "托尔", "托夫", "托福", "托福考", "托荤咸食", "托架", "托克", "托克托", "托拉", "托辣斯", "托莱多", "托赖", "托老院", "托老中心", "托勒", "托里", "托利党人", "托利米尔", "托领", "托罗斯山", "托洛茨基", "托洛斯基", "托马", "托米", "托墨", "托木尔", "托盘", "托起", "托瑞赛", "托瑞丝", "托腮", "托塞洛", "托色", "托实", "托熟", "托斯卡", "托塔天王", "托胎", "托特", "托维", "托叶", "托幼", "托育", "托运", "托运行李", "托住", "托子", "托足", "拖斗", "脱发", "脱谷机", "坨里", "跎纤", "唾面", "唾面自干", "唾沫直咽", "瓦尔基里", "瓦里", "瓦松", "外强中干", "外烟", "外欲", "外御其侮", "外制", "弯管面", "完全愈复", "顽卤", "顽筑舞笈", "挽词", "挽詞", "挽额", "挽夫", "挽歌", "挽联", "挽聯", "挽曲", "挽诗", "挽詩", "晚钟", "绾发", "碗面", "万里", "万里长征", "万历", "万年", "万年历表", "万旗", "万俟", "万坛", "万扎", "万只", "腕表", "王干发", "王侯后", "王后", "王熙松", "王于真", "王正杰", "王子面", "网御", "网志", "往日無仇", "忘生舍死", "望后石", "威奇托", "偎干", "煨干", "韦后", "为中台", "维斗", "维克", "维克托", "维系", "伟晶岩", "苇苕系巢", "苇席", "尾注", "卫星钟", "未干", "胃脏", "喂饱", "喂哺", "喂动物", "喂鹅", "喂饭", "喂给", "喂过", "喂鸡", "喂了", "喂驴", "喂马", "喂母乳", "喂奶", "喂你", "喂乳", "喂食", "喂它", "喂我", "喂鸭", "喂羊", "喂养", "喂鱼", "喂猪", "慰借", "魏郁奇", "文采", "文采郁郁", "文丑", "文锦复阱", "文斯范恩", "炆面", "稳吃三注", "稳扎", "问卷", "问卷大调査", "问卷调査", "翁干晃", "翁郁容", "窝里", "窝里窝囊", "蜗杆", "我系", "沃依采克", "卧薪尝胆", "握发", "乌东査克", "乌发", "乌干达", "乌狗吃食白狗当灾", "乌兰巴托", "乌里", "乌龙面", "乌洛托品", "乌松", "乌托邦", "污蔑", "巫咸", "诬蔑", "屋梁", "无干", "无价", "无价事", "无价珍珠", "无精打采", "无厘头", "无梁", "无梁斗", "无欲", "吴采璋", "吴皓升", "吴里克", "吴荣杰", "吴育升", "無言不仇", "五采", "五斗", "五谷", "五行生克", "五棵松", "五里", "五辟", "五弦", "五脏", "五扎", "五只", "五周", "伍采克", "武丑", "武大郎吃毒药", "武后", "武里省", "武松", "舞后", "舞水端里", "兀术", "勿里洞岛", "物欲", "西后", "西里", "西历", "西利古里", "西松", "西征", "西周", "西周钟", "吸干", "吸烟", "希伯来历", "希拉克", "希斯仑", "矽岩", "矽质岩", "息谷", "晞发", "欷吁", "稀里", "稀里哗啦", "傒幸", "徯幸", "席卷", "席棚", "席志成", "袭卷", "洗荡", "洗发", "洗发皂", "洗脸台", "洗手台", "喜欢表", "喜欢钟", "戏彩娱亲", "系臂", "系臂之宠", "系泊", "系船桩", "系辞", "系带", "系到", "系而不食", "系发带", "系风捕景", "系风捕影", "系缚", "系个", "系裹", "系好", "系怀", "系获", "系结", "系紧", "系颈", "系颈阙庭", "系扣", "系裤子", "系缆", "系牢", "系了", "系累", "系恋", "系铃解铃", "系铃人", "系留", "系马", "系命", "系捻儿", "系念", "系囚", "系上", "系绳", "系世", "系数", "系丝带", "系蹄", "系条", "系头巾", "系为", "系爲", "系系", "系鞋带", "系心", "系腰", "系一番", "系一片", "系一线", "系一种", "系有", "系于", "系于一发", "系狱", "系爪", "系着", "系争", "系指", "系趾", "系踵", "系住", "细不容发", "细炼", "细如发", "细咽", "细致", "舄卤", "潟卤", "虾干", "虾须", "侠气干云", "狎妓冶游", "下摆", "下采", "下课钟", "下里", "下梁", "下仑路", "下咽", "下游", "下周", "夏后氏", "夏里夫", "夏历", "夏于乔", "夏于喬", "仙后", "仙迹", "仙台", "仙岩", "先尝", "先干为敬", "先签", "先天下之忧而忧后天下之乐而乐", "先小人后君子", "纤夫", "纤户", "纤维植物", "鲜谷王", "鲜于", "贤后", "弦动", "弦断", "弦歌", "弦乐", "弦器", "弦琴", "弦声", "弦索", "弦线", "弦音", "弦轴", "咸安区", "咸池", "咸丰", "咸和", "咸亨", "咸镜", "咸卤", "咸宁", "咸认为", "咸五登三", "咸信", "咸兴", "咸阳", "咸宜", "賢后", "显示表", "显示钟", "县志", "羡叹", "乡里", "乡愿", "相并", "相冲", "相干", "相奸", "相克", "相里", "相托", "香斗", "香干", "香榭里大道", "香烟", "香郁", "香皂", "箱扣", "详注", "响弦", "响钟", "向导", "向迩", "向晦", "向明", "向慕", "向往", "向应", "向者", "项链", "橡斗", "橡子面", "消费欲", "宵征", "萧参", "萧行范篆", "销毁", "小便斗", "小尝", "小丑", "小丑跳梁", "小范", "小几", "小价", "小杰", "小栗旬", "小千世界", "小松", "小型钟", "小余", "小云", "小泽征尔", "小只", "小钟", "校核", "歇斯底里", "邪不干正", "邪辟", "斜管面", "鞋扣", "写字台", "泄欲", "械系", "谢福松", "谢里", "蟹黄鲍鱼面", "心荡神摇", "心高遮了太阳", "心花怒发", "心系", "心细似发", "心弦", "心相系", "心脏", "心脏复苏术", "辛丑", "辛辣面", "辛里希", "新干", "新喀里多尼亚", "新历", "新扎", "信马游缰", "信天游", "信托", "信托贸易", "衅钟", "星辰表", "星斗", "星回", "星历", "星历表", "星罗云布", "星移斗换", "星移斗转", "星占学", "刑克", "刑辟", "刑于", "形单影只", "形孤影只", "形影相吊", "兴云布雨", "幸臣", "幸存", "幸感歌姬", "幸进", "幸免", "幸幸", "幸运", "幸运胡", "性泼凶顽", "性欲", "姓岳", "凶德", "凶地", "凶多吉少", "凶服", "凶怪", "凶耗", "凶荒", "凶礼", "凶门", "凶逆", "凶年", "凶年饥岁", "凶气", "凶日", "凶煞", "凶身", "凶神", "凶事", "凶竖", "凶死", "凶肆", "凶岁", "凶信", "凶讯", "凶燄", "凶宅", "凶兆", "凶终隙末", "雄斗斗", "休仑湖", "休戚", "修胡刀", "修杰楷", "修金", "修敬", "修名", "修脯", "修润", "修樾", "宿松", "秀发", "袖扣", "袖一卷", "吁咈", "吁嗟", "吁了", "吁气", "吁叹", "吁吁", "吁嘘", "吁俞", "须发", "须根", "须后水", "须胡", "须鲸", "须毛", "须眉", "须弥山", "须髯", "须鲨", "须生", "须须", "须子", "虚冲", "墟里", "徐汇", "徐汇区", "徐家汇", "徐余伟", "徐赞升", "许人丰", "许圣杰", "许愿", "许愿起经", "旭日初升", "恤典", "恤荒", "恤金", "续签", "续弦", "蓄发", "蓄胡", "蓄须", "宣布", "宣传周", "喧哄", "玄参", "玄武岩", "玄针", "玄制", "悬臂梁", "悬梁", "悬钟", "悬胄", "旋干转坤", "旋回", "旋里", "旋辟", "削发", "削面", "薛松干", "踅门了户", "雪窗萤几", "雪松", "血才干", "血参", "血已干", "血症", "熏草", "熏风", "熏风徐来", "熏腐", "熏赫", "熏黑", "熏鸡", "熏烤", "熏笼", "熏鑪", "熏人", "熏肉", "熏陶", "熏陶成性", "熏天", "熏习", "熏熏", "熏衣草", "熏鱼儿", "熏烝", "熏蒸", "熏蒸剂", "熏蒸室", "熏制", "薰修", "巡回", "巡回检査", "压杆", "压力表", "压面棍", "压胄子", "鸦窝里出凤凰", "崖广", "哑子托梦", "雅游", "雅致", "雅筑", "亚里", "亚美尼亚历", "亚青杯", "亚松森", "亚洲杯", "咽不了", "咽到", "咽干", "咽肌", "咽进", "咽苦吞甘", "咽了", "咽气", "咽唾", "咽下", "咽着", "咽住", "烟草", "烟厂", "烟袋", "烟袋杆儿", "烟蒂", "烟斗", "烟杆", "烟缸", "烟管", "烟管面", "烟害", "烟灰", "烟碱", "烟禁", "烟酒", "烟卷", "烟民", "烟农", "烟屁股", "烟圈", "烟丝", "烟头", "烟熏", "烟熏火燎", "烟蚜", "烟叶", "烟纸店", "烟嘴", "腌里巴臜", "腌臜", "腌䐶", "腌制", "严云农", "言辩而确", "言大而夸", "言云", "岩仓使节团", "岩层", "岩床", "岩村", "岩村明宪", "岩基", "岩浆", "岩浆岩", "岩礁", "岩濑健", "岩脉", "岩棉", "岩圈", "岩溶", "岩石", "岩土", "岩屑", "岩心", "岩盐", "岩羊", "沿门托钵", "研制", "筵几", "眼干", "眼花了乱", "眼前花发", "眼酸", "偃仆", "偃松", "艳后", "唁吊", "验核", "餍于游乐", "燕几", "扬谷", "羊须疮", "阳春面", "阳历", "杨采妮", "杨凌示范区", "杨日松", "杨士梁", "杨松", "杨苏棣", "杨文志", "洋参", "洋面", "洋烟", "养发", "妖后", "妖里妖气", "腰杆", "腰间系", "腰扣", "腰酸", "腰系", "腰一卷", "侥天之幸", "侥幸", "姚采颖", "姚升志", "摇荡", "摇杆", "咬姜呷醋", "药而愈", "药面儿", "药用植物", "药皂", "要干了", "耶娘", "椰枣干", "噎饥", "爷饭娘羹", "爷羹娘饭", "爷娘", "也斗了胆", "野姜", "叶步梁", "叶恭弘", "叶轮机械", "叶叶", "叶叶琹", "叶音", "叶韵", "叶状植物", "叶子", "叶子杰", "叶子烟", "页岩", "夜光表", "液晶表", "液面", "一表人物", "一别", "一别头", "一并", "一冲性子", "一出剧", "一出子", "一斗", "一斗斗", "一发", "一发千钧", "一发之差", "一发之间", "一杆", "一杆进洞", "一干", "一干而尽", "一干二净", "一个", "一锅面", "一毫一发", "一号木杆", "一哄", "一块面", "一厘一毫", "一里", "一签", "一日叫娘", "一树百获", "一坛", "一坛坛", "一天钟", "一托气", "一托头", "一物克一物", "一赞", "一扎", "一只", "一周", "伊府面", "伊拉克", "伊里奇", "伊面", "伊斯兰历", "伊于胡底", "伊于湖底", "衣摆", "衣不兼采", "衣不完采", "衣不重采", "衣斗木", "衣锦夜游", "衣锦昼游", "衣扣", "衣衫已干", "衣物渐干", "衣物已干", "医托", "噫吁戏", "宜云", "胰脏", "移星换斗", "遗传钟", "遗葑菲采", "遗迹", "疑系", "乙丑", "已系", "已占", "已占算", "以功复过", "蚁后", "倚托", "倚闲", "蟻后", "亿只", "义气干霄", "仡栗", "亦舍下", "亦云", "异采", "异苔同岑", "抑扬升降性", "邑里", "译制", "译注", "逸游自恣", "逸致", "意大利直面", "意面", "因奸成孕", "阴干", "阴历", "阴占", "茵借", "荫蔽", "荫监", "荫生", "荫袭", "音声如钟", "吟叹", "银杯", "银发", "银丝卷", "银须", "银制", "银朱", "淫欲", "嚚暗", "隐几", "飮胄", "印制", "英里", "樱花杯", "鹰雕", "萦回", "萦系", "影后", "影评人周", "影占", "影只形单", "应有尽有", "应占", "应钟", "硬面", "硬咽", "佣钿", "佣金", "佣钱", "庸暗", "永历", "永志", "永志不忘", "咏叹", "泳气钟", "攸戚相关", "悠暗", "悠荡", "悠活丽致", "悠悠荡荡", "悠游", "悠游表", "尤里", "由余", "犹如表", "犹如钟", "犹太历", "油面", "油松", "莜面", "游尘", "游出", "游到", "游芳来", "游蜂浪蝶", "游过", "游过来", "游过去", "游皓玮", "游鸿明", "游鸿儒", "游回", "游击", "游击队", "游击区", "游击手", "游击战", "游进", "游进来", "游进去", "游来", "游来游去", "游乐器", "游离电子", "游龙", "游履", "游美洲", "游明金", "游牧民族", "游乃海", "游欧洲", "游泮", "游禽类", "游去", "游僧攒住持", "游上", "游水", "游完", "游文宏", "游锡", "游锡坤", "游锡昆", "游锡堃", "游戏", "游戏机", "游戏机台", "游下", "游亚洲", "游盈隆", "游泳", "游囿伦", "游鱼", "游昭钦", "游志宏", "游中国", "友于", "有采", "有仇不报非君子", "有出好戏", "有发头陀寺", "有够赞", "有尽有让", "有进有出", "有云", "有征", "有征无战", "有只", "羑里", "牖里", "又干", "又干又硬", "又云", "幼托", "诱奸", "迂回", "纡回", "于贝尔", "于斌", "于波", "于晨楠", "于成龙", "于成龍", "于从濂", "于從濂", "于大宝", "于丹", "于道泉", "于德海", "于尔岑", "于尔根", "于尔里克", "于爾里克", "于飞", "于风政", "于凤桐", "于凤至", "于格", "于根伟", "于光远", "于广洲", "于归", "于国", "于国桢", "于海洋", "于汉超", "于浩威", "于衡", "于洪区", "于化虎", "于会泳", "于慧", "于吉", "于佳卉", "于家", "于家堡", "于坚", "于堅", "于江震", "于嗟", "于杰", "于禁", "于靖", "于娟", "于军", "于軍", "于康震", "于克勒", "于孔兼", "于勒", "于里察", "于凌奎", "于冕", "于敏", "于明涛", "于默奥", "于娜", "于品海", "于奇库杜克", "于谦", "于謙", "于晴", "于仁泰", "于若木", "于山", "于慎行", "于式枚", "于树洁", "于帅", "于帥", "于双戈", "于思", "于斯", "于斯达尔", "于斯達爾", "于斯纳尔斯贝里", "于斯納爾斯貝里", "于斯塔德", "于素秋", "于台烟", "于涛", "于濤", "于特森", "于天仁", "于田", "于阗", "于韦斯屈莱", "于韋斯屈萊", "于伟国", "于西翰", "于湘兰", "于小彤", "于小伟", "于欣源", "于姓", "于秀敏", "于徐", "于学忠", "于荫霖", "于永波", "于右任", "于幼军", "于于", "于余曲折", "于馀曲折", "于玉立", "于远伟", "于越", "于樂", "于泽尔", "于赠", "于贈", "于占元", "于振", "于震环", "于震寰", "于正昌", "于正升", "于志宁", "于寘", "于子千", "汙蔑", "余碧芬", "余秉谚", "余炳贤", "余灿荣", "余车", "余发扬", "余干", "余歌沧", "余光", "余光生", "余光中", "余兰香", "余里", "余男", "余珮琳", "余三", "余三胜", "余三勝", "余上沅", "余思", "余思敏", "余天", "余威", "余威德", "余文", "余吾镇", "余贤明", "余宪宗", "余筱萍", "余姓", "余秀菁", "余雪兰", "余雪明", "余英时", "余余", "余苑绮", "余月", "余政宪", "余只", "余子", "余子明", "鱼干", "鱼鳞松", "鱼游釜底", "鱼游釜中", "鱼胄", "逾闲荡检", "愚暗", "舆尸", "与国同休", "宇宙志", "语汇", "语云", "玉斗", "玉里", "玉历", "玉米须", "玉制", "郁达夫", "郁馥", "郁离子", "郁李", "郁烈", "郁穆", "郁朴", "郁郁", "郁郁菲菲", "郁郁青青", "郁哉", "预制", "欲海", "欲壑难填", "欲火", "欲加之罪", "欲加之罪何患无词", "欲令智昏", "欲念", "欲女", "欲求", "欲求不满", "欲善其事必先利其器", "欲望", "欲障", "御敌", "御寒", "御寇", "御史大夫", "御侮", "御制", "寓禁于征", "愈合", "元后", "袁友范", "袁于令", "原钟", "原子钟", "远县才至", "远征", "愿而恭", "愿朴", "曰云", "约翰松", "月历", "月相表", "月岩", "岳飞", "岳坟", "岳父", "岳家", "岳珂", "岳庙", "岳母", "岳氏", "岳阳", "岳云", "岳丈", "煴斗", "云城区", "云尔", "云翻雨复", "云何", "云乎", "云然", "云吞", "云吞面", "云为", "云爲", "云溪", "云须", "云云", "芸辉", "芸苔", "芸薹", "耘荡", "允准", "陨获", "酝借", "蕴借", "熨斗", "杂合面儿", "杂酱面", "杂面", "杂志", "再制", "在核", "赞呗", "赞不绝口", "赞词", "赞辞", "赞道", "赞的", "赞歌", "赞个不", "赞口不", "赞乐", "赞了", "赞两句", "赞美", "赞佩", "赞赏", "赞颂", "赞叹", "赞歎", "赞我", "赞羡", "赞许", "赞扬", "赞一句", "赞一声", "赞一赞", "赞语", "赞誉", "赞自己", "脏发", "脏腑", "脏器", "臧谷亡羊", "凿岩", "凿岩机", "早起的鸟儿有虫吃", "早占勿药", "皂化", "皂荚", "造曲", "造岩矿物", "造钟", "噪动", "噪诈", "泽里可", "泽卤", "泽渗漓而下降", "扎成", "扎带", "扎囮", "扎尔达里", "扎根", "扎裹", "扎好", "扎脚", "扎结", "扎紧", "扎了", "扎马剌丁", "扎马鲁丁", "扎欧扎翁", "扎起", "扎上", "扎实", "扎铁", "扎下", "扎线带", "扎营", "扎在", "扎扎", "扎扎实实", "扎诈", "扎寨", "札夸威", "轧制", "炸毁", "炸酱面", "榨干", "詹志维", "薝卜", "展采", "占拜", "占卜", "占城", "占断", "占房", "占风使帆", "占凤", "占卦", "占候", "占课", "占了", "占了卜", "占梦", "占强", "占亲", "占人", "占上", "占射", "占身", "占世界", "占筮", "占万", "占星", "占验", "占有", "占有五不验", "占有欲", "占主", "占主导地位", "战略伙伴", "站干岸儿", "张斗辉", "张基郁", "张木松", "张琴松", "张三丰", "张文松", "张志", "张志家", "长发", "长干曲", "长干巷", "长胡", "长几", "长历", "长绳系景", "长绳系日", "长寿烟", "长吁", "长须", "长征", "丈母娘", "仗托", "赵坤郁", "照签", "遮复", "折尺", "折叠", "折好", "折合", "折痕", "折节读书", "折进", "折进来", "折进去", "折篷", "折裙", "折扇", "折台", "折梯", "折页", "折椅", "折纸", "折子", "折奏", "这出电影", "这出好戏", "这出剧", "这只", "这钟", "针砭", "针关", "针芥相投", "针灸", "针口", "针扣", "针芒", "针叶植物", "珍珠岩", "甄后", "枕借", "枕席", "缜致", "鬒发", "振荡", "振杰", "赈饥", "震荡", "镇荣里", "黮暗", "征尘", "征程", "征伐", "征帆", "征夫", "征服", "征驾", "征剿", "征敛", "征马", "征旆", "征辟", "征衫", "征戍", "征讨", "征途", "征衣", "征战", "征彸", "蒸干", "蒸面", "整出剧", "整发", "整发用品", "整只", "整周", "正梁", "正凶", "郑家钟", "郑凯云", "郑荣松", "郑易里", "郑余豪", "症结", "鄭凱云", "之欲", "之钟", "支杆", "支努干", "支烟", "只査", "只吃", "只鸡絮酒", "只立", "只轮不反", "只轮不返", "只日", "只身", "只声不出", "只手", "只言片语", "只言片字", "只眼", "只影", "只占", "只占吉", "只占神问卜", "只占算", "只准", "只字", "枝不得大于干", "织席", "直摆", "直发", "直系", "直系军阀", "植发", "植物志", "殖谷", "止谤莫如自修", "纸烟", "纸扎", "纸制", "指挥台", "指亲托故", "指水盟松", "志哀", "志诚君子", "志冲斗牛", "志悼", "志高气扬", "志庆", "志喜", "志异", "志之不忘", "制版", "制备", "制表", "制冰", "制播", "制材", "制茶", "制成", "制程", "制出", "制得", "制毒", "制法", "制革", "制剂", "制假", "制件", "制浆", "制冷", "制面", "制面具", "制坯", "制片", "制品", "制取", "制售", "制酸性", "制糖", "制陶", "制图", "制为", "制爲", "制鞋", "制盐", "制氧", "制药", "制衣", "制造", "制纸", "制钟", "制作", "制做", "治愈", "栉发工", "致密", "窒欲", "踬仆", "薙发", "中国制", "中环杯", "中仑", "中千世界", "中型钟", "中游", "忠人之托", "终端台", "终身有托", "钟摆", "钟被", "钟壁", "钟表", "钟表盘", "钟不", "钟差", "钟楚红", "钟的", "钟点", "钟顶", "钟鼎", "钟发音", "钟福松", "钟鼓", "钟关", "钟行", "钟好", "钟匠", "钟口", "钟快", "钟乐", "钟楼", "钟漏", "钟螺", "钟律", "钟慢", "钟没", "钟面", "钟鸣", "钟模", "钟纽", "钟盘", "钟珮瑄", "钟敲", "钟琴", "钟磬", "钟乳洞", "钟乳石", "钟山", "钟上", "钟身", "钟声", "钟速", "钟塔", "钟太", "钟体", "钟调", "钟停", "钟头", "钟王", "钟下", "钟相", "钟响", "钟形", "钟腰", "钟意", "钟有", "钟在寺里", "钟罩", "钟左右", "钟座", "种放", "种谷", "种师道", "种师中", "种子植物", "重复", "重见复出", "重罗面", "重折", "重制", "州里", "周报", "周而复始", "周二", "周回", "周会", "周记", "周济", "周杰", "周刊", "周考", "周历", "周六", "周末", "周年", "周期", "周全方便", "周人", "周人之急", "周日", "周三", "周上", "周数", "周四", "周岁", "周五", "周薪", "周休", "周一", "周游世界", "周召共和", "周中", "周周", "周转", "周走秀", "洲际杯", "咒愿", "胄甲", "胄科", "昼伏夜游", "皱别", "皱折", "朱贝尔", "朱笔", "朱唇", "朱唇皓齿", "朱脣皓齿", "朱轓皂盖", "朱干玉戚", "朱红", "朱卷", "朱俊", "朱口皓齿", "朱理安历", "朱仑街", "朱批", "朱色", "朱砂", "朱天文", "朱郁信", "朱谕", "珠斗烂班", "猪八戒吃人参果", "猪肝面", "猪脚面", "猪舌面", "猪只", "竹苞松茂", "竹几", "竹席", "竹制", "主梁", "主钟曲线", "煮面", "煮粥焚须", "芧栗", "住扎", "注标", "注冊", "注册", "注定", "注记", "注脚", "注解", "注名", "注明", "注批", "注入", "注入式敎学法", "注上", "注生娘娘", "注失", "注释", "注疏", "注文", "注销", "注译", "注云", "驻扎", "柱梁", "祝发", "祝赞", "铸钟", "筑邦", "筑北", "筑波", "筑肥", "筑后", "筑後", "筑前", "筑西", "筑阳", "筑陽", "筑州", "筑紫", "抓奸", "专鉴", "专征", "转速表", "转台", "转游", "转注", "装岩机", "装折", "壮面", "撞府冲州", "撞球杆", "撞钟", "准保", "准保护", "准保释", "准不准你", "准不准谁", "准不准他", "准不准它", "准不准她", "准不准我", "准不准许", "准此", "准分子", "准伏", "准假", "准将", "准决斗", "准考证", "准普尔", "准入", "准三后", "准算", "准尉", "准许", "准以", "准予", "准折", "准奏", "捉发", "捉奸", "桌几", "桌历", "浊积岩", "擢发", "姿采", "髭胡", "髭须", "子姜炒鸡", "子游", "子云", "子之丰兮", "梓里", "紫姜", "紫云", "紫云苗族布依族自治县", "自动表", "自然卷", "自制", "自制炸弹", "字汇", "渍已干", "宗周", "宗周钟", "综核", "总发", "总杆赛", "总杆数", "总汇", "总台", "总统杯", "纵横交布", "纵欲", "粽粑叶", "奏折", "足协杯", "足总杯", "组合服装", "祖冲之", "钻杆", "最后一天", "左光斗", "左里克", "左邻右里", "左手不托右手", "左右采之", "作幸", "坐标", "坐如钟", "坐台", "坐钟", "座钟", "做出", "做出好戏"];
    var _s2t_t = ["傌", "㑶", "偑", "㑳", "倲", "㑯", "儸", "𠗣", "劏", "劃", "劚", "噚", "喎", "㘔", "㘚", "㜄", "媰", "𡞵", "𡢃", "㜏", "孋", "𡠹", "㠏", "𡾱", "嵾", "幓", "㥮", "懤", "慺", "掆", "㩳", "撝", "擓", "擽", "㩜", "棡", "椲", "𣙎", "樢", "樫", "殰", "殨", "瀇", "濧", "灡", "澾", "濄", "𣾷", "瀰", "潚", "鸂", "燶", "煱", "獱", "璯", "𤫩", "𤪺", "䁻", "瞜", "碽", "磾", "稏", "穇", "𥢢", "筴", "籔", "䊷", "紬", "縳", "絅", "䋙", "䋚", "綐", "綵", "䋻", "䋹", "繿", "繸", "䍦", "䎱", "膞", "𦪙", "薵", "薳", "藭", "罃", "螮", "𧝞", "𧜗", "𧜵", "䙡", "襬", "訢", "鿁", "𧩙", "䜀", "讌", "貙", "𧵳", "䝼", "𧶧", "賰", "躎", "𨊰", "𨊸", "𨋢", "釾", "鏺", "䥱", "𨯅", "𨦫", "𨧜", "䥇", "鐯", "鐥", "钁", "䦛", "䦟", "靦", "𩞯", "𩣑", "騧", "䯀", "䱽", "𩶘", "鮣", "鰆", "鰌", "鰧", "䱷", "鳾", "鵁", "鴷", "鶄", "鶪", "鷈", "鷿", "龑", "萬", "與", "醜", "專", "業", "叢", "東", "絲", "丟", "兩", "嚴", "喪", "個", "豐", "臨", "爲", "麗", "舉", "麼", "義", "烏", "樂", "喬", "習", "鄉", "書", "買", "亂", "了", "爭", "於", "虧", "雲", "亙", "亞", "產", "畝", "親", "褻", "嚲", "億", "僅", "僕", "仇", "從", "侖", "倉", "儀", "們", "價", "仿", "衆", "優", "夥", "會", "傴", "傘", "偉", "傳", "俥", "俔", "傷", "倀", "倫", "傖", "僞", "佇", "體", "餘", "佛", "傭", "僉", "俠", "侶", "僥", "偵", "側", "僑", "儈", "儕", "儂", "儘", "俊", "俁", "儔", "儼", "倆", "儷", "倈", "儉", "修", "借", "債", "傾", "傯", "僂", "僨", "償", "儎", "儻", "儐", "儲", "儺", "僵", "兒", "克", "兌", "兗", "黨", "蘭", "關", "興", "茲", "養", "獸", "囅", "內", "岡", "冊", "寫", "軍", "農", "冬", "馮", "衝", "決", "況", "凍", "淨", "悽", "準", "涼", "凌", "減", "湊", "凜", "幾", "鳳", "鳧", "憑", "凱", "兇", "出", "擊", "鑿", "芻", "劃", "劉", "則", "剛", "創", "刪", "別", "剗", "剄", "刮", "制", "剎", "劊", "㓨", "劌", "剴", "劑", "剮", "劍", "剝", "劇", "勸", "辦", "務", "勱", "動", "勵", "勁", "勞", "勢", "勳", "勩", "勻", "匭", "匱", "區", "醫", "千", "升", "華", "協", "單", "賣", "卜", "佔", "盧", "滷", "臥", "衛", "卻", "卷", "巹", "廠", "廳", "歷", "厲", "壓", "厭", "厙", "龎", "廁", "釐", "廂", "厴", "廈", "廚", "廄", "廝", "縣", "叄", "參", "靉", "靆", "雙", "發", "變", "敘", "疊", "只", "臺", "葉", "號", "嘆", "嘰", "籲", "喫", "合", "吊", "同", "後", "向", "嚇", "呂", "嗎", "唚", "噸", "聽", "啓", "吳", "吶", "嘸", "囈", "嘔", "嚦", "唄", "員", "咼", "嗆", "嗚", "周", "詠", "嚨", "嚀", "噝", "吒", "諮", "鹹", "咽", "哄", "響", "啞", "噠", "嘵", "嗶", "噦", "譁", "噲", "嚌", "噥", "喲", "脣", "嘜", "嗊", "嘮", "啢", "嗩", "喚", "嘖", "嗇", "囀", "齧", "嘓", "囉", "嘽", "嘯", "喂", "噴", "嘍", "嚳", "囁", "噯", "噓", "嚶", "囑", "嚕", "噪", "囂", "回", "團", "園", "困", "囪", "圍", "圇", "國", "圖", "圓", "聖", "壙", "場", "阪", "壞", "坐", "塊", "堅", "壇", "壢", "壩", "塢", "墳", "墜", "壟", "壠", "壚", "壘", "墾", "堊", "墊", "埡", "墶", "壋", "塏", "堖", "塒", "壎", "堝", "垵", "塹", "墮", "壪", "牆", "壯", "聲", "殼", "壺", "壼", "處", "備", "復", "夠", "夫", "頭", "誇", "夾", "奪", "奩", "奐", "奮", "獎", "奧", "奸", "妝", "婦", "媽", "嫵", "嫗", "嬀", "姍", "姜", "奼", "婁", "婭", "嬈", "嬌", "孌", "娘", "娛", "媧", "嫺", "嫿", "嬰", "嬋", "嬸", "媼", "嬃", "嬡", "嬪", "嬙", "嬤", "孫", "學", "孿", "寧", "寶", "實", "寵", "審", "憲", "宮", "家", "寬", "賓", "寢", "對", "尋", "導", "壽", "將", "爾", "塵", "嘗", "堯", "尷", "屍", "盡", "局", "層", "屓", "屜", "屆", "屬", "屢", "屨", "嶼", "歲", "豈", "嶇", "崗", "峴", "嶴", "嵐", "島", "巖", "嶺", "嶽", "崬", "巋", "嶨", "嶧", "峽", "嶢", "嶠", "崢", "巒", "峯", "嶗", "崍", "嶮", "嶄", "嶸", "嶔", "嶁", "巔", "巨", "鞏", "巰", "幣", "布", "帥", "師", "幃", "帳", "簾", "幟", "帶", "幀", "席", "幫", "幬", "幘", "幗", "冪", "襆", "幹", "並", "幸", "廣", "莊", "慶", "牀", "廬", "廡", "庫", "應", "廟", "龐", "廢", "庵", "廎", "廩", "開", "異", "棄", "弒", "張", "彌", "弦", "弳", "彎", "彈", "強", "歸", "當", "錄", "彠", "彥", "彲", "彩", "徹", "徵", "徑", "徠", "御", "憶", "懺", "志", "憂", "念", "愾", "懷", "態", "慫", "憮", "慪", "悵", "愴", "憐", "總", "懟", "懌", "戀", "恆", "恤", "懇", "惡", "慟", "懨", "愷", "惻", "惱", "惲", "悅", "愨", "懸", "慳", "悞", "憫", "驚", "懼", "慘", "懲", "憊", "愜", "慚", "憚", "慣", "愈", "慍", "憤", "憒", "願", "懾", "憖", "懣", "懶", "懍", "戇", "戔", "戲", "戧", "戰", "戚", "戩", "戱", "戶", "才", "扎", "撲", "託", "扣", "執", "擴", "捫", "掃", "揚", "擾", "折", "撫", "拋", "摶", "摳", "掄", "搶", "護", "報", "擡", "抵", "擔", "拐", "擬", "攏", "揀", "擁", "攔", "擰", "撥", "擇", "掛", "摯", "攣", "掗", "撾", "撻", "挾", "撓", "擋", "撟", "掙", "擠", "揮", "撏", "挨", "挽", "挩", "撈", "損", "撿", "換", "搗", "據", "擄", "摑", "擲", "撣", "摻", "摜", "攬", "搵", "撳", "攙", "擱", "摟", "揯", "攪", "搜", "攜", "攝", "攄", "擺", "搖", "擯", "攤", "攖", "撐", "攆", "擷", "擼", "攛", "㩵", "擻", "攢", "敵", "敓", "斂", "斆", "數", "齋", "斕", "鬥", "斬", "斷", "無", "舊", "時", "曠", "暘", "昆", "曇", "暱", "晝", "曨", "顯", "晉", "曬", "曉", "曄", "暈", "暉", "暫", "𣈶", "暗", "曖", "曲", "術", "朱", "樸", "機", "殺", "雜", "權", "杆", "槓", "條", "來", "楊", "榪", "杯", "傑", "鬆", "板", "極", "構", "樅", "樞", "棗", "櫪", "梘", "棖", "槍", "楓", "梟", "櫃", "檸", "檉", "梔", "柵", "標", "棧", "櫛", "櫳", "棟", "櫨", "櫟", "欄", "樹", "棲", "慄", "樣", "核", "欒", "椏", "橈", "楨", "檔", "榿", "橋", "樺", "檜", "槳", "樁", "樳", "梁", "夢", "檮", "棶", "槤", "檢", "梲", "櫺", "槨", "槼", "櫝", "槧", "槶", "欏", "樿", "橢", "槮", "樓", "欖", "榲", "櫬", "櫚", "櫸", "樧", "檟", "檻", "檳", "櫧", "橫", "檣", "櫻", "櫫", "櫥", "櫓", "櫞", "檁", "歡", "歟", "歐", "欲", "殲", "歿", "殤", "殘", "殞", "殮", "殫", "殯", "毆", "毀", "轂", "畢", "斃", "氈", "毿", "毿", "氌", "氣", "氫", "氬", "氳", "匯", "漢", "湯", "洶", "沈", "溝", "沒", "灃", "漚", "瀝", "淪", "滄", "渢", "潙", "滬", "泛", "濘", "注", "淚", "澩", "瀧", "瀘", "濼", "瀉", "潑", "澤", "涇", "潔", "灑", "窪", "浹", "淺", "漿", "澆", "湞", "溮", "濁", "測", "澮", "濟", "瀏", "滻", "渾", "滸", "濃", "潯", "濜", "塗", "涗", "濤", "澇", "淶", "漣", "潿", "渦", "溳", "渙", "滌", "潤", "澗", "漲", "澀", "澱", "淵", "淥", "漬", "瀆", "漸", "澠", "漁", "瀋", "滲", "溫", "遊", "灣", "溼", "濚", "潰", "濺", "漵", "漊", "潷", "滾", "滯", "灩", "灄", "滿", "瀅", "濾", "濫", "灤", "濱", "灘", "澦", "漓", "瀠", "瀟", "瀲", "濰", "潛", "瀦", "瀂", "瀾", "瀨", "瀕", "灝", "滅", "燈", "靈", "竈", "災", "燦", "煬", "爐", "燉", "煒", "熗", "點", "煉", "熾", "爍", "爛", "烴", "燭", "煙", "煩", "燒", "燁", "燴", "燙", "燼", "熱", "煥", "燜", "燾", "熅", "薰", "愛", "爺", "牘", "犛", "牽", "犧", "犢", "狀", "獷", "獁", "猶", "狽", "獮", "獰", "獨", "狹", "獅", "獪", "猙", "獄", "猻", "獫", "獵", "獼", "玀", "豬", "貓", "蝟", "獻", "獺", "璣", "璵", "瑒", "瑪", "瑋", "環", "現", "瑲", "璽", "琺", "瓏", "璫", "琿", "璡", "璉", "瑣", "瓊", "瑤", "璦", "璸", "瓔", "瓚", "甕", "甌", "電", "畫", "暢", "疇", "癤", "療", "瘧", "癘", "瘍", "癧", "瘲", "瘡", "瘋", "皰", "痾", "症", "癰", "痙", "癢", "瘂", "癆", "瘓", "癇", "癡", "癉", "瘮", "瘞", "瘻", "癟", "癱", "癮", "癭", "癩", "癬", "癲", "皁", "皚", "皺", "皸", "盞", "鹽", "監", "蓋", "盜", "盤", "瞘", "眥", "矓", "睜", "睞", "瞼", "瞶", "瞞", "矚", "矩", "矯", "磯", "礬", "礦", "碭", "碼", "磚", "硨", "硯", "碸", "礪", "礱", "礫", "礎", "硜", "碩", "硤", "磽", "磑", "礄", "確", "磠", "礆", "礙", "磧", "磣", "鹼", "禮", "禡", "禕", "禰", "禎", "禱", "禍", "稟", "祿", "禪", "離", "私", "禿", "稈", "秋", "種", "祕", "積", "稱", "穢", "穠", "穭", "稅", "穌", "穩", "穡", "穗", "穭", "窮", "竊", "竅", "窵", "窯", "竄", "窩", "窺", "竇", "窶", "豎", "競", "篤", "筍", "筆", "筧", "箋", "籠", "籩", "築", "篳", "篩", "簹", "箏", "籌", "篔", "籤", "篠", "簡", "籙", "簀", "篋", "籜", "籮", "簞", "簫", "簣", "簍", "籃", "籛", "籬", "籪", "籟", "糴", "類", "秈", "糶", "糲", "粵", "糞", "糧", "糉", "糝", "餱", "餈", "系", "緊", "縶", "縕", "緪", "糹", "糾", "紆", "紅", "紂", "纖", "紇", "約", "級", "紈", "纊", "紀", "紉", "緯", "紜", "紘", "純", "紕", "紗", "綱", "納", "紝", "縱", "綸", "紛", "紙", "紋", "紡", "紵", "紖", "紐", "紓", "線", "紺", "紲", "紱", "練", "組", "紳", "細", "織", "終", "縐", "絆", "紼", "絀", "紹", "繹", "經", "紿", "綁", "絨", "結", "絝", "繞", "絰", "絎", "繪", "給", "絢", "絳", "絡", "絕", "絞", "統", "綆", "綃", "絹", "繡", "綌", "綏", "絛", "繼", "綈", "績", "緒", "綾", "緓", "續", "綺", "緋", "綽", "鞝", "緄", "繩", "維", "綿", "綬", "繃", "綢", "綯", "綹", "綣", "綜", "綻", "綰", "綠", "綴", "緇", "緙", "緗", "緘", "緬", "纜", "緹", "緲", "緝", "縕", "繢", "緦", "綞", "緞", "緶", "線", "緱", "縋", "緩", "締", "縷", "編", "緡", "緣", "縉", "縛", "縟", "縝", "縫", "縗", "縞", "纏", "縭", "縊", "縑", "繽", "縹", "縵", "縲", "纓", "縮", "繆", "繅", "纈", "繚", "繕", "繒", "繮", "繾", "繰", "繯", "繳", "纘", "罌", "網", "羅", "罰", "罷", "羆", "羈", "羥", "羨", "羣", "翹", "翽", "翬", "耮", "耬", "聳", "恥", "聶", "聾", "職", "聹", "聯", "聵", "聰", "肅", "腸", "膚", "骯", "餚", "腎", "腫", "脹", "脅", "胄", "膽", "背", "勝", "胡", "朧", "腖", "臚", "脛", "膠", "脈", "膾", "髒", "臍", "腦", "膿", "臠", "腳", "脫", "腡", "臉", "臘", "醃", "膕", "齶", "膩", "靦", "膃", "騰", "臏", "羶", "臢", "致", "輿", "舍", "艤", "艦", "艙", "艫", "艱", "豔", "藝", "節", "羋", "薌", "蕪", "蘆", "芲", "芸", "蓯", "葦", "藶", "莧", "萇", "蒼", "苧", "蘇", "苔", "檾", "薴", "蘋", "範", "莖", "蘢", "蔦", "塋", "煢", "繭", "荊", "薦", "薘", "莢", "蕘", "蓽", "萴", "蕎", "薈", "薺", "蕩", "榮", "葷", "滎", "犖", "熒", "蕁", "藎", "蓀", "蔭", "蕒", "葒", "葤", "藥", "蒞", "萊", "蓮", "蒔", "萵", "薟", "獲", "蕕", "瑩", "鶯", "蓴", "蘀", "蘿", "螢", "營", "縈", "蕭", "薩", "蔥", "蒕", "蕆", "蕢", "蔣", "蔞", "醟", "蒙", "藍", "薊", "蘺", "蕷", "鎣", "驀", "虆", "蔑", "薔", "蘞", "藺", "藹", "薀", "蘄", "蘊", "藪", "蘚", "蘊", "櫱", "虜", "慮", "虛", "蟲", "虯", "蟣", "蝨", "雖", "蝦", "蠆", "蝕", "蟻", "螞", "蠁", "蠶", "蠔", "蜆", "蠱", "蠣", "蟶", "蠻", "蟄", "蛺", "蟯", "螄", "蠐", "蛻", "蝸", "蠟", "蠅", "蟈", "蟬", "蠍", "螻", "蠑", "螿", "蟎", "蠨", "釁", "銜", "補", "表", "襯", "袞", "襖", "嫋", "褘", "襪", "襲", "襏", "裝", "襠", "褌", "褳", "襝", "褲", "襉", "褸", "襤", "襴", "見", "觀", "覎", "規", "覓", "視", "覘", "覽", "覺", "覬", "覡", "覿", "覥", "覦", "覯", "覲", "覷", "觴", "觸", "觶", "誾", "讋", "譽", "謄", "訁", "計", "訂", "訃", "認", "譏", "訐", "訌", "討", "讓", "訕", "訖", "託", "訓", "議", "訊", "記", "訒", "講", "諱", "謳", "詎", "訝", "訥", "許", "訛", "論", "訩", "訟", "諷", "設", "訪", "訣", "證", "詁", "訶", "評", "詛", "識", "詗", "詐", "訴", "診", "詆", "謅", "詞", "詘", "詔", "詖", "譯", "詒", "誆", "誄", "試", "詿", "詩", "詰", "詼", "誠", "誅", "詵", "話", "誕", "詬", "詮", "詭", "詢", "詣", "諍", "該", "詳", "詫", "諢", "詡", "譸", "誡", "誣", "語", "誚", "誤", "誥", "誘", "誨", "誑", "說", "誦", "誒", "請", "諸", "諏", "諾", "讀", "諑", "誹", "課", "諉", "諛", "誰", "諗", "調", "諂", "諒", "諄", "誶", "談", "讅", "誼", "謀", "諶", "諜", "謊", "諫", "諧", "謔", "謁", "謂", "諤", "諭", "諼", "讒", "諮", "諳", "諺", "諦", "謎", "諞", "諝", "謨", "讜", "謖", "謝", "謠", "謗", "諡", "謙", "謐", "謹", "謾", "謫", "譾", "謬", "譚", "譖", "譙", "讕", "譜", "譎", "讞", "譴", "譫", "讖", "谷", "豶", "貝", "貞", "負", "貟", "貢", "財", "責", "賢", "敗", "賬", "貨", "質", "販", "貪", "貧", "貶", "購", "貯", "貫", "貳", "賤", "賁", "貰", "貼", "貴", "貺", "貸", "貿", "費", "賀", "貽", "賊", "贄", "賈", "賄", "貲", "賃", "賂", "贓", "資", "賅", "贐", "賕", "賑", "賚", "賒", "賦", "賭", "齎", "贖", "賞", "賜", "贔", "賙", "賡", "賠", "賧", "賴", "賵", "贅", "賻", "賺", "賽", "賾", "贗", "贊", "贇", "贈", "贍", "贏", "贛", "赬", "趙", "趕", "趨", "趲", "躉", "躍", "蹌", "躒", "踐", "躂", "蹺", "蹕", "躚", "躋", "躊", "蹤", "躓", "躑", "躡", "蹣", "躕", "躥", "躪", "躦", "軀", "轀", "車", "軋", "軌", "軒", "軑", "軔", "轉", "軛", "輪", "軟", "轟", "軲", "軻", "轤", "軸", "軹", "軼", "軤", "軫", "轢", "軺", "輕", "軾", "載", "輊", "轎", "輈", "輇", "輅", "較", "輒", "輔", "輛", "輦", "輩", "輝", "輥", "輞", "輬", "輟", "輜", "輳", "輻", "輯", "轀", "輸", "轡", "轅", "轄", "輾", "轆", "轍", "轔", "辭", "闢", "辯", "辮", "邊", "遼", "達", "遷", "過", "邁", "運", "還", "這", "進", "遠", "違", "連", "遲", "邇", "逕", "跡", "適", "選", "遜", "遞", "邐", "邏", "遺", "遙", "鄧", "鄺", "鄔", "郵", "鄒", "鄴", "鄰", "鬱", "郟", "鄶", "鄭", "鄆", "酈", "鄖", "鄲", "酇", "醞", "醱", "醬", "酸", "釅", "釃", "釀", "醞", "採", "釋", "裏", "鑑", "鑾", "鏨", "釒", "釓", "釔", "針", "釘", "釗", "釙", "釕", "釷", "釺", "釧", "釤", "鈒", "釩", "釣", "鍆", "釹", "鍚", "釵", "鈃", "鈣", "鈈", "鈦", "鉅", "鈍", "鈔", "鍾", "鈉", "鋇", "鋼", "鈑", "鈐", "鑰", "欽", "鈞", "鎢", "鉤", "鈧", "鈁", "鈥", "鈄", "鈕", "鈀", "鈺", "錢", "鉦", "鉗", "鈷", "鉢", "鈳", "鉕", "鈽", "鈸", "鉞", "鑽", "鉬", "鉭", "鉀", "鈿", "鈾", "鐵", "鉑", "鈴", "鑠", "鉛", "鉚", "鉋", "鈰", "鉉", "鉈", "鉍", "鈮", "鈹", "鐸", "鉶", "銬", "銠", "鉺", "鋩", "錏", "銪", "鋮", "鋏", "鋣", "鐃", "銍", "鐺", "銅", "鋁", "銱", "銦", "鎧", "鍘", "銖", "銑", "鋌", "銩", "銛", "鏵", "銓", "鎩", "鉿", "銚", "鉻", "銘", "錚", "銫", "鉸", "銥", "鏟", "銃", "鐋", "銨", "銀", "銣", "鑄", "鐒", "鋪", "鋙", "錸", "鋱", "鏈", "鏗", "銷", "鎖", "鋰", "鋥", "鋤", "鍋", "鋯", "鋨", "鏽", "銼", "鋝", "鋒", "鋅", "鋶", "鐦", "鐗", "銳", "銻", "鋃", "鋟", "鋦", "錒", "錆", "鍺", "鍩", "錯", "錨", "錛", "錡", "鍀", "錁", "錕", "錩", "錫", "錮", "鑼", "錘", "錐", "錦", "鑕", "杴", "錈", "鍃", "錇", "錟", "錠", "鍵", "鋸", "錳", "錙", "鍥", "鍈", "鍇", "鏘", "鍶", "鍔", "鍤", "鍬", "鍾", "鍛", "鎪", "鍠", "鍰", "鎄", "鍍", "鎂", "鏤", "鎡", "鐨", "鎇", "鏌", "鎮", "鎛", "鎘", "鑷", "钂", "鐫", "鎳", "鎿", "鎦", "鎬", "鎊", "鎰", "鎵", "鑌", "鎔", "鏢", "鏜", "鏝", "鏍", "鏰", "鏞", "鏡", "鏑", "鏃", "鏇", "鏐", "鐔", "钁", "鐐", "鏷", "鑥", "鐓", "鑭", "鐠", "鑹", "鏹", "鐙", "鑊", "鐳", "鐶", "鐲", "鐮", "鐿", "鑔", "鑣", "鑞", "鑱", "鑲", "長", "門", "閂", "閃", "閆", "閈", "閉", "問", "闖", "閏", "闈", "閒", "閎", "間", "閔", "閌", "悶", "閘", "鬧", "閨", "聞", "闥", "閩", "閭", "闓", "閥", "閣", "閡", "閫", "鬮", "閱", "閬", "闍", "閾", "閹", "閶", "鬩", "閿", "閽", "閻", "閼", "闡", "闌", "闃", "闠", "闊", "闋", "闔", "闐", "闒", "闕", "闞", "闤", "隊", "陽", "陰", "陣", "階", "際", "陸", "隴", "陳", "陘", "陝", "隯", "隉", "隕", "險", "隨", "隱", "隸", "雋", "難", "僱", "雛", "雕", "讎", "靂", "霧", "霽", "黴", "霢", "靄", "靚", "靝", "靜", "面", "靨", "韃", "鞽", "韉", "韝", "韋", "韌", "韍", "韓", "韙", "韞", "韜", "韻", "頁", "頂", "頃", "頇", "項", "順", "須", "頊", "頑", "顧", "頓", "頎", "頒", "頌", "頏", "預", "顱", "領", "頗", "頸", "頡", "頰", "頲", "頜", "潁", "熲", "頦", "頤", "頻", "頮", "頹", "頷", "頴", "穎", "顆", "題", "顒", "顎", "顓", "顏", "額", "顳", "顢", "顛", "顙", "顥", "纇", "顫", "顬", "顰", "顴", "風", "颺", "颭", "颮", "颯", "颶", "颸", "颼", "颻", "飀", "飄", "飆", "飈", "飛", "饗", "饜", "飠", "飣", "飢", "飥", "餳", "飩", "餼", "飪", "飫", "飭", "飯", "飲", "餞", "飾", "飽", "飼", "飿", "飴", "餌", "饒", "餉", "餄", "餎", "餃", "餏", "餅", "餑", "餖", "餓", "餘", "餒", "餕", "餜", "餛", "餡", "館", "餷", "饋", "餶", "餿", "饞", "饁", "饃", "餺", "餾", "饈", "饉", "饅", "饊", "饌", "饢", "馬", "馭", "馱", "馴", "馳", "驅", "馹", "駁", "驢", "駔", "駛", "駟", "駙", "駒", "騶", "駐", "駝", "駑", "駕", "驛", "駘", "驍", "罵", "駰", "驕", "驊", "駱", "駭", "駢", "驫", "驪", "騁", "驗", "騂", "駸", "駿", "騏", "騎", "騍", "騅", "騌", "驌", "驂", "騙", "騭", "騤", "騷", "騖", "驁", "騮", "騫", "騸", "驃", "騾", "驄", "驏", "驟", "驥", "驦", "驤", "髏", "髖", "髕", "鬢", "鬹", "魘", "魎", "魚", "魛", "魢", "魷", "魨", "魯", "魴", "䰾", "魺", "鮁", "鮃", "鮎", "鱸", "鮋", "鮓", "鮒", "鮊", "鮑", "鱟", "鮍", "鮐", "鮭", "鮚", "鮳", "鮪", "鮞", "鮦", "鰂", "鮜", "鱠", "鱭", "鮫", "鮮", "鮺", "鯗", "鱘", "鯁", "鱺", "鰱", "鰹", "鯉", "鰣", "鰷", "鯀", "鯊", "鯇", "鮶", "鯽", "鯒", "鯖", "鯪", "鯕", "鯫", "鯡", "鯤", "鯧", "鯝", "鯢", "鮎", "鯛", "鯨", "鰺", "鯴", "鯔", "鱝", "鰈", "鰏", "鱨", "鯷", "鰮", "鰃", "鰓", "鱷", "鰍", "鰒", "鰉", "鰁", "鱂", "鯿", "鰠", "鰲", "鰭", "鰨", "鰥", "鰩", "鰟", "鰜", "鰳", "鰾", "鱈", "鱉", "鰻", "鰵", "鱅", "䲁", "鰼", "鱖", "鱔", "鱗", "鱒", "鱯", "鱤", "鱧", "鱣", "䲘", "鳥", "鳩", "雞", "鳶", "鳴", "鳲", "鷗", "鴉", "鶬", "鴇", "鴆", "鴣", "鶇", "鸕", "鴨", "鴞", "鴦", "鴒", "鴟", "鴝", "鴛", "鷽", "鴕", "鷥", "鷙", "鴯", "鴰", "鵂", "鴴", "鵃", "鴿", "鸞", "鴻", "鵐", "鵓", "鸝", "鵑", "鵠", "鵝", "鵒", "鷳", "鵜", "鵡", "鵲", "鶓", "鵪", "鵾", "鵯", "鵬", "鵮", "鶉", "鶊", "鵷", "鷫", "鶘", "鶡", "鶚", "鶻", "鶖", "鶿", "鶥", "鶩", "鷊", "鷂", "鶲", "鶹", "鶺", "鷁", "鶼", "鶴", "鷖", "鸚", "鷓", "鷚", "鷯", "鷦", "鷲", "鷸", "鷺", "䴉", "鸇", "鷹", "鸌", "鸏", "鸛", "鸘", "鹺", "麥", "麩", "麴", "麪", "麼", "黃", "黌", "黶", "黷", "黲", "黽", "黿", "鼂", "鼉", "鞀", "鼴", "齇", "齊", "齏", "齒", "齔", "齕", "齗", "齟", "齡", "齙", "齠", "齜", "齦", "齬", "齪", "齲", "齷", "龍", "龔", "龕", "龜", "䃮", "䥑", "鿓", "鎶", "𠁞", "儣", "𠌥", "俓", "㒓", "𠏢", "𣍐", "儭", "𠠎", "剾", "𠞆", "𪟖", "勑", "嗰", "哯", "噅", "㘉", "嚧", "囃", "𡅏", "𡃕", "𡄔", "𡄣", "㗲", "𡓾", "𡑭", "壗", "𡔖", "壈", "㜷", "㜗", "㜢", "孎", "孻", "𡮉", "𡮣", "𡳳", "𦘧", "嵼", "𡽗", "嶈", "嶘", "㢝", "㦛", "𢤱", "𢣚", "𢣭", "愻", "憹", "𢠼", "憢", "懀", "㦎", "懎", "𤢻", "戰", "𢷮", "𢶫", "摋", "擫", "𢹿", "擣", "斅", "斸", "曥", "𣋋", "𦢈", "腪", "脥", "臗", "槫", "桱", "欍", "𣠲", "楇", "橯", "樤", "樠", "欓", "㰙", "㯤", "𣞻", "檭", "𣝕", "欘", "𣠩", "殢", "𣯴", "𣯩", "氭", "湋", "潕", "㵗", "澅", "𣿉", "𪷓", "𤅶", "濆", "灙", "𤁣", "瀃", "熓", "㷍", "爄", "熌", "爖", "熚", "熉", "㷿", "𤒎", "𤓩", "熡", "㸇", "𤑳", "𤛮", "𤢟", "獩", "玁", "㺏", "瓕", "瓛", "𤳸", "癐", "𤸫", "㿗", "㿧", "皟", "麬", "䀉", "𥌃", "䀹", "𥊝", "瞤", "䁪", "䂎", "礒", "𥖅", "𥕥", "碙", "𥞵", "𥨐", "竚", "𥪂", "籅", "䉙", "籋", "篘", "𥵊", "𥸠", "䉲", "篸", "𥵃", "𥼽", "䊭", "𥽖", "𥿊", "緷", "綇", "綀", "繟", "緍", "縺", "緸", "𦂅", "䋿", "縎", "緰", "䌈", "𦃄", "䌋", "䌰", "縬", "繓", "䌖", "繏", "䌟", "䌝", "䌥", "繻", "䍽", "朥", "膢", "𦣎", "𦪽", "蓧", "䕳", "爇", "𦾟", "蘟", "檾", "𧕟", "䗿", "𧎈", "蠙", "蠀", "蠾", "𧔥", "䙱", "襰", "𧟀", "詀", "𧳟", "䞈", "買", "𧶔", "賬", "䝻", "賟", "贃", "𨇁", "躘", "𨄣", "𨅍", "𨈊", "𨈌", "䠱", "𨇞", "躝", "軉", "軗", "𨊻", "𨏠", "輄", "𨎮", "𨏥", "䢨", "𨣞", "𨣧", "𨢿", "𨣈", "𨤻", "鎷", "釳", "𨥛", "鈠", "鈋", "鈲", "鈯", "鉁", "龯", "銶", "鋉", "鍄", "𨧱", "錂", "鏆", "鎯", "鍮", "鎝", "𨫒", "鐄", "鏉", "鐎", "鐏", "𨮂", "䥩", "䦳", "𨳕", "𨳑", "閍", "閐", "䦘", "𨴗", "𨵩", "𨵸", "𨶀", "𨶏", "𨶲", "𨶮", "𨷲", "𨽏", "䧢", "䪏", "𩏪", "𩎢", "䪘", "䪗", "顂", "𩓣", "顃", "䫴", "颰", "𩗀", "䬞", "𩘹", "𩘀", "颷", "颾", "𩘺", "𩘝", "䬘", "䬝", "𩙈", "𩚛", "𩚥", "𩚵", "𩛆", "𩛩", "𩟐", "𩜦", "䭀", "䭃", "𩜇", "𩜵", "𩝔", "餸", "𩞄", "𩞦", "𩠴", "𩡣", "𩡺", "駎", "𩤊", "䮾", "駚", "𩢡", "䭿", "𩢾", "驋", "䮝", "𩥉", "駧", "𩢸", "駩", "𩢴", "𩣏", "𩣫", "駶", "𩣵", "𩣺", "䮠", "騔", "䮞", "驄", "騝", "騪", "𩤸", "𩤙", "䮫", "騟", "𩤲", "騚", "𩥄", "𩥑", "𩥇", "龭", "䮳", "𩧆", "䯤", "𩭙", "𩰀", "鬖", "𩯳", "𩰹", "𩳤", "𩴵", "魥", "𩵩", "𩵹", "鯶", "𩶱", "鮟", "𩶰", "鮕", "鯄", "䲖", "鮸", "𩷰", "𩸃", "𩸦", "鯱", "䱙", "䱬", "䱰", "鱇", "𩽇", "䲰", "鳼", "𩿪", "𪀦", "鴲", "鴜", "𪁈", "鷨", "𪀾", "𪁖", "鵚", "𪂆", "𪃏", "𪃍", "鷔", "𪄕", "𪈼", "𪄆", "𪇳", "䴬", "麲", "麨", "䴴", "麳", "𪋿", "䵳", "𪔵", "𪘀", "𪘯", "𠿕", "凙", "㔋", "勣", "𧷎", "㓄", "𠬙", "唓", "㖮", "嚛", "𠽃", "噹", "嘺", "嘪", "噞", "嗹", "㗿", "嘳", "𡃄", "㘓", "𡃤", "𡂡", "嚽", "𡅯", "囒", "圞", "墲", "埬", "堚", "塿", "𡓁", "壣", "𧹈", "孇", "嬣", "嬻", "孾", "寠", "㞞", "屩", "崙", "𡸗", "輋", "巗", "𡹬", "㟺", "巊", "巘", "𡿖", "幝", "幩", "廬", "㢗", "廧", "𢍰", "彃", "徿", "𢤩", "㦞", "憸", "𢣐", "𢤿", "𢯷", "摐", "擟", "𢶒", "掚", "撊", "㨻", "㩋", "撧", "𢺳", "攋", "㪎", "曊", "膹", "梖", "櫅", "欐", "檵", "櫠", "欇", "𣜬", "欑", "毊", "霼", "濿", "溡", "𤄷", "𣽏", "㵾", "灒", "熂", "煇", "𤑹", "𤓌", "爥", "𤒻", "𤘀", "𤜆", "犞", "獊", "𤠮", "㺜", "猌", "瑽", "瓄", "瑻", "璝", "㻶", "𤬅", "畼", "𤳷", "痮", "𤷃", "㿖", "𤺔", "瘱", "盨", "睍", "眝", "矑", "矉", "𥏝", "𥖲", "礮", "𥗇", "𥜰", "𥜐", "䅐", "䅳", "𥢷", "䆉", "竱", "鴗", "𥶽", "䉑", "𥯤", "䉶", "𥴼", "簢", "簂", "䉬", "𥴨", "𥻦", "𩏷", "糺", "䊺", "紟", "䋃", "𥾯", "䋔", "絁", "絙", "絧", "絥", "繷", "繨", "纚", "𦀖", "綖", "絺", "䋦", "𦅇", "綟", "緤", "緮", "䋼", "𦃩", "縍", "繬", "縸", "縰", "繂", "𦅈", "繈", "繶", "纁", "纗", "䍤", "羵", "𦒀", "䎙", "𦔖", "聻", "𦟼", "𦡝", "𦧺", "艣", "𦱌", "蔿", "蒭", "蕽", "蕳", "葝", "蔯", "蕝", "薆", "藷", "䗅", "蠦", "蟜", "𧒯", "蟳", "蟂", "蟘", "䙔", "襗", "襓", "襘", "襀", "襵", "𧞫", "覼", "覛", "𧡴", "𧢄", "覹", "䚩", "𧭹", "訑", "訞", "訜", "詓", "諫", "𧦝", "𧦧", "䛄", "詑", "譊", "詷", "譑", "誂", "譨", "誺", "誫", "諣", "誋", "䛳", "誷", "𧩕", "誳", "諴", "諰", "諯", "謏", "諥", "謱", "謸", "𧩼", "謉", "謆", "謯", "𧫝", "譆", "𧬤", "譞", "𧭈", "譾", "豵", "貗", "贚", "䝭", "𧸘", "賝", "䞋", "贉", "贑", "䞓", "䟐", "䟆", "𧽯", "䟃", "䠆", "蹳", "蹻", "𨂐", "蹔", "𨇽", "𨆪", "𨇰", "𨇤", "軏", "軕", "轣", "軜", "軷", "軨", "軬", "𨎌", "軿", "𨌈", "輢", "輖", "輗", "輨", "輷", "輮", "𨍰", "轊", "轇", "轐", "轗", "轠", "遱", "鄟", "鄳", "醶", "釟", "釨", "鈇", "鈛", "鏦", "鈆", "𨥟", "鉔", "鉠", "𨪕", "銈", "銊", "鐈", "銁", "𨰋", "鉾", "鋠", "鋗", "𫒡", "錽", "錤", "鐪", "錜", "𨨛", "錝", "錥", "𨨢", "鍊", "鐼", "鍉", "𨰲", "鍒", "鎍", "䥯", "鎞", "鎙", "𨰃", "鏥", "䥗", "鏾", "鐇", "鐍", "𨬖", "𨭸", "𨭖", "𨮳", "𨯟", "鑴", "𨰥", "𨲳", "開", "閒", "閗", "閞", "𨴹", "閵", "䦯", "闑", "𨼳", "𩀨", "霣", "𩅙", "靧", "䪊", "鞾", "𩎖", "韠", "𩏂", "韛", "韝", "𩏠", "𩑔", "䪴", "䪾", "𩒎", "顗", "頫", "䫂", "䫀", "䫟", "頵", "𩔳", "𩓥", "顅", "𩔑", "願", "顣", "䫶", "䫻", "𩗓", "𩗴", "䬓", "飋", "𩟗", "飦", "䬧", "餦", "𩚩", "飵", "飶", "𩛌", "餫", "餔", "餗", "𩛡", "饠", "餧", "餬", "餪", "餵", "餭", "餱", "䭔", "䭑", "𩝽", "饘", "饟", "馯", "馼", "駃", "駞", "駊", "駤", "駫", "駻", "騃", "騉", "騊", "騄", "騠", "騜", "騵", "騴", "騱", "騻", "䮰", "驓", "驙", "驨", "鬠", "𩯁", "鱮", "魟", "鰑", "鱄", "魦", "魵", "𩶁", "䱁", "䱀", "鮅", "鮄", "鮤", "鮰", "鰤", "鮆", "鮯", "𩻮", "鯆", "鮿", "鮵", "䲅", "𩸄", "鯬", "𩸡", "䱧", "鯞", "鰋", "鯾", "鰦", "鰕", "鰫", "鰽", "𩻗", "𩻬", "鱊", "鱢", "𩼶", "鱲", "鳽", "鳷", "鴀", "鴅", "鴃", "鸗", "𩿤", "鴔", "鸋", "鴥", "鴐", "鵊", "鴮", "𪀖", "鵧", "鴳", "鴽", "鶰", "䳜", "鵟", "䳤", "鶭", "䳢", "鵫", "鵰", "鵩", "鷤", "鶌", "鶒", "鶦", "鶗", "𪃧", "䳧", "𪃒", "䳫", "鷅", "𪆷", "鷐", "鷩", "𪅂", "鷣", "鷷", "䴋", "𪉸", "麷", "䴱", "𪌭", "䴽", "𪍠", "䵴", "𪓰", "䶕", "齧", "齩", "𫜦", "齰", "齭", "齴", "𪙏", "齾", "龓", "䶲", "㑮", "𠐊", "㛝", "㜐", "媈", "嬦", "𡟫", "婡", "嬇", "孆", "孄", "嶹", "𦠅", "潣", "澬", "㶆", "灍", "爧", "爃", "𤛱", "㹽", "珼", "璾", "𤩂", "璼", "璊", "𥢶", "絍", "綋", "綡", "緟", "𦆲", "䖅", "䕤", "訨", "詊", "譂", "誴", "䜖", "䡐", "䡩", "䡵", "𨞺", "𨟊", "釚", "釲", "鈖", "鈗", "銏", "鉝", "鉽", "鉷", "䤤", "銂", "鐽", "𨧰", "𨩰", "鎈", "䥄", "鑉", "閝", "韚", "頍", "𩖰", "䫾", "䮄", "騼", "𩦠", "𩵦", "魽", "䱸", "鱆", "𩿅", "齯", "僤", "墠", "娙", "嵽", "廞", "彄", "暐", "錀", "鑪", "𨧀", "𨨏", "𨭎", "𨭆", "鮈", "鮀", "鮠", "鯻", "鷭", "阿斗", "阿杰", "阿卡提里", "阿拉乾山脈", "阿拉克", "阿里", "阿里斯托芬", "阿孃", "阿森松島", "阿松森島", "阿托品", "阿咸", "阿紮倫卡", "阿芝特剋人", "阿芝特剋語", "哀弔", "哀慼", "哀輓", "埃夫伯里", "埃格爾松", "埃及曆", "埃克托", "埃拉托塞尼斯", "捱打", "捱到", "捱得", "捱餓", "捱過", "捱飢抵餓", "捱苦", "捱了", "捱罵", "捱滿", "捱磨", "捱日子", "捱三頂四", "捱上", "捱時間", "捱整", "捱揍", "矮桿品種", "矮几", "艾迴", "艾里賽宮", "艾里森", "艾瑞里", "愛彼錶", "愛睏", "愛麗捨宮", "愛慾", "礙難照准", "安納托利亞", "安瀋鐵路", "菴藹", "菴菴", "菴婪", "菴廬", "菴羅樹園", "菴舍", "案几", "案准", "暗吃一驚", "闇火", "暗釦", "闇藍髮", "闇劣", "闇亂", "闇倫", "闇昧", "闇冥", "闇莫", "闇淺", "闇然", "闇弱", "闇誦", "暗歎", "闇跳", "暗箱操作", "暗中行事", "盎盂相繫", "敖盪", "遨遊四海", "熬薑呷醋", "熬製", "翱遊四海", "鼇頭獨占", "奧里里亞", "奧里薩", "奧特朗托", "奧托", "八斗", "八蜡", "八里", "八辟", "八紮", "八隻", "八週", "八字鬍", "巴斗", "巴拉克", "巴拉松", "巴厘島", "巴里", "巴松管", "巴托麗", "巴托莉", "巴游", "芭托莉", "笆斗", "拔髮", "拔鬚", "拔宅上昇", "把飯叫饑", "吧檯", "吧托女", "白髮", "白髮銀鬚", "白粉麪", "白桿兵", "白乾", "白骨松", "白果松", "白鬍", "白里安", "白麪", "白面無鬚", "白皮松", "白千層", "白朮", "白松", "白鬚", "白雲", "白雲岩", "百不當一", "百穀", "百花曆", "百里", "百鍊", "百辟", "百葉", "百葉捲", "百紮", "百隻", "柏節松操", "擺佈", "擺盪", "擺烏龍", "擺鐘", "拜斗", "拜覆", "拜岳", "拜占庭", "頒佈", "斑岩", "闆闆", "板栗", "板岩", "辦公檯", "辦伙", "半乾", "半個世紀", "半里", "半托", "半隻", "拌麪", "綁紮", "棒子麪", "包乾", "包穀", "包伙", "包紮", "包占", "褒采一介", "褒讚", "薄倖", "寶丰", "寶里寶氣", "寶曆", "寶誌", "保險桿", "抱大足桿", "抱朴", "杯佈", "杯乾", "杯麪", "盃賽", "悲慼", "悲筑", "碑誌", "北斗", "北迴", "北里", "北征", "貝爾托內", "貝里", "貝那芬托", "貝冑", "備嘗艱苦", "備禦", "備註", "揹榜", "揹包", "背出", "揹出去", "揹帶", "揹負", "揹回", "揹饑荒", "揹筐", "揹來", "揹簍", "揹你", "揹人", "背痠", "揹他", "揹她", "揹我", "揹物", "揹小孩", "揹債", "揹着", "揹走", "被髮", "被複", "被人揹", "被頭散髮", "焙乾", "唄讚", "本里", "本週", "畚斗", "逼併", "鼻樑", "鼻菸", "比桿賽", "比干", "彼得里皿", "秕穀", "筆桿", "筆管麪", "筆捲", "筆禿墨乾", "畢昇", "庇里牛斯", "庇廕", "碧眼紫鬚", "弊倖", "壁誌", "壁鐘", "避凶就吉", "避凶趨吉", "嬖倖", "臂一捲", "砭鍼", "編髮", "編余", "編鐘", "鞭辟近裏", "鞭辟入裏", "扁擬穀盜蟲", "變速桿", "變質岩", "便吃乾", "便辟", "遍佈", "辨姦論", "辮髮", "標占", "標誌", "標緻", "標註", "標準", "標準尺寸", "標準單位", "標準桿", "錶板", "錶廠", "錶帶", "錶的嘀嗒", "錶的歷史", "錶店", "錶冠", "錶行", "錶殼", "錶快", "錶款", "錶鏈", "錶慢", "錶蒙子", "錶盤", "錶速", "錶停", "錶王", "表演", "表演慾", "錶針", "錶轉", "彆彆扭扭", "彆口氣", "彆扭", "彆拗", "彆氣", "彆強", "別日南鴻纔北去", "彆着", "別隻", "別緻", "彆嘴", "玢岩", "濱松市", "鬢髮", "冰斗", "冰磧岩", "冰岩", "餅乾", "稟覆", "併案", "幷包", "並不", "併不併", "併產", "併成", "併除", "併到", "併疊", "併發", "併購", "併骨", "併合", "併火", "並肩", "併肩子", "併兼", "併捲機", "併科", "併力", "併攏", "併名", "併入", "併紗", "併吞", "併網", "併爲", "併線", "併一不二", "併贓拿賊", "併贓治罪", "幷州", "病癒", "撥穀", "撥絃", "波盪", "波爾干", "波髮藻", "波拉克", "波里", "剝製", "菠蘿乾", "伯里克利", "伯孃", "伯余", "泊松", "泊松分佈", "博彙", "簸盪", "卜征", "逋髮", "補釦", "補註", "捕蟲植物", "捕風繫影", "捕虜岩", "捕影繫風", "哺餵", "不併", "不憚強禦", "不弔", "不負所托", "不幹", "不乾不淨", "不乾不淨吃了沒病", "不干己事", "不乾膠", "不干你", "不干他", "不干它", "不干她", "不干我", "不穀", "不好干預", "不氣干", "不捨", "不食乾腊", "不討采", "不通弔慶", "不畏強禦", "不繫", "不讚", "不佔", "不占吉凶", "不占算", "不占凶吉", "不準", "不准翻印", "不准沒", "不准你", "不准誰", "不准他", "不准它", "不准她", "不准問", "不准我", "佈擺", "布班尼斯瓦", "佈菜", "佈慈", "佈道", "佈德", "佈防", "布覆", "布干維爾", "佈崗", "佈告", "布穀", "布穀鳥鐘", "佈劃", "佈會", "佈教", "佈景", "佈局", "佈扣", "布拉克", "佈雷", "布里", "布里斯托", "佈列", "布魯", "布魯托", "布倫", "布倫托海", "佈滿", "布囊", "佈囊其口", "佈讓", "佈散", "佈哨", "佈設", "佈施", "佈勢", "佈署", "布托", "佈網", "布希總統", "佈下", "佈線", "佈雪", "佈一個", "佈疑陣", "佈於", "佈陣", "佈政", "佈置", "步步高昇", "步斗踏罡", "步罡踏斗", "擦乾", "猜三划五", "纔不", "纔出", "纔此", "纔打", "才當曹斗", "纔到", "纔得到", "纔得兩年", "纔等", "纔讀", "纔對", "纔多", "纔敢", "才幹", "纔乾杯", "纔乾旱", "纔乾淨", "纔乾透", "纔剛", "纔給", "纔跟", "纔夠", "纔怪", "纔過來", "纔過去", "纔行", "纔好", "纔回", "纔會", "纔將", "纔講", "纔開", "纔看", "纔可", "纔來", "纔料", "纔買", "纔沒", "纔拿", "才能", "纔能夠", "纔能勇敢追", "纔能有", "纔派", "纔起來", "才氣縱橫", "纔去", "纔上到", "纔上來", "纔上去", "纔始", "纔是", "纔鬆下", "纔算", "纔爲", "纔下來", "纔下去", "纔想", "纔像", "纔信", "纔要", "才用", "纔用到", "纔有", "纔再", "纔在", "纔則", "裁併", "裁製", "采采", "采椽不斲", "采地", "采蘩", "採風", "采風錄", "采葛", "採光", "采光剖璞", "采緝", "采及葑菲", "采蘭贈芍", "采烈", "采苓", "采綠", "采女", "采芑", "采芹", "采色", "采聲", "采詩", "採石", "采石之役", "采石之戰", "采石之戰", "采菽", "采頭", "采薇", "採薪", "采薪之疾", "采薪之憂", "采衣", "采邑", "採製", "彩筆", "彩筆生", "綵筆生花", "綵綢", "綵船", "綵帶", "綵緞", "綵鳳", "綵樓", "綵鸞", "綵女", "綵牌樓", "綵棚", "綵球", "彩色世界", "綵勝", "綵線", "綵衣", "綵繒", "菜乾", "菜薹", "蔡松坡", "蔘茸", "蔘綏", "蔘湯", "餐松啖柏", "餐松食柏", "餐松飲澗", "餐檯", "慘慼", "蒼髮", "蒼朮", "蒼松", "藏曆", "藏矇歌兒", "操縱桿", "操縱檯", "操作檯", "操作鐘", "曹郁芬", "漕輓", "艸木丰丰", "草菴", "草荐", "草簽", "草蓆", "測量桿", "杈桿兒", "茶几", "茶麪", "茶托", "茶已乾", "查覈", "查報表", "查表", "查不出", "查出", "查兌克", "查號臺", "查回", "查獲", "查價", "查卷", "查克拉", "查問出", "查無實據", "查修", "查詢檯", "查找週期", "搽穰捲兒", "察覈", "產後", "產後檢查", "產製", "剷板", "剷草", "剷剷", "剷車", "剷出", "剷除", "剷刀", "剷倒", "剷掉", "剷鬥", "剷斷", "剷煤", "剷平", "剷起", "剷搶", "剷球", "剷傷", "剷射", "剷土", "剷下", "剷雪", "剷刈", "剷鑿", "腸繫膜", "腸臟", "嚐遍", "嚐嚐", "嚐出", "嚐到", "嚐點", "嚐個", "嚐盡", "嚐來嚐去", "嚐了", "嚐了嚐", "嚐起來", "嚐鮮", "厂部", "唱唸", "超基性岩", "超級盃", "超讚", "朝鐘", "潮菸", "炒麪", "車斗", "車伕", "車裏", "車里雅賓斯克", "車仔麪", "扯麪", "扯縴", "撤併", "沉積岩", "陳沖", "陳漢昇", "陳杰", "陳鍊", "陳昇", "陳世杰", "陳萬松", "陳幸", "陳倖嫚", "陳尹杰", "陳瀛鐘", "陳郁秀", "晨鐘", "襯托", "稱歎", "稱讚", "蟶乾", "成岩作用", "呈准", "承製", "乘凶完配", "秤桿", "秤平斗滿", "吃板刀麪", "吃飽了飯撐的", "吃飽沒事幹", "吃不出", "吃不了", "吃出", "吃錯藥", "吃得出", "吃得了", "吃地面", "吃釘板", "吃豆乾", "吃飯別忘了種穀人", "吃飯家伙", "吃飯傢伙", "吃乾醋", "吃乾了", "吃掛絡兒", "吃過麪", "吃合家歡", "吃後悔藥", "吃回頭草", "吃幾碗乾飯", "吃薑", "吃盡", "吃虧就是佔便宜", "吃虧上當", "吃辣麪", "吃了", "吃裏扒外", "吃裏爬外", "吃麪", "吃槍藥", "吃敲才", "吃人蟲", "吃人一個蛋恩情無法斷", "吃傷了", "吃完麪", "吃閒飯", "吃閒話", "吃煙", "吃藥", "吃一頓挨一頓", "喫虧的是乖占便宜的是呆", "遲迴", "尺布斗粟", "尺寸斗粟", "齒髮", "齒危髮秀", "斥鹵", "赤繩繫足", "赤朮", "赤松", "沖鼻", "沖廁所", "沖茶", "沖沖", "沖牀工", "沖淡", "沖滌", "沖掉", "沖斷", "沖服", "沖溝", "衝冠", "衝冠髮怒", "沖和", "沖懷", "沖壞", "沖毀", "沖積", "沖劑", "沖襟", "沖決", "沖剋", "沖空機", "沖垮", "沖擴", "沖涼", "沖淋浴", "沖齡", "沖流", "沖昧", "沖模", "沖末", "沖默", "沖年", "沖泡", "沖人", "沖弱", "沖繩", "沖蝕", "沖刷", "沖水", "沖稅", "沖塌", "沖天", "沖田", "沖洗", "沖喜", "沖銷", "沖霄", "沖瀉", "沖虛", "沖印", "沖澡", "沖帳", "衝州撞府", "沖走", "舂穀", "虫部", "蟲吃牙", "抽斗", "抽乾", "抽菸", "仇讎", "丑旦", "丑角", "丑年", "丑牛", "醜奴兒", "丑日", "丑三", "丑時", "丑月", "偢采", "瞅下錶", "瞅下鐘", "臭侷", "臭燻燻", "出陳佈新", "出醜", "出醜狼藉", "齣兒", "出分子", "出家", "出家人", "出家人吃八方", "出口貿易", "出事情", "出水管", "出太陽", "齣戲", "出污泥而不染", "出征", "初征", "除舊佈新", "觸鬚", "搋麪", "川穀", "川后", "傳佈", "傳位于四太子", "船伕", "船孃", "船隻", "船鐘", "遄征", "串游", "窗明几淨", "窗明几亮", "牀蓆", "闖鍊", "創穫", "創鉅", "創傷後壓力", "創意盃", "吹髮", "吹乾", "吹鬍", "炊臼之鏚", "炊煙", "炊煙裊裊", "垂髮", "捶鍊", "捶檯拍凳", "錘鍊", "春捲", "春秋大一統", "春筍怒發", "春武里府", "脣乾", "脣燥舌乾", "脣若抹硃", "脣若塗硃", "脣似抹硃", "淳于", "醇郁", "詞采", "詞彙", "辭采", "辭彙", "慈悲喜捨", "磁製", "此仆彼起", "此係", "刺蔘", "刺干", "賜卹", "蔥蔥郁郁", "蔥薑蒜", "粗管麪", "粗鹵", "粗麪", "粗面岩", "粗製", "醋栗", "醋罈", "催併", "脆穀樂", "淬鍊", "存托股", "存摺", "寸髮千金", "錯彩鏤金", "搭乾鋪", "達欣盃", "答覆", "打捱", "打併", "打出", "打出弔入", "打出頭棒子", "打穀", "打鬨", "打簧錶", "打槳桿", "打卡鐘", "打游飛", "打游擊", "打製", "打中伙", "打鐘", "大本鐘", "大笨鐘", "大病初癒", "大不里士", "大采", "大沖", "大蟲", "大蟲不吃伏肉", "大蟲吃小蟲", "大丑", "大斗", "大發慈悲", "大夫松", "大鑒", "大金髮薹", "大蜡", "大里", "大理岩", "大曆", "大利麪", "大麻里", "大明曆", "大目乾連", "大辟", "大千世界", "大麴", "大水沖倒龍王殿", "大水沖倒龍王廟", "大水沖溺", "大同世界", "大武崙", "大咸", "大型鐘", "大凶", "大言非夸", "大衍曆", "大一統誌", "大讚", "大摺兒", "大隻", "大鐘", "大周后", "大專盃", "呆緻緻", "逮繫", "代表人物", "代簽", "帶髮修行", "帶凶", "待覈", "袋錶", "戴錶", "戴髮含齒", "戴維斯盃", "丹蔘", "丹干", "擔擔麪", "擔干紀", "擔仔麪", "單夫隻婦", "單位信托", "單絃", "單于", "單隻", "單週", "單子葉植物", "膽大如斗", "膽大於天", "但得一片橘皮吃且莫忘了洞庭湖", "但云", "彈珠檯", "彈子檯", "澹盪", "噹噹", "噹噹噹", "噹的一聲", "噹的一響", "當家", "當家纔知柴米價", "噹啷", "噹一聲", "當週", "擋禦", "黨蔘", "党懷英", "党進", "党太尉", "党太尉吃匾食", "党項", "党項", "党支書", "盪出", "盪船", "蕩蕩", "盪盪悠悠", "盪到", "盪滌", "盪風", "蕩覆", "盪垢滌污", "盪寒", "蕩檢逾閑", "盪酒", "盪開", "盪口", "盪來盪去", "蕩氣迴陽", "盪鞦韆", "蕩瑕滌穢", "盪漾", "盪悠悠", "盪舟", "叨唸", "搗鬼弔白", "倒八字鬚", "倒唸", "倒懸捱命", "禱唸", "盜鐘", "道里", "稻穀", "得采", "德干", "德高而譭來", "德里", "德勝頭迴", "的鐘", "燈綵", "等閒人物", "低盪", "低迴", "低價買進", "嘀嗒的錶", "滴乾", "滴里耷拉", "滴里搭拉", "滴里嘟嚕", "狄拉克", "狄志杰", "滌盪", "滌穢盪瑕", "滌瑕盪垢", "滌瑕盪穢", "詆譭", "牴觸", "牴牾", "抵禦", "地覆天翻", "地塞米松", "地無三里平", "地一捲", "地誌", "帝后", "遞迴", "第八齣", "第二齣", "第九齣", "第六齣", "第七齣", "第三齣", "第十齣", "第四齣", "第五齣", "第一齣", "顛顛仆仆", "顛覆", "顛仆", "攧攧仆仆", "點半鐘", "點多鐘", "點菸", "點鐘", "電錶", "電覆", "電鬍刀", "電腦檯", "電鬚刀", "電鐘", "電子錶", "電子鐘", "殿鐘自鳴", "刁斗", "刁姦", "貂覆額", "鵰鶚", "鵰悍", "鵰具座", "雕樑", "鵰翎", "鵰心雁爪", "弔詞", "弔奠", "吊斗", "弔古", "弔詭", "弔賀迎送", "弔鶴", "弔喉", "弔謊", "弔祭", "吊腳", "吊腳兒", "弔腳兒事", "弔拷", "弔客", "弔民", "弔旗", "弔撒", "弔喪", "弔書", "吊死", "弔死問孤", "弔死問疾", "弔頭", "弔慰", "弔文", "弔問", "弔孝", "弔唁", "弔宴", "弔喭", "弔腰撒跨", "弔影", "吊譽沽名", "弔者大悅", "弔紙", "吊鐘", "掉髮", "爹孃", "疊層岩", "蹀里蹀斜", "丁丑", "丁固生松", "叮噹", "頂針", "頂針捱住", "訂製", "釘釦", "定製", "東倉里", "東干", "東里", "東山里", "東昇", "東征", "東芝醫療繫", "東周", "東周鐘", "鼕鼕", "冬季世界", "董里府", "董氏封髮", "動盪", "棟樑", "都捨下", "斗柄", "斗車", "斗城", "斗儲", "斗大", "斗膽", "斗的", "斗燈", "斗店", "斗頓", "斗方", "鬥分子", "斗府", "斗概", "斗拱", "斗栱", "鬥鬨", "斗斛之祿", "斗箕", "斗極", "斗酒", "斗酒隻雞", "斗居", "斗絕", "斗魁", "斗笠", "斗量", "斗六", "斗羅大陸", "斗門", "斗南", "鬥牛", "斗牛之間", "斗蓬裝", "斗篷", "斗渠", "斗然", "斗山", "斗筲", "斗杓", "斗升", "斗食", "斗室", "斗數", "斗藪", "斗粟尺布", "斗粟囊金", "斗尾港", "斗紋", "斗香", "斗小馬", "斗菸絲", "斗齋", "斗帳", "斗折蛇行", "斗真", "斗重山齊", "斗轉參橫", "斗轉星移", "斗子", "豆乾", "豆干肉絲", "豆麪", "豆簽", "毒殭指", "獨佔鼇頭", "讀書三余", "賭檯", "杜老誌道", "杜雅里克", "短髮", "短几", "短鬚", "斷髮", "斷絃", "斷紙余墨", "鍛鍊", "堆案盈几", "對錶", "對摺", "對準", "對準錶", "對準鐘", "多采", "多層複", "多吃多佔", "多箇", "多少隻", "多凶少吉", "多隻", "奪盃", "俄製", "額我略曆", "噁心", "惡意", "惡意譭謗", "噁唑啉", "恩同父母", "恩准", "摁釦", "而云", "二斗", "二噁英", "二缶鐘惑", "二桿子", "二里", "二崙", "二孃", "二撇鬍", "二手菸", "二絃", "二葉松", "二隻", "二週", "髮辮", "發表", "發表慾", "髮鬢", "發佈", "發采揚明", "髮菜", "髮釵", "髮帶", "髮雕", "髮短心長", "髮匪", "髮膚", "發覆", "發乾", "髮根", "髮箍", "發光", "髮光可鑑", "發號", "發號佈令", "發皓齒", "髮際", "髮髻", "髮夾", "髮箋", "髮膠", "髮腳", "髮結", "髮姐", "髮禁", "髮捲", "髮卡", "發睏", "髮臘", "髮蠟", "髮廊", "髮量", "髮亂釵橫", "發矇", "發麪", "髮漂", "髮妻", "髮圈", "髮如飛蓬", "髮乳", "髮色", "髮紗", "發上", "髮上衝冠", "髮上指冠", "髮梢", "髮式", "髮飾", "髮梳", "髮束", "髮霜", "髮絲", "髮套", "髮挽雙髻", "髮網", "髮爲血之本", "髮尾", "髮屋", "髮香", "髮型", "髮鬚斑", "髮鬚都", "髮鬚皆", "髮鬚俱", "髮鬚已", "髮癬", "髮已霜白", "發引", "髮引千鈞", "髮纓", "髮踊沖冠", "髮油", "髮簪", "發展中國", "髮長", "髮針", "發政施仁", "髮指", "髮質", "髮狀", "廢后", "法拉托", "翻覆", "翻來覆去", "翻手作雲覆手雨", "翻檯", "翻雲覆雨", "煩複", "繁複", "繁鐘", "反剷", "反斗", "反反覆覆", "反覆", "反捲", "反亂併", "返還占有", "返里", "返照回光", "飯後", "飯後鐘", "飯糰", "氾濫", "泛水", "汎水淩山", "范冰冰", "范陳柏", "范成大", "范德格拉夫", "范德林特", "范德薩", "范德瓦耳斯", "范德維德", "范登堡", "范範之輩", "范甘迪", "范綱武", "范戈德", "范公偁", "范公堤", "范光羣", "范國銓", "范哈能", "范皓闐", "范洪森", "范家謙", "范嘉驊", "范姜", "范進", "范靖瑤", "范雎", "范可欽", "范寬", "范蠡", "范倫鐵諾", "范履霜", "范尼斯特魯伊", "范佩西", "范琪斐", "范綺馨", "范士丹", "范斯坦", "范特爾", "范特西", "范瑋琪", "範文", "范文程", "范文芳", "范文虎", "范文瀾", "范文藤", "范文同", "范文照", "范文正公", "范湘暄", "范小姐", "范曉萱", "范筱梵", "范欣妤", "范陽", "范曄", "范逸臣", "范增", "范張雞黍", "范正祥", "范織欽", "范植谷", "范植偉", "范志毅", "范仲淹", "方便麪", "方纔", "方几", "方里", "方岳", "方誌", "防颱", "防禦", "彷彿", "仿製", "放懞掙", "放鬆", "放松管制", "飛芻輓粒", "飛芻輓糧", "飛芻輓粟", "飛行鐘", "飛樑", "飛糧輓秣", "飛昇", "飛紮", "飛征", "非簽不可", "非游離輻射傷害", "菲捨爾", "肥皂", "肥皂荚", "肥皂剧", "肥皂丝", "肥筑方言", "肺臟", "廢后", "費爾干納", "費里克斯", "分半鐘", "分佈", "分多鐘", "分鐘", "分子鐘", "焚燬", "粉面", "粉面硃脣", "糞穢衊面", "丰標", "丰采", "豐城", "豐城貫斗", "丰度", "丰情", "丰茸", "丰容", "丰若有肌柔若無骨", "丰神", "豐溪里", "丰儀", "丰儀", "丰韻", "丰韻", "丰姿", "風采", "風斗", "風乾", "風颳", "風后", "風捲", "風入松", "風土誌", "風物誌", "風雲人物", "封后", "封妻廕子", "風采", "峯迴", "葑菲之采", "蜂后", "逢凶化吉", "縫製", "鳳皇于蜚", "鳳凰于蜚", "鳳台", "鳳尾松", "鳳占", "奉干", "奉公剋己", "佛里特", "佛曆", "佛羅里達", "佛鐘", "伕力", "伕役", "膚髮", "弗裏得裏希", "弗里德里希", "弗里敦", "弗里曼", "伏几", "伏屍流血", "扶幼週", "扶余", "拂盪", "拂鬚", "拂鐘無聲", "服飾週", "服務檯", "服裝週", "浮樑", "浮簽", "浮托", "浮游", "符采", "福生于微", "福廕", "撫屍", "撫尸慟哭", "撫松", "撫卹", "腑臟", "腐乾", "父母在不遠游", "負圖之托", "婦人生鬚", "附膻逐臭", "附膻逐穢", "附膻逐腥", "附註", "覆按", "覆敗", "覆杯", "覆被", "複本", "複比", "覆庇之恩", "覆蔽", "複壁", "複變函數", "覆瓿", "複測", "複查", "複查", "覆車", "複稱", "覆成", "覆呈", "覆幬", "複詞", "複次", "複道", "覆電", "覆鼎", "複對數", "復發", "複發率", "複發性", "複方", "複肥", "複分解", "複分數", "複分析", "複輔音", "複復", "覆蓋", "複閣", "複共軛", "複果", "覆海移山", "覆醢", "覆函", "複函數", "複合", "複覈", "復婚", "複婚制", "複基因", "複檢", "覆醬瓿", "覆蕉尋鹿", "複句", "複決", "覆軍", "複利", "複列", "複流", "覆鹿尋蕉", "覆鹿遺蕉", "覆露", "覆冒", "覆沒", "覆滅", "複名", "覆命", "複畝珍", "複目", "覆墓", "覆逆", "複拍子", "覆盤", "覆盆", "復辟", "複頻", "複品牌", "複平面", "複評", "複錢", "覆去翻來", "複仞年如", "複襦", "複賽", "複色", "覆上", "複審", "複式", "複試", "複視", "複述", "複數", "覆水", "複誦", "復甦", "覆文", "複習", "複線", "複相關", "覆校", "複寫", "覆信", "複姓", "複選", "複穴", "複循環發電", "複訓", "複鹽", "複眼", "複驗", "複葉", "複以百萬", "複議", "複意", "複音", "複印", "複用", "覆盂", "覆雨翻雲", "覆育", "復元", "複元音", "復員", "複員軍人", "複閱", "複韻", "複雜", "覆載", "覆在", "覆帳", "覆轍", "複診", "複殖目", "複殖吸蟲", "複製", "複種", "覆舟", "覆住", "復子明辟", "複字鍵", "覆宗", "複綜語", "覆餗", "傅里葉", "傅勻余", "富里", "覆雨翻雲", "馥郁", "該鐘", "改唸", "改簽", "肝臟", "扞禦", "桿秤", "桿刀", "桿菌", "桿茵", "桿直", "桿狀", "趕麪棍", "趕製", "敢斗了膽", "橄欖岩", "擀麪", "乾阿奶", "干礙", "乾熬", "乾巴", "乾爸", "乾白", "乾杯", "乾貝", "乾繃", "乾逼", "乾煸", "干扁豆角", "乾癟", "乾冰", "乾剝剝", "幹不", "乾不乾杯", "乾不乾淨", "乾布", "乾擦", "乾材", "乾菜", "乾草", "乾茶錢", "乾柴", "乾產", "乾唱", "乾炒牛河", "干城", "乾池", "乾芻", "乾茨臘", "乾此杯", "乾此罈", "乾醋", "乾脆", "乾村沙", "乾打雷", "乾打壘", "乾旦", "幹得", "乾得很", "乾得兩杯", "乾得三杯", "乾得一杯", "乾的", "乾燈盞", "乾等", "乾瞪眼", "乾地", "乾弟", "乾點", "乾電", "乾吊着下巴", "幹掉", "乾掉那杯", "乾掉那碗", "乾掉一杯", "乾掉一瓶", "乾掉一碗", "乾掉這杯", "乾掉這碗", "乾爹", "乾斷", "乾兒", "干犯", "乾飯", "乾肥", "乾粉", "乾乾", "乾綱", "乾稿", "干戈", "乾哥", "乾溝", "乾股", "乾卦", "乾館", "乾果", "幹過", "乾過杯", "乾過一杯", "乾過癮", "乾旱", "乾嚎", "乾號", "乾耗", "乾和", "乾涸", "乾紅", "乾餱", "乾花", "乾回付", "乾噦", "乾貨", "乾霍亂", "乾姬松茸", "乾急", "乾幾杯", "乾幾手", "乾幾碗", "乾季", "干將", "乾薑", "乾焦", "乾酵母", "乾結", "乾姐", "乾疥", "幹盡", "乾盡一杯", "乾盡一壺", "乾盡一罈", "乾盡一碗", "乾井", "乾淨", "乾淨", "干凱文", "乾咳", "乾渴", "乾刻版", "乾枯", "乾哭", "乾酪", "幹了", "乾了杯", "乾了這杯", "乾了這碗", "乾了這一杯", "乾了這一瓶", "乾雷", "乾冷", "乾禮", "乾李子", "干連", "乾涼", "乾糧", "乾兩杯", "乾量", "乾料", "乾撂臺", "乾裂", "乾餾", "乾落", "乾媽", "乾毛巾", "乾冒煙", "乾沒", "乾梅", "乾妹", "乾麪", "乾篾片", "幹那", "乾那杯", "乾那一杯", "干撓", "幹你", "幹你孃", "乾孃", "乾奴才", "乾暖", "乾女", "乾嘔", "乾皮症", "乾啤", "乾片", "乾撇下", "干鏚", "乾漆", "乾虔", "乾喬", "乾親", "干卿底事", "干卿何事", "乾球溫度", "乾渠", "干擾", "乾熱", "乾熱", "乾肉", "乾顙", "乾澀", "乾燒", "干涉", "乾伸舌", "乾生氣", "乾生受", "乾生子", "乾屍", "乾溼", "乾溼髮", "干時", "乾屎橛", "乾式", "乾手淨腳", "乾瘦", "乾數杯", "乾爽", "乾絲", "乾死", "乾闥婆", "乾颱", "乾薹", "乾罈子", "乾堂嬸", "乾啼", "乾田", "乾透", "乾圖", "乾土", "乾塢", "乾洗", "干係", "乾鮮", "乾象", "乾笑", "乾薪", "乾性", "干休", "乾癬", "乾血漿", "乾啞", "乾嚥", "乾眼", "乾曜", "干謁", "幹一", "乾一杯", "乾一罈", "乾一碗", "乾衣", "干邑", "乾癮", "乾硬", "乾又熱", "乾魚", "干與", "干預", "乾圓潔淨", "乾雲蔽日", "乾造", "乾燥", "乾躁", "乾宅", "乾站着", "乾折", "幹這", "乾這杯", "乾這一杯", "幹着", "乾着急", "干政", "干支", "乾支剌", "乾支支", "乾枝", "乾製", "乾重", "乾子", "乾姊", "乾坐着", "岡田准", "剛纔", "剛乾", "鋼釦", "鋼樑", "鋼製", "港製", "槓桿", "槓桿", "高柏松", "高几", "高麗蔘", "高良薑", "高清愿", "高郁淨", "高政昇", "高週波", "糕乾", "咯噹", "哥里", "割捨", "歌后", "歌鐘", "格里", "格里高利曆", "格里曆", "格列高利曆", "隔週", "葛里芬", "葛羅托斯基", "葛斯范桑", "個別教學", "箇舊", "個人電腦", "個人消費", "箇中", "個鐘", "各類鐘", "各里", "各簽", "根鬚", "根菸", "跟斗", "耕穫", "骾朴", "更待干罷", "更鐘", "工欲善其事", "工緻", "公佈", "公斗", "公里", "公曆", "公認會計准", "公孫丑", "公仔麪", "公制單位", "功緻", "宮裏", "宮里藍", "拱托", "共剷", "共和曆", "共同管道", "共禦外侮", "貢菸", "供製", "狗釦", "狗孃養的", "狗占馬坑", "購併", "購買慾", "咕咕鐘", "沽名干譽", "孤征", "辜濂松", "古蹟", "古里古怪", "古切里", "古書云", "古書云", "古絃", "古語云", "古云", "古鐘", "穀氨酸", "谷氨醯胺", "穀保家商", "穀倉", "穀艙", "穀草", "穀場", "穀旦", "穀道", "穀粉", "穀風", "穀圭", "穀貴餓農", "穀貴餓農穀賤傷農", "穀賤傷農", "穀糠", "穀殼", "谷口", "谷口耕岩", "穀類", "穀粒", "穀梁", "穀米", "穀苗", "穀皮", "穀人", "穀日", "穀神", "穀食", "穀穗", "穀物", "穀雨", "穀子", "股慄", "股栗膚粟", "骨幹分子", "骨灰罈", "骨罈", "骨頭裏掙出來的錢纔做得肉", "鼓盪", "鼓譟", "鶻崙吞棗", "故里", "故云", "顧藉", "颳大風", "颳倒", "颳得", "颳風", "刮鬍", "颳了", "颳起", "颳去", "刮鬚", "颳雪", "颳着", "颳走", "寡慾", "掛錶", "掛斗", "掛冠", "掛冠歸里", "掛角讀書", "掛曆", "掛麪", "掛鐘", "柺棒", "柺棍", "柺杖", "柺子", "怪里怪氣", "關弓與我确", "關係", "關岳", "關征", "觀光週", "觀葉植物", "官地爲寀", "官曆", "官准", "館穀", "管道昇", "管絃", "冠軍盃", "冠冑", "灌製", "光采", "光桿", "光鹵石", "光緻緻", "咣噹", "廣佈", "广部", "廣捨", "歸併", "硅質岩", "癸丑", "柜柳", "櫃檯", "貴价", "桂圓乾", "桂仔云", "郭采潔", "郭松燾", "郭子乾", "鍋伙", "國際漫遊撥接服務", "國曆", "國立臺灣圖書館", "國樑", "國之楨榦", "果乾", "果子乾", "裹紮", "過沖", "過樑", "過水麪", "哈里", "哈里里", "還辟", "海蔘", "海淀", "海乾", "海里", "海馬迴", "海松", "海鮮麪", "海于格松", "海岳名言", "含齒戴髮", "含油岩", "函覆", "韓復榘", "韓國製", "韓昇洙", "韓侂冑", "韓製", "漢彌登鐘", "旱乾", "旱菸", "捍禦", "行個方便", "行李捲", "行事曆", "行萬裏路勝讀萬捲書", "行佣", "行鍼", "行鍼佈線", "航海曆", "蒿里", "毫髮", "豪氣干雲", "好不容易纔", "好乾", "好睏", "好凶", "好一齣", "號誌", "皓齒硃脣", "皓髮", "喝采", "喝倒采", "喝乾", "禾穀", "合八字", "合併", "合成樂器", "合吃", "閤兒", "閤府", "閤家", "合曆", "合簽", "閤眼", "合於", "合于時宜", "閤中", "合作", "合作伙伴", "何干", "何杰金氏病", "何小昇", "和姦", "和麪", "和平里", "和絃", "劾繫", "河干", "河裏", "河裏孩兒岸上孃", "河昇鎮", "荷里活", "覈保", "覈報", "覈編", "覈撥", "覈查", "覈定", "覈對", "覈覆", "覈計", "覈價", "覈減", "覈批", "覈審", "覈實", "覈示", "覈收", "覈算", "覈銷", "覈驗", "覈准", "覈資", "覈字", "涸乾", "賀后罵殿", "赫弗里希", "鶴弔", "鶴髮", "鶴骨松姿", "黑暗世界", "黑髮", "黑亮髮", "黑麪", "黑松", "黑鬚", "黑曜岩", "很乾", "橫樑", "烘乾", "烘托", "烘燻", "烘雲托月", "烘製", "弘曆", "紅髮", "紅繩繫足", "紅絲暗繫", "紅松", "紅鬚綠眼", "紅葉", "紅葉盃", "紅鐘", "洪覆", "洪昇", "洪士杰", "洪适", "洪鐘", "鬨動", "鬨夥", "鬨鬧", "鬨然", "鬨堂", "鬨笑", "澒濛", "喉乾舌燥", "后安路", "後襬", "后北街", "后帝", "后髮座", "后妃", "后豐", "后豐", "后冠", "后海灣", "后海灣", "后皇", "后稷", "后角", "后街", "后里", "後孃", "后辟", "后平路", "后土", "后王", "后辛", "后羿", "後製", "後座", "後座繫", "厚朴", "忽捨下", "狐藉虎威", "鬍碴子", "胡吃海喝", "胡吃悶睡", "鬍匪", "胡椒", "胡椒麪", "胡杰", "胡里胡塗", "鬍髯", "鬍梢", "衚衕", "胡托莫", "鬍鬚", "胡云", "鬍渣", "鬍髭", "鬍子", "鵠髮", "餬口", "糊里糊塗", "虎皮松", "虎鬚", "護髮", "戽斗", "花菴詞選", "花采", "花髮老", "花崗岩", "花鬨", "花椒麪", "花捲", "花馬弔嘴", "花托", "花葯", "花鐘", "划不來", "划船", "划單人艇", "劃到", "划到岸", "划到江心", "劃得", "划得來", "划動", "劃過", "划過來", "划過去", "划行", "划槳", "划進", "划具", "划來", "劃了", "划了一會", "划龍舟", "划起", "划拳", "划雙人", "划水", "划算", "划艇", "划向", "劃一", "划一槳", "划着", "划子", "划走", "華髮", "華覈", "華里", "華嚴鐘", "嘩的", "嘩地", "嘩嘩", "嘩啦", "滑桿", "滑藉", "畫樑雕棟", "懷錶", "懷鐘", "環遊世界", "換擋桿", "換檔桿", "換髮", "換隻", "擐繫", "慌里慌張", "皇后", "皇極", "皇極曆", "皇曆", "皇辟", "黃東樑", "黃髮", "黃乾黑瘦", "黃金週", "黃俊杰", "黃曆", "黃梁", "黃樑美夢", "黃珮筑", "黃麴毒素", "黃鬚", "黃旭昇", "黃岩", "黃郁涵", "黃郁茹", "黃育杰", "黃鈺筑", "黃鐘", "黃鐘譭棄", "黃鈺筑", "湟潦生苹", "晃盪", "灰髮", "灰鬍", "灰濛", "輝綠岩", "輝長岩", "迴避", "迴飆", "迴腸", "回沖", "迴帶", "迴盪", "迴遞性", "迴風", "回覆", "迴光返照", "迴歸", "迴護", "迴環", "回黃倒皂", "迴魂仙夢", "回教會議組織", "回教世界", "回捲", "迴廊", "回曆", "迴流", "迴路", "迴鑾", "迴盲瓣", "迴夢", "迴清倒影", "迴圈", "迴繞", "回聲", "迴聲探測", "迴誦", "迴天", "迴腕", "迴文", "迴紋針", "迴斡", "迴翔", "迴響", "迴向", "迴心", "迴形夾", "迴旋", "迴穴", "迴雪", "迴陽蕩氣", "迴音", "迴應", "回佣", "迴游", "迴匝", "迴轉", "洄闇", "洄游", "譭棄", "譭誣", "燬犀", "燬炎", "譭譽", "譭鐘爲鐸", "彙報", "彙編", "彙集", "彙輯", "彙刊", "彙算", "彙映", "彙整", "彙總", "彙纂", "會弔", "會簽制度", "繪里", "繪製", "燴麪", "惠里香", "會干擾", "渾樸自然", "渾儀註", "餛飩麪", "活釦", "活塞桿", "火盃", "火併", "火柴桿", "火成岩", "火斗", "火炬松", "火繩桿", "火中取栗", "伙房", "伙伕", "伙食", "伙頭", "或繫之牛", "獲准", "霍克松", "霍里", "擊鐘", "饑荒", "饑饉", "饑民", "饑年", "機辟", "機器壓製", "機械錶", "機械鐘", "雞蛋裏挑骨頭", "雞蛋麪", "雞姦", "雞尸牛從", "雞絲", "雞絲麪", "雞腿麪", "雞隻", "積穀", "積穀防饑", "積金至斗", "基里巴斯", "基里蘭柯", "基岩", "齎志沒地", "箕斗", "稽覈", "激盪", "羈繫", "吉里", "吉普斯夸", "吉田松陰", "吉凶", "吉凶慶弔", "吉占", "極樂世界", "急征重斂", "集中托運", "集註", "几案", "幾齣", "几凳", "幾桿", "幾乎完全", "几几", "几淨窗明", "機率", "几面上", "几旁", "幾匹馬", "几上", "幾世紀", "几榻", "几席", "几筵", "几椅", "几杖", "幾隻", "几子", "己丑", "蟣蝨相弔", "脊樑", "計時錶", "紀里谷", "紀曆", "紀念", "紀念週", "技術單位", "技術潛水", "忌菸", "季咸", "蹟蹈", "覬倖", "祭弔", "祭尸", "加捲", "加拉干達", "加里", "加簽", "加註", "佳里", "傢伙", "傢俱", "傢俱", "傢什", "傢俬", "傢伙", "嘉柏隆里", "嘉穀", "夾註", "頰鬚", "甲冑", "賈后", "假髮", "賈后", "價電子", "架樑", "架鐘", "尖管麪", "姦盜邪淫", "姦非", "姦夫", "姦伏", "姦婦", "姦情", "姦殺", "姦屍", "姦通", "姦污", "姦凶", "姦淫", "堅緻", "間不容髮", "艱鉅", "艱苦備嚐", "監繫", "監製", "兼併", "兼籌幷顧", "兼容幷蓄", "兼收幷蓄", "箋註", "煎麪", "繭栗", "儉确之教", "檢覆", "剪綵", "剪髮", "剪其髮", "簡併", "簡朝崙", "簡余晏", "戩穀", "鹼性岩", "翦綵", "見覆", "見鑒", "件鐘", "建築機械", "荐饑", "荐居", "荐臻", "劍桿", "艦隻", "鑑核備查", "箭桿", "江采蘋", "江干", "江乾區", "薑餅", "薑茶", "薑桂", "薑還是老", "薑黃", "薑就是老", "薑辣", "薑老辣", "薑麻園", "薑末", "薑母", "薑片", "薑切片", "薑蓉", "薑石年", "薑是老", "薑絲", "薑湯", "薑糖", "姜文杰", "姜郁美", "薑愈老愈辣", "薑越老越辣", "薑汁", "殭蠶", "僵仆", "殭屍", "彊禦", "獎盃", "交併", "交鬨", "澆製", "膠捲", "焦乾", "焦穫", "礁岩", "嚼穀", "角里", "絞乾", "矯情干譽", "腳伕", "腳釦", "腳鍊", "腳痠", "腳註", "儌倖", "徼倖", "轎伕", "教學鐘", "嗟吁", "街里街坊", "節慾", "傑弗里喬叟", "傑里米", "傑里森", "杰倫", "杰特", "拮据", "結采", "結綵", "結髮", "結晶岩", "結釦", "結托", "結紮", "截髮", "解發", "解髮佯狂", "解釦", "解鈴繫鈴", "介係詞", "介冑", "戒菸", "藉草枕塊", "藉詞", "藉此", "藉端", "藉槁", "藉故", "藉卉", "藉機", "藉藉", "藉口", "藉寇兵", "藉甚", "藉手", "借托", "藉以", "藉由", "藉着", "藉助", "借箸", "藉箸代籌", "藉資", "斤斗", "金盃", "金錶", "金伯利岩", "金斗", "金髮", "金裏奇", "金鍊", "金崙溪", "金馬崙道", "金錢松", "金昇圭", "金屬桿", "金屬製", "金印如斗", "金鐘", "觔斗", "津樑", "筋斗", "緊繫", "緊緻", "錦囊佳製", "儘底下", "儘夠", "儘管", "儘教", "儘儘", "儘可", "儘快", "儘裏", "儘量", "儘落尾", "儘讓", "儘速", "儘先", "儘想", "儘性", "盡意", "儘意隨心", "儘有可能", "儘早", "儘子", "儘自", "近日無讎", "浸製", "禁燬", "禁菸", "禁慾", "經摺", "荊尸", "菁英盃", "旌卹", "驚歎", "驚讚", "驚鐘", "精采", "精奇里江", "精製", "精緻", "鯨鬚", "井榦", "頸鍊", "景緻", "警報鐘", "警示鐘", "警世鐘", "警鐘", "淨髮", "淨髮", "敬鑒", "敬輓", "敬菸", "迥然迴異", "揪髮", "揪鬚", "九穀", "九里", "九鍊成鋼", "九紮", "九隻", "酒乾掉", "酒乾盡", "酒乾了", "酒帘", "酒氣熏人", "酒麴", "酒罈", "酒已乾", "酒游花", "舊錶", "舊曆", "舊鐘", "就吃乾", "就係", "拘繫", "居里", "侷促", "侷限", "咀嚥", "莒光週", "舉薦征辟", "舉手可采", "鉅變", "鉅額", "鉅富", "鉅公", "鉅奸", "鉅艦", "鉅獎", "鉅款", "鉅虧", "鉅鹿", "鉅商", "鉅黍", "鉅貪", "鉅萬", "鉅細", "鉅獻", "鉅祥", "鉅野", "鉅業", "鉅債", "鉅製", "鉅著", "鉅子", "鉅作", "拒菸", "據榦而窺井底", "據云", "聚葯雄蕊", "捲包", "捲餅", "捲不起", "捲層雲", "捲纏", "捲成", "捲尺", "捲到", "捲動", "捲髮", "捲風", "捲鋼", "捲過", "捲回", "卷甲", "捲甲重來", "捲進", "捲開", "捲款", "捲來", "捲來捲去", "捲浪", "捲了", "捲簾", "捲攏", "捲落葉", "捲毛", "捲棚", "捲鋪蓋", "捲舖蓋", "捲起", "捲翹", "捲曲", "捲去", "捲刃", "捲入", "捲上", "捲舌", "捲縮", "捲逃", "捲筒", "捲圖", "捲土", "捲腿褲", "捲尾猴", "捲吸作用", "捲線器", "捲心", "捲袖", "卷鬚", "捲旋", "捲菸", "捲煙畫片", "捲菸", "捲揚", "捲一捲", "捲衣袖", "捲雲", "捲紙", "捲住", "捲走", "絕緣檯", "軍艦岩", "君子于役", "鈞覆", "鈞鑒", "菌托", "喀拉喀托火山", "卡迪拉克", "卡里", "卡里紮德", "卡洛里", "卡耶里", "開誠佈公", "開弔", "開發爲", "開發中國", "開鬨", "開伙", "揩乾", "揩檯抹凳", "凱迪拉克", "凱里", "鎧冑", "刊佈", "看錶", "看下錶", "看下鐘", "看鐘", "康采恩", "扛大樑", "抗禦", "炕蓆", "考覈", "烤乾", "珂里", "柯里", "科斗", "科托努", "科學麪", "蚵仔麪線", "可乾拭", "可乾飲", "可可托海", "可週", "剋薄", "剋剝", "克分子", "剋夫", "剋核", "克己奉公", "剋剋", "剋扣", "克拉克", "克萊", "克萊查克", "克里", "克里斯蒂安松", "克里斯托", "克里特克里島", "剋落", "剋期", "剋日", "剋嗇", "剋死", "剋星", "剋意", "克原子", "剋制", "刻半鐘", "刻多鐘", "刻鐘", "峇峇孃惹", "峇里島", "客製化", "墾丁盃", "墾複", "空濛", "空前絕後", "空前絕后後", "空鐘", "涳濛", "孔章望斗", "控捲", "控制", "控制桿", "控制檯", "控制慾", "口乾", "口唸", "口燥脣乾", "口占", "口鐘", "叩鐘", "釦環", "扣剋", "釦眼", "釦針", "釦子", "枯乾", "苦蔘", "苦瓜乾", "苦鹵", "褲釦", "夸誕", "夸爾", "夸父", "夸姣", "夸克", "夸麗", "夸毗", "夸人", "夸容", "夸特", "夸脫", "誇讚", "快吃乾", "快乾", "快捨下", "快速麪", "狂併潮", "奎松市", "坤錶", "崑劇", "崑崙", "崑腔", "崑曲", "崑山", "崑蘇", "崑調", "崑崙", "髡髮", "捆紮", "綑紮", "睏乏", "睏覺", "睏倦", "睏意", "括髮", "拉桿", "拉克施爾德鐘", "拉里", "拉鍊", "拉麪", "拉涅里", "拉昇", "拉縴", "拉祖里", "邋里邋遢", "腊之以爲餌", "蠟查", "蜡祭", "蜡月", "辣椒麪", "來複", "萊彩", "萊綵北堂", "萊里達", "欄干", "闌干", "藍采和", "藍髮", "藍托斯", "籃虹盃", "啷噹", "狼餐虎嚥", "狼飧虎嚥", "狼吞虎嚥", "浪蝶游蜂", "浪琴錶", "撈乾", "撈麪", "勞力士錶", "老闆", "老鵰", "老斗", "老鴰窩裏出鳳凰", "老薑", "老懞", "老孃", "老爺鐘", "樂器鐘", "泐覆", "勒里勒得", "瞭解", "瞭然", "瞭如", "瞭若指掌", "瞭望", "雷夫范恩斯", "誄讚", "淚乾", "冷地里", "冷麪", "厘金", "梨乾", "灕江", "灕然", "灕水", "灕湘", "禮斗", "禮讚", "李乾龍", "李連杰", "李連杰", "李鍊福", "李盟乾", "李三娘", "李鐘奭", "李鍾郁", "李鍾郁", "里昂", "里奧", "里包恩", "里貝利", "里布", "里程", "里程錶", "里黨", "里爾", "里耳", "里弗賽德", "里港", "里根", "里閈", "里豪", "里肌", "里加", "里賈納", "里居", "里君", "里科", "里克特", "里扣", "里拉", "里老", "里鄰長", "里路", "里閭", "里美", "里門", "里蒙諾夫", "里民", "里名", "里納", "里尼", "里弄", "里歐", "里奇蒙", "里契蒙", "里人", "里仁", "里舍", "里社", "里士滿", "里氏", "里斯", "里談巷議", "里特維寧科", "里瓦幾亞條約", "里維拉", "里希特霍芬", "里巷", "里胥", "里亞", "里諺", "里語", "里約", "里宰", "里長", "里正", "里茲", "理次髮", "理髮", "理個", "理個髮", "理事長盃", "理完髮", "理一次髮", "理一個髮", "力征", "曆本", "曆法", "曆紀", "曆命", "歷史人物", "曆始", "曆室", "曆書", "曆頭", "曆尾", "曆象", "曆獄", "曆元", "立方厘米", "利得彙", "利默里克", "利托", "利慾", "瀝乾", "儷采", "栗苞", "栗暴", "栗爆", "栗喇", "栗烈", "栗碌", "栗色", "栗鼠", "栗田雄介", "栗尾", "栗薪", "栗鑿", "栗子", "礫岩", "粒變岩", "連桿", "連三併四", "連繫", "蓮鬚", "聯賽盃", "聯繫", "鎌倉", "鍊度", "鍊鋼", "鍊汞", "鍊金", "鍊鋁", "鍊貧", "鍊師", "鍊鐵", "鍊銅", "鍊冶", "煉製", "鍊甲", "鏈釦", "鍊形", "鍊墜", "良价", "涼麪", "涼蓆", "樑棟", "樑架", "樑龍", "梁木", "樑木其壞", "樑上", "梁文沖", "樑柱", "樑子", "兩當一", "兩撇鬍", "兩紮", "兩隻", "兩週", "亮鐘", "晾乾", "遼瀋", "寮寀", "燎髮", "料斗", "廖于誠", "鄰里", "林沖", "林芳郁", "林乾閔", "林杰樑", "林俊杰", "林榮松", "林松", "林郁方", "林占梅", "林正杰", "林鐘", "臨海水土誌", "淋沖", "磷酸鹽岩", "鱗游", "躪藉", "靈蹟", "靈脩", "靈慾", "淩策", "淩昌焕", "淩馳", "淩河", "淩惠平", "凌藉", "淩濛初", "淩如焕", "淩十八", "淩氏", "淩水", "淩統", "淩退思", "淩小姐", "淩姓", "凌雲", "淩云翰", "凌志", "淩志美", "零隻", "領釦", "領檯", "領袖慾", "令狐沖", "令岳", "溜鬚", "劉峯松", "劉松藩", "劉松年", "劉偉杰", "劉倖如", "劉占吉", "劉志昇", "留髮", "流佈", "流乾", "流紋岩", "流血浮尸", "流血漂鹵", "柳斌杰", "柳昇耀", "六沖", "六穀", "六通四辟", "六絃", "六鬚鮎", "六鬚鮎", "六慾", "六紮", "六隻", "六週", "龍捲", "龍蝦麪", "龍鬚", "龍鬚麪", "龍遊", "龍游淺水", "礱穀機", "鏤金錯采", "漏斗", "嚕嚕囌囌", "嚕囌", "盧貝松", "盧郁佳", "蘆蓆", "鹵簿", "鹵代烴", "鹵地", "鹵鈍", "鹵化", "鹵莽", "滷麪", "滷牛肉", "鹵人", "鹵素", "滷製", "鹵族", "甪里", "陸均松", "陸游", "彔彔", "錄製", "鹿門采藥", "路里", "路誌", "露覆", "亂髮", "亂鬨", "崙背", "崙豐村", "輪迴", "輪姦", "羅漢松", "羅興樑", "蘿蔔", "蘿蔔乾", "螺桿", "螺旋麪", "裸岩", "犖确", "洛克希德馬丁", "洛鐘東應", "絡腮鬍", "落髮", "落腮鬍", "落托", "落葉", "落葉松", "落葉植物", "囉囉囌囌", "囉囌", "驢蒙虎皮", "閭里", "呂后", "呂宋菸", "呂岩", "呂后", "郘鐘", "捋采", "鋁製", "屢仆屢起", "律曆志", "綠髮", "麻桿", "麻醬麪", "麻栗坡", "馬錶", "馬丁杜里荷", "馬伕", "馬乾", "馬後", "馬后練服", "馬拉巴栗", "馬拉松", "馬里安納海溝", "馬里克", "馬里蘭", "馬里內斯科", "馬里奇", "馬鬣松", "馬尼托巴", "馬普托", "馬斯垂剋", "馬尾松", "馬紮", "馬占山", "碼錶", "馬占山", "埋佈", "埋頭尋錶", "埋頭尋鐘", "買進對沖", "買菸", "邁科里", "麥卡托", "麥科里", "麥托姆", "賣柺", "賣姦", "脈岩", "顢里顢頇", "滿布", "滿佈疑雲", "滿頭洋髮", "滿洲里", "曼尼托巴省", "慢嚥", "忙併", "莽鹵", "毛髮", "毛薑", "毛里求斯", "毛里塔尼亞", "茂松", "貿易伙伴", "沒采", "沒地里", "沒爹沒孃", "沒精打采", "沒量斗", "沒簽", "沒摺至", "梅乾", "梅里", "媒人口無量斗", "每隻", "每週", "美髮", "美國製", "美后", "美里", "美崙", "美製", "美洲盃", "門斗", "悶錶", "矇蔽", "懞懂", "蒙古大夫", "濛鴻", "矇混", "矇聵", "矇矓", "矇昧", "濛昧不清", "濛濛", "懞懞懂懂", "矇矇黑", "矇矇亮", "矇矇矓矓", "矇騙", "矇事", "濛汜", "濛鬆雨", "矇瞍", "蒙特卡洛", "矇頭", "蒙托羅拉", "濛霧", "矇眼", "蒙在", "矇在鼓裏", "懞直", "矇住", "孟德爾松", "夢迴", "夢蘭叶吉", "夢繫", "夢有五不占", "瀰漫", "彌矇", "瀰瀰", "瀰山遍野", "迷姦", "迷濛", "縻繫", "米德爾伯裏", "米穀", "米里", "米麪", "祕製", "密佈", "密摺", "密緻", "棉製", "免冑", "麪霸", "面白無鬚", "麪包", "麪餅", "麪茶", "麪廠", "麪點", "麪店", "麪坊", "麪肥", "麪粉", "麪缸", "麪疙瘩", "麪館", "麪糊", "麪灰", "麪價", "麪漿", "麪醬", "麪餃", "麪筋", "麪碼兒", "麪坯兒", "麪皮", "麪票", "麪人", "麪食", "麪塑", "麪攤", "麪湯", "麪條", "麪糰", "麪碗", "面向", "面嚮對象的技術", "面嚮對象語言", "面譽", "面譽背譭", "面皂", "麪杖", "面子", "麪子藥", "苗栗", "秒錶", "秒鐘", "民政里", "民族誌", "閔采爾", "閔凶", "抿髮", "名錶", "名覆金甌", "名錄服務", "明查暗訪", "明窗淨几", "明窗淨几", "明覆", "明伙畫供", "明釦", "明瞭", "鳴鐘", "冥淩", "冥濛", "溟濛", "謬讚", "摸鐘", "摹紮特", "模範人物", "模里西斯", "模製", "摩根費里曼", "摩里西斯", "摩托", "磨變岩", "磨鍊", "磨製", "魔錶", "魔術數字", "抹乾", "莫干山", "莫吉托", "莫里", "莫余毒也", "墨盪子", "墨斗", "墨沈", "墨瀋未乾", "默唸", "某隻", "母后", "母鐘", "木樑", "木偶戲紮", "木製", "木鐘", "目牛游刃", "墓誌", "幕後人物", "穆罕默德曆", "拿坡里", "拿破崙", "拿下錶", "拿下鐘", "拿賊要贓拿姦要雙", "哪一齣", "哪隻", "那齣電影", "那齣好戲", "那齣劇", "那個", "那個貓兒不吃腥", "那捲", "那隻", "納采", "乃里", "奶捲", "奶孃", "男用錶", "南宮适", "南迴", "南京鐘", "南里", "南山盃", "南征", "南筑", "難捱", "難捨", "難嚥", "難廕", "鬧錶", "鬧鬨", "鬧鐘", "內鬨", "內臟", "內製", "能捨", "能征慣戰", "能征善戰", "尼采", "尼克", "尼克松", "泥灰岩", "泥岩", "泥質岩", "倪嗣沖", "擬製", "你纔子發昏", "你斗了膽", "你係", "逆鐘", "惄如調饑", "拈鬚", "年穀", "年曆", "捻鬚", "捻鍼", "唸啊", "唸吧", "唸白", "唸錯", "唸叨", "唸到", "唸的", "唸對", "唸佛", "唸經", "唸了", "念念", "唸唸有詞", "唸詩", "唸書", "唸誦", "唸完", "唸曰", "唸咒", "唸作", "孃的", "孃兒", "孃家", "孃舅", "孃老子", "孃親", "孃胎", "孃姨", "釀製", "鳥松", "嫋嫋", "裊裊炊煙", "裊裊上升", "裊繞", "裊窕", "尿斗", "捏製", "甯成子", "甯悼子", "甯斧成", "甯浩", "甯惠子", "甯猛力", "甯戚", "甯調元", "甯武子", "甯越", "甯中則", "甯莊子", "凝灰岩", "凝鍊", "擰乾", "牛驥同皂", "牛柳麪", "牛肉麪", "牛隻", "鈕釦", "鈕釦", "拗彆", "農曆", "農民曆", "濃髮", "濃郁", "醲郁", "弄乾", "弄鬼弔猴", "弄麪吃", "奴兒干", "怒髮衝冠", "怒髮沖天", "怒氣沖發", "暖盪撩鍋", "諾里", "糯米糰", "女丑", "女生外嚮", "歐伯托", "歐幾里得", "歐幾里德", "歐里庇得斯", "歐里桑", "歐洲盃", "藕覆", "帕累托法則", "帕累托最優", "拍出", "拍齣好戲", "拍檯拍凳", "排檔桿", "排骨麪", "排鬚", "徘迴", "潘威誌", "潘岳", "盤迴", "盤松", "蟠采", "蟠龍松", "旁註", "匏繫", "跑檯子", "泡麪", "泡製", "炮製", "陪弔", "培爾松", "培里克里斯", "裴里諾", "裴松之", "佩脫拉克", "配膳檯", "配製", "砰噹", "烹製", "彭咸", "彭于晏", "蓬髮", "篷蓋佈", "膨土岩", "碰鐘", "批覆", "批覈", "批迴", "批註", "批准", "披髮", "披覆", "披紅掛綵", "披頭散髮", "披榛採蘭", "劈里", "皮特拉克", "皮托管", "皮製", "毗婆尸佛", "疲睏", "埤塘里", "脾臟", "匹馬隻輪", "辟惡", "辟咡", "辟穀", "辟舉", "辟君三舍", "辟歷施鞭", "辟纑", "辟邏", "辟命", "辟匿", "闢辟", "辟然", "辟人之士", "辟色", "辟世", "辟書", "辟違", "辟邪", "辟言", "辟易", "辟淫", "辟引", "辟雍", "辟廱", "辟召", "辟支佛", "辟芷", "偏信則闇", "偏倖", "片麻岩", "片言隻語", "片岩", "片語隻辭", "漂盪", "漂游", "飄遊四海", "撇弔", "拚捨", "品嚐", "品彙", "平康里", "平面測量", "評覈", "評註", "蘋果電腦", "苹縈", "憑弔", "憑几", "憑藉", "憑閑", "憑摺", "頗覆", "破錶", "破壞慾", "掊斗折衡", "裒剋", "撲鼕", "鋪蓋捲兒", "仆倒", "仆地", "僕伕", "仆街", "仆然", "葡萄乾", "蒲松齡", "朴茨茅斯", "朴刀", "朴恩惠", "朴鳳柱", "朴父", "朴吉淵", "朴槿惠", "朴京琳", "朴璐美", "朴茂", "朴詩妍", "朴世莉", "朴樹", "朴泰桓", "朴硝", "朴新陽", "朴宣英", "朴永訓", "朴仔樹", "朴贊浩", "朴真熙", "朴正恩", "朴正熙", "朴正祥", "朴志胤", "朴智星", "朴忠", "朴周永", "朴資茅斯", "朴子", "埔里", "普勒托利亞", "普里", "普利艾托", "七里河", "七里香", "七娘", "七孃媽", "七罈", "七絃", "七紮", "七隻", "七週", "淒滄", "淒寒", "淒冷", "淒厲", "淒涼", "淒雨", "戚里", "慼慼", "慼墅堰", "欺矇", "齊后破環", "齊梁世界", "齊王捨牛", "其次辟地", "奇盃", "奇蹟", "奇里安", "奇台", "棋佈", "棋羅星佈", "啓發", "啓發式教學法", "起鬨", "起撲桿", "氣衝斗牛", "氣衝牛斗", "氣克斗牛", "氣若游絲", "氣勢熏灼", "氣吞牛斗", "氣揚采飛", "氣一衝", "棄捨", "磧鹵", "恰纔", "千層麪", "千迴百折", "千迴百轉", "千鈞", "千鈞一髮", "千里", "千紮", "千隻", "遷思迴慮", "牽繫", "牽一髮", "鉛製", "慳吝苦剋", "謙沖", "簽報", "簽唱", "簽呈", "簽出", "簽單", "簽到", "簽得", "簽訂", "簽定", "簽賭", "簽發", "簽過", "簽好", "簽結", "簽了", "簽名", "簽派室", "簽入", "簽上", "簽收", "簽書會", "簽署", "簽退", "簽妥", "簽完", "簽爲", "簽下", "簽些", "簽寫", "簽押", "簽印", "簽有", "簽約", "簽在", "簽章", "簽帳", "簽着", "簽證", "簽証", "簽註", "簽字", "前車覆後車戒", "前車之覆", "前後左右", "前仆後繼", "前仆後起", "錢穀", "乾清宫", "乾象曆", "潛水鐘", "潛游", "槍桿", "槍托", "強姦", "強投松", "強嚥", "強制保險", "磽确", "敲鐘", "喬松", "蕎麥麪", "蕎麪", "橋樑", "巧曆", "竊占", "竊鐘掩耳", "侵併", "侵入岩", "親孃", "親倖", "親征", "秦少游", "琴斷朱絃", "琴桿", "琴絃", "琴鐘", "噙齒戴髮", "青帘", "青苹", "青山一髮", "青松", "輕釦", "氫鹵酸", "傾覆重器", "清查不當黨產", "清晨盃", "清桿運動", "清檯", "情采", "情繫", "情慾", "慶弔", "慶曆", "磬鐘", "窮髮", "窮里", "邱富郁", "邱郁婷", "秋不乾", "秋髮", "鞦韆", "秋征", "囚繫", "求知慾", "虯鬚", "球后", "球檯", "趨吉避凶", "麴車", "麴塵", "曲道", "麴道士", "麴酒", "曲捲", "麴菌", "麴黴", "麴櫱", "麴錢", "麴生", "曲松", "麴秀才", "麴院", "取捨", "圈釦", "圈樑", "權力慾", "權慾薰心", "全綵", "全斗煥", "全乾", "全球定位系統衛星測量", "詮註", "痊癒", "鬈髮", "犬隻", "卻纔", "确瘠", "確係", "闕里", "裙襬", "羣后", "羣謀咸同", "羣辟", "然身死纔數月耳", "髯鬍", "染髮", "染干", "繞樑", "人蔘", "人物誌", "人慾", "人云", "人云亦云", "仁杰", "忍饑受餓", "忍饑受渴", "認製修", "日本國誌", "日本製", "日復一日", "日干", "日進斗金", "日曆", "日誌", "日製", "榮登后座", "榮獲冠軍", "溶岩", "熔燬", "熔鍊", "熔岩", "鎔岩", "揉麪", "鞣製", "肉乾", "肉羹麪", "肉絲麪", "肉慾", "如干", "如坐鍼氈", "茹誌鵑", "儒略改革曆", "儒略曆", "乳孃", "辱游", "入托", "阮咸", "软肥皂", "瑞郎方麪", "瑞簽", "瑞士捲", "潤髮", "若干", "弱智賴于涵", "撒佈", "薩巴托", "薩里", "腮斗", "腮托", "賽里木湖", "三複", "三里河", "三里屯", "三孃教子", "三辟", "三尸", "三統曆", "三瓦四舍", "三絃", "三熏三沐", "三葉松", "三浴三熏", "三元里", "三紮", "三徵七辟", "三隻", "三週", "散佈", "桑乾", "桑托里尼島", "桑托榮", "桑托斯", "喪蕩游魂", "喪鐘", "色慾", "銫鐘", "沙蔘", "沙坑桿", "沙里夫", "沙崙", "沙岩", "砂鍋麪", "砂岩", "傻里傻氣", "曬乾", "曬穀", "曬菸", "山崩鐘應", "山斗", "山裏", "山里站", "山樑", "山岩", "山羊鬍", "山羊鬚", "山仔后", "山中無曆日", "山重水複", "膻中", "鱔魚麪", "商曆", "賞讚", "上衝", "上沖下洗", "上覆", "上閤屋", "上課鐘", "上栗縣", "上鍊", "上樑", "上樑不正下樑歪", "上野樹里", "上游", "上週", "燒乾", "燒燬", "燒製", "筲斗", "韶山沖", "邵廷采", "舌乾脣焦", "舌乾脣焦", "舌一捲", "蛇髮女妖", "蛇綠混雜岩", "蛇綠岩", "蛇皮松", "蛇紋岩", "捨安就危", "捨本", "捨不得", "捨車保帥", "捨出", "捨得", "捨短從長", "捨短錄長", "捨短取長", "捨短用長", "捨墮", "捨己", "捨近即遠", "捨近謀遠", "捨近求遠", "捨近務遠", "捨舊迎新", "捨命", "捨棄", "捨去", "捨身", "捨生", "捨實", "捨死忘生", "捨我復誰", "捨我其誰", "舍下", "捨下你", "捨下他", "捨下她", "捨下我", "捨正從邪", "設言托意", "射鵰", "射覆", "射干", "攝製", "申覆", "身教重於言教", "身繫囹圄", "深成岩", "深山何處鐘", "什錦麪", "神采", "神鵰", "神魂盪颺", "神蹟", "神曲", "神麴茶", "神聖週", "神搖魂盪", "瀋海", "瀋河", "沈積", "沈積岩", "瀋吉線", "瀋山線", "瀋水", "瀋陽", "瀋州", "審覈", "嬸孃", "腎臟", "升斗", "昇汞", "昇華", "昇平", "升遷管道", "昇天", "昇仙", "升學壓力", "昇陽", "生旦淨末丑", "生棟覆屋", "生髮", "生薑", "生力麪", "生態環境游", "生田斗", "生物鐘", "繩釦", "省慾去奢", "聖盃", "聖哈辛托", "聖蹟", "聖帕強颱", "聖神降臨週", "聖修伯里", "勝部冠軍", "勝蹟", "胜鍵", "胜肽", "盛价", "盛讚", "尸臣", "尸諫", "尸解", "尸鳩", "尸居龍見", "尸居餘氣", "尸利", "尸祿", "尸陀林", "尸位", "尸饔", "尸祝", "失之毫釐", "失之毫厘謬以千里", "師孃", "師生盃", "師云而云", "詩云", "詩讚", "詩鐘", "施佳昇", "施仁佈恩", "施仁佈澤", "施捨", "施粥捨飯", "溼地松", "詩云", "十分乾", "十干", "十里", "十紮", "十隻", "十週", "石柺", "石灰岩", "石几", "石枯松老", "石樑", "石松", "石屋製果", "石英錶", "石英岩", "石英鐘", "石鐘乳", "時尚週", "時憲曆", "時鐘", "時裝週", "拾瀋", "食蟲植物", "食麪", "食野之苹", "食慾", "史查克", "史蹟", "史托姆", "史托瑟", "史托蘇兒", "史托騰柏格", "史托威", "史游", "示範單位", "示範教學", "示覆", "世綵堂", "世紀鐘", "世界盃", "市長盃", "事蹟", "事可幹", "事情干脆", "事情可幹", "飾釦", "試驗檯", "試製", "拭乾", "是隻", "適當的", "視如寇讎", "嗜酸乳干菌", "嗜慾", "收穫", "手錶", "手工檯", "手鍊", "手面", "手面賺吃", "手術", "手術檯", "手痠", "手一捲", "手摺", "手製", "手冢治虫", "守禦", "首都盃", "首隻", "壽麪", "受命於天", "售後服務", "獸姦", "獸慾", "書檯", "殊域周咨錄", "梳髮", "梳妝檯", "淑郁", "舒捲", "輸征", "鼠麴草", "朮赤", "束髮", "束脩", "樹樑", "恕乏价催", "數箇", "數與虜确", "數週", "數字鐘", "數罪併罰", "甩髮", "雙鵰", "雙柺", "雙后前兵開局", "雙胜類", "雙摺", "雙週", "雙子葉植物", "水錶", "水纔乾", "水成岩", "水斗", "水乾", "水谷", "水穀之海", "水管麪", "水裏", "水里鄉", "水里鄉", "水一衝", "水已乾", "水準測量", "睡眠慾", "順德者吉逆天者凶", "順鐘向", "順硃兒", "說岳", "絲恩髮怨", "絲髮", "絲桿", "絲托索", "絲絃", "私慾", "斯伯丁盃", "斯迪里", "斯干", "斯里", "斯里查潘", "斯托", "斯瓦希里", "死麪", "死傷相藉", "四分曆", "四海皆准", "四里", "四面", "四面鐘", "四捨六入", "四捨五入", "四凶", "四紮", "四隻", "寺鐘", "飼餵", "松巴哇", "松柏", "松北", "松本", "松材線蟲", "松大輔", "松風", "松岡", "松高路", "松果", "松鶴", "松花", "松化石", "松篁交翠", "松雞", "松江", "松膠", "松焦油", "松節油", "松井秀", "松菌", "松筠之操", "松筠之節", "松科", "鬆口", "松口蘑", "鬆釦", "松類", "松遼平原", "松林", "松嶺", "松隆子", "松露", "松蘿", "松毛", "松煤", "松明", "松蘑", "松漠", "松木", "松潘", "松皮癬", "松浦", "松喬", "松青", "松丘", "松球", "松毬", "松瓤", "松仁", "松茸", "松蕊", "松山", "松獅犬", "松石", "松鼠", "松樹", "松濤", "松桃苗族自治縣", "松桃縣", "松田", "松尾芭蕉", "松紋", "松溪", "松下", "松香", "松雪泰子", "松蕈", "松鴉", "松煙", "松陽", "松葉", "松油", "松魚", "松原", "松贊干布", "松藻蟲", "松針", "松枝", "松脂", "松指部", "松智路", "松竹", "松子", "送報伕", "誦唸", "頌繫", "頌讚", "蒐藏", "搜查證", "蒐購", "蒐集", "蒐括", "蒐羅", "蒐證", "溲麪", "蘇盃", "蘇俄在中國", "蘇格蘭", "蘇格蘭摺耳貓", "蘇哈托", "甦家屯", "蘇崑", "蘇里", "甦仙區", "甦醒", "蘇揚托", "酥簽", "素髮", "素藉", "塑膠製", "溯游", "痠懶", "痠麻", "痠軟", "痠疼", "痠痛", "蒜髮", "蒜薹", "算髮", "算曆", "雖覆能復", "歲凶", "歲聿云暮", "脺臟", "碎髮", "碎屑岩", "繐裳", "繐幃飄井幹", "繐帷", "繐帳", "孫杰", "筍乾", "縮影微捲", "所佈的", "所佈之", "所繫", "所云", "索爾茲伯里平原", "索爾茲伯里石環", "索里士", "索馬里", "索馬里", "索麪", "索托", "鎖釦", "他鐘", "塔克", "塔克拉馬干", "塔克拉瑪干", "塔里班", "塔里木", "塔里契亞努", "塔什干", "塔什庫爾干塔吉克自治縣", "塔什庫爾干鄉", "塔什庫爾干自治縣", "塔鐘", "胎髮", "檯安", "臺北市長", "檯筆", "檯布", "檯秤", "檯燈", "檯凳", "颱風", "臺鑒", "檯曆", "檯面", "檯盤", "臺澎金馬", "檯球", "台山", "檯扇", "臺灣", "臺灣教育學院", "臺灣台", "臺製", "臺中", "臺中教育大學", "檯鐘", "台州", "駘藉", "太沖", "太初曆", "太乾", "太后", "太麻里", "太陽黑子週", "泰斗", "貪慾", "罈白干", "罈陳年", "罈高粱", "罈好酒", "罈騞", "罈佳釀", "罈老酒", "罈美酒", "罈女兒紅", "罈燒刀子", "罈燒刀子", "罈罈罐罐", "罈子", "潭祉叶吉", "歎服", "歎號", "歎絕", "歎賞", "嘆爲", "歎爲觀止", "歎羨", "探知慾", "碳酸岩", "歎吁", "湯麪", "湯糰", "湯下麪", "糖醋里脊", "燙次髮", "燙髮", "燙個", "燙個髮", "燙麪", "燙完髮", "燙一次髮", "燙一個髮", "陶製", "特別公積金", "特里", "特內里費", "特松加", "特製", "特准", "騰昇", "藤製", "提樑", "提製", "提子乾", "題簽", "剃髮", "剃鬍", "剃鬚", "天冬醯胺", "天翻地覆", "天覆", "天干", "天乾物燥", "天后", "天剋地衝", "天曆", "天台", "天文學鐘", "天文鐘", "天心岩", "天要落雨孃要嫁人", "天要下雨孃要嫁人", "田穀", "甜麪醬", "甜水麪", "捵麪", "挑大樑", "挑了", "挑了隻", "條几", "調錶", "調查團", "調絃", "調製", "跳錶", "跳梁", "跳樑猖獗之小醜", "跳樑小醜", "跳隻舞", "鐵板", "鐵板麪", "鐵板牛柳", "鐵板牛肉", "鐵桿", "鐵柺", "鐵釦", "鐵托", "鐵製", "鐵鐘", "聽絃", "停製", "通佈圖", "通姦", "通曆", "通心麪", "同步衛星", "同儕壓力", "同吃", "同人", "同人誌", "同休共慼", "銅斗兒", "銅釦", "銅製", "銅鐘", "瞳矇", "偷嚐禁果", "偷樑換柱", "頭髮", "頭髮鬍子一把抓", "頭花髮", "頭巾弔在水裏", "投托", "禿髮", "禿妃之髮", "圖書館週", "涂長望", "涂爾干", "涂逢年", "涂鴻欽", "涂鴻欽", "涂惠元", "涂惠源", "塗漿檯", "涂謹申", "涂謹申", "涂居賢", "涂坤", "涂美倫", "涂敏恆", "涂敏恆", "涂善妮", "涂紹煃", "涂天相", "涂文生", "涂醒哲", "涂姓", "涂序瑄", "涂永輝", "涂羽卿", "涂月", "塗澤", "涂澤民", "涂澤民", "涂長望", "涂壯勳", "涂壯勳", "土穀祠", "土托魚", "土製", "糰粉", "團伙", "團體冠軍", "糰子", "團圞", "推陳佈新", "推誠佈公", "推誠佈信", "推輦歸里", "推託", "推托之詞", "推輓", "推絃", "腿痠", "吞併", "吞嚥", "屯紮", "托杯", "托比麥奎爾", "托比亞斯", "托鉢人", "托鉢僧", "托鉢修會", "托鉢", "托出", "托膽", "托登漢隊", "托地", "托蒂", "托兒", "托爾", "托夫", "託福", "托福考", "托葷鹹食", "托架", "托克", "托克托", "托拉", "托辣斯", "托萊多", "托賴", "托老院", "托老中心", "托勒", "托裏", "托利黨人", "托利米爾", "托領", "托羅斯山", "托洛茨基", "托洛斯基", "托馬", "托米", "托墨", "托木爾", "托盤", "托起", "托瑞賽", "托瑞絲", "托腮", "托塞洛", "托色", "托實", "托熟", "托斯卡", "托塔天王", "托胎", "托特", "托維", "托葉", "托幼", "托育", "託運", "托運行李", "托住", "托子", "托足", "拖斗", "脫髮", "脫穀機", "坨里", "跎縴", "唾面", "唾面自乾", "唾沫直嚥", "瓦爾基里", "瓦里", "瓦松", "外強中乾", "外菸", "外慾", "外禦其侮", "外製", "彎管麪", "完全癒復", "頑鹵", "頑筑舞笈", "輓詞", "輓詞", "輓額", "輓夫", "輓歌", "輓聯", "輓聯", "輓曲", "輓詩", "輓詩", "晚鐘", "綰髮", "碗麪", "萬里", "萬里長征", "萬曆", "萬年", "萬年曆錶", "万旗", "万俟", "萬罈", "萬紮", "萬隻", "腕錶", "王乾發", "王侯后", "王后", "王熙松", "王于真", "王正杰", "王子麪", "網禦", "網誌", "往日無讎", "忘生捨死", "望后石", "威奇托", "偎乾", "煨乾", "韋后", "爲中颱", "維斗", "維克", "維克托", "維繫", "偉晶岩", "葦苕繫巢", "葦蓆", "尾註", "衛星鐘", "未乾", "胃臟", "餵飽", "餵哺", "餵動物", "餵鵝", "餵飯", "餵給", "餵過", "餵雞", "餵了", "餵驢", "餵馬", "餵母乳", "餵奶", "餵你", "餵乳", "餵食", "餵它", "餵我", "餵鴨", "餵羊", "餵養", "餵魚", "餵豬", "慰藉", "魏郁奇", "文采", "文采郁郁", "文丑", "文錦覆阱", "文斯范恩", "炆麪", "穩吃三注", "穩紮", "問卷", "問卷大調查", "問卷調查", "翁乾晃", "翁郁容", "窩裏", "窩里窩囊", "蝸桿", "我係", "沃依采克", "臥薪嚐膽", "握髮", "烏東查克", "烏髮", "烏干達", "烏狗吃食白狗當災", "烏蘭巴托", "烏里", "烏龍麪", "烏洛托品", "烏松", "烏托邦", "污衊", "巫咸", "誣衊", "屋樑", "無干", "無價", "無价事", "無價珍珠", "無精打采", "無厘頭", "無樑", "無樑斗", "無慾", "吳采璋", "吳皓昇", "吳里克", "吳榮杰", "吳育昇", "無言不讎", "五采", "五斗", "五穀", "五行生剋", "五棵松", "五里", "五辟", "五絃", "五臟", "五紮", "五隻", "五週", "伍采克", "武丑", "武大郎吃毒藥", "武后", "武里省", "武松", "舞后", "舞水端里", "兀朮", "勿里洞島", "物慾", "西后", "西里", "西曆", "西利古里", "西松", "西征", "西周", "西周鐘", "吸乾", "吸菸", "希伯來曆", "希拉剋", "希斯崙", "矽岩", "矽質岩", "息穀", "晞髮", "欷吁", "稀里", "稀里嘩啦", "傒倖", "徯倖", "席捲", "蓆棚", "席誌成", "襲捲", "洗盪", "洗髮", "洗髮皂", "洗臉檯", "洗手檯", "喜歡錶", "喜歡鐘", "戲綵娛親", "係臂", "繫臂之寵", "繫泊", "繫船樁", "繫辭", "繫帶", "繫到", "繫而不食", "繫髮帶", "繫風捕景", "繫風捕影", "繫縛", "繫個", "繫裹", "繫好", "繫懷", "係獲", "繫結", "繫緊", "繫頸", "係頸闕庭", "係扣", "繫褲子", "繫纜", "繫牢", "繫了", "繫累", "繫戀", "繫鈴解鈴", "繫鈴人", "繫留", "繫馬", "繫命", "繫捻兒", "繫念", "繫囚", "繫上", "繫繩", "繫世", "係數", "繫絲帶", "係蹄", "繫條", "繫頭巾", "係爲", "係爲", "繫系", "繫鞋帶", "繫心", "繫腰", "係一番", "係一片", "繫一線", "係一種", "繫有", "繫於", "繫於一髮", "繫獄", "繫爪", "繫着", "係爭", "係指", "繫趾", "係踵", "繫住", "細不容髮", "細鍊", "細如髮", "細嚥", "細緻", "舄鹵", "潟鹵", "蝦乾", "蝦鬚", "俠氣干雲", "狎妓冶游", "下襬", "下采", "下課鐘", "下里", "下樑", "下崙路", "下嚥", "下游", "下週", "夏后氏", "夏里夫", "夏曆", "夏于喬", "夏于喬", "仙后", "仙蹟", "仙台", "仙岩", "先嚐", "先乾爲敬", "先簽", "先天下之憂而憂后天下之樂而樂", "先小人後君子", "縴夫", "縴戶", "纖維植物", "鮮穀王", "鮮于", "賢后", "絃動", "絃斷", "絃歌", "絃樂", "絃器", "絃琴", "絃聲", "絃索", "絃線", "絃音", "絃軸", "咸安區", "咸池", "咸豐", "咸和", "咸亨", "咸鏡", "鹹鹵", "咸寧", "咸認爲", "咸五登三", "咸信", "咸興", "咸陽", "咸宜", "賢后", "顯示錶", "顯示鐘", "縣誌", "羨歎", "鄉里", "鄉愿", "相併", "相沖", "相干", "相姦", "相剋", "相里", "相托", "香斗", "香乾", "香榭里大道", "香菸", "香郁", "香皂", "箱釦", "詳註", "響絃", "響鐘", "嚮導", "嚮邇", "嚮晦", "嚮明", "嚮慕", "嚮往", "嚮應", "曏者", "項鍊", "橡斗", "橡子麪", "消費慾", "宵征", "蕭蔘", "蕭行范篆", "銷燬", "小便斗", "小嚐", "小丑", "小醜跳樑", "小范", "小几", "小价", "小杰", "小栗旬", "小千世界", "小松", "小型鐘", "小余", "小云", "小澤征爾", "小隻", "小鐘", "校覈", "歇斯底里", "邪不干正", "邪辟", "斜管麪", "鞋釦", "寫字檯", "泄慾", "械繫", "謝福松", "謝里", "蟹黃鮑魚麪", "心蕩神搖", "心高遮了太陽", "心花怒發", "心繫", "心細似髮", "心絃", "心相繫", "心臟", "心臟復甦術", "辛丑", "辛辣麪", "辛里希", "新干", "新喀里多尼亞", "新曆", "新紮", "信馬游繮", "信天游", "信託", "信托貿易", "釁鐘", "星辰錶", "星斗", "星迴", "星曆", "星曆錶", "星羅雲佈", "星移斗換", "星移斗轉", "星占學", "刑剋", "刑辟", "刑于", "形單影隻", "形孤影隻", "形影相弔", "興雲佈雨", "倖臣", "倖存", "倖感歌姬", "倖進", "倖免", "倖幸", "幸運", "幸運鬍", "性潑凶頑", "性慾", "姓岳", "凶德", "凶地", "凶多吉少", "凶服", "凶怪", "凶耗", "凶荒", "凶禮", "凶門", "凶逆", "凶年", "凶年饑歲", "凶氣", "凶日", "凶煞", "凶身", "凶神", "凶事", "凶豎", "凶死", "凶肆", "凶歲", "凶信", "凶訊", "凶燄", "凶宅", "凶兆", "凶終隙末", "雄斗斗", "休崙湖", "休慼", "修鬍刀", "修杰楷", "脩金", "脩敬", "脩名", "脩脯", "脩潤", "脩樾", "宿松", "秀髮", "袖釦", "袖一捲", "吁咈", "吁嗟", "吁了", "吁氣", "吁嘆", "吁吁", "吁噓", "吁俞", "鬚髮", "鬚根", "鬚後水", "鬚鬍", "鬚鯨", "鬚毛", "鬚眉", "須彌山", "鬚髯", "鬚鯊", "鬚生", "鬚鬚", "鬚子", "虛沖", "墟里", "徐匯", "徐彙區", "徐家彙", "徐余偉", "徐讚昇", "許人丰", "許聖杰", "許願", "許愿起經", "旭日初昇", "卹典", "卹荒", "卹金", "續簽", "續絃", "蓄髮", "蓄鬍", "蓄鬚", "宣佈", "宣傳週", "喧鬨", "玄蔘", "玄武岩", "玄鍼", "玄製", "懸臂樑", "懸樑", "懸鐘", "懸冑", "旋乾轉坤", "旋迴", "旋里", "旋辟", "削髮", "削麪", "薛松乾", "踅門瞭戶", "雪窗螢几", "雪松", "血纔乾", "血蔘", "血已乾", "血癥", "燻草", "薰風", "熏風徐來", "熏腐", "燻赫", "燻黑", "燻雞", "燻烤", "熏籠", "燻鑪", "燻人", "燻肉", "薰陶", "熏陶成性", "熏天", "熏習", "熏熏", "熏衣草", "燻魚兒", "熏烝", "燻蒸", "熏蒸劑", "熏蒸室", "熏製", "薰脩", "巡迴", "巡迴檢查", "壓桿", "壓力錶", "壓麪棍", "壓冑子", "鴉窩裏出鳳凰", "崖广", "啞子托夢", "雅游", "雅緻", "雅筑", "亞里", "亞美尼亞曆", "亞青盃", "亞松森", "亞洲盃", "嚥不了", "嚥到", "咽乾", "嚥肌", "嚥進", "嚥苦吞甘", "嚥了", "嚥氣", "嚥唾", "嚥下", "嚥着", "嚥住", "菸草", "菸廠", "菸袋", "菸袋桿兒", "菸蒂", "菸斗", "煙桿", "菸缸", "煙管", "煙管麪", "菸害", "菸灰", "菸鹼", "菸禁", "菸酒", "菸捲", "菸民", "菸農", "菸屁股", "菸圈", "菸絲", "菸頭", "煙燻", "煙熏火燎", "菸蚜", "菸葉", "菸紙店", "菸嘴", "腌裏巴臢", "腌臢", "腌䐶", "醃製", "嚴云農", "言辯而确", "言大而夸", "言云", "岩倉使節團", "岩層", "岩牀", "巖村", "岩村明憲", "岩基", "岩漿", "岩漿岩", "岩礁", "岩瀨健", "岩脈", "岩棉", "岩圈", "岩溶", "岩石", "岩土", "岩屑", "岩心", "岩鹽", "岩羊", "沿門托鉢", "研製", "筵几", "眼乾", "眼花瞭亂", "眼前花發", "眼痠", "偃仆", "偃松", "豔后", "唁弔", "驗覈", "饜於游樂", "燕几", "揚穀", "羊鬚瘡", "陽春麪", "陽曆", "楊采妮", "楊淩示範區", "楊日松", "楊士樑", "楊松", "楊甦棣", "楊文誌", "洋蔘", "洋麪", "洋菸", "養髮", "妖后", "妖里妖氣", "腰桿", "腰間繫", "腰釦", "腰痠", "腰繫", "腰一捲", "僥天之倖", "僥倖", "姚采穎", "姚昇志", "搖盪", "搖桿", "咬薑呷醋", "藥而癒", "藥麪兒", "藥用植物", "藥皂", "要乾了", "耶孃", "椰棗乾", "噎饑", "爺飯孃羹", "爺羹孃飯", "爺孃", "也斗了膽", "野薑", "葉步樑", "叶恭弘", "葉輪機械", "葉葉", "葉叶琹", "叶音", "叶韻", "葉狀植物", "葉子", "葉子杰", "葉子菸", "頁岩", "夜光錶", "液晶錶", "液麪", "一表人物", "一別", "一彆頭", "一併", "一沖性子", "一齣劇", "一齣子", "一斗", "一斗斗", "一發", "一髮千鈞", "一髮之差", "一髮之間", "一杆", "一桿進洞", "一干", "一乾而盡", "一乾二淨", "一箇", "一鍋麪", "一毫一髮", "一號木桿", "一鬨", "一塊麪", "一釐一毫", "一里", "一簽", "一日叫孃", "一樹百穫", "一罈", "一罈罈", "一天鐘", "一托氣", "一托頭", "一物剋一物", "一讚", "一紮", "一隻", "一週", "伊府麪", "伊拉克", "伊里奇", "伊麪", "伊斯蘭曆", "伊于胡底", "伊于湖底", "衣襬", "衣不兼采", "衣不完采", "衣不重采", "衣斗木", "衣錦夜游", "衣錦晝游", "衣釦", "衣衫已乾", "衣物漸乾", "衣物已乾", "醫托", "噫吁戲", "宜云", "胰臟", "移星換斗", "遺傳鐘", "遺葑菲采", "遺蹟", "疑係", "乙丑", "已係", "已佔", "已占算", "以功覆過", "蟻后", "倚托", "倚閑", "蟻后", "億隻", "義氣干霄", "仡栗", "亦捨下", "亦云", "異采", "異薹同岑", "抑揚昇降性", "邑里", "譯製", "譯註", "逸游自恣", "逸緻", "意大利直麪", "意麪", "因姦成孕", "陰乾", "陰曆", "陰占", "茵藉", "廕庇", "廕監", "廕生", "廕襲", "音聲如鐘", "吟歎", "銀盃", "銀髮", "銀絲捲", "銀鬚", "銀製", "銀硃", "淫慾", "嚚闇", "隱几", "飮冑", "印製", "英里", "櫻花盃", "鷹鵰", "縈迴", "縈繫", "影后", "影評人週", "影占", "影隻形單", "應有盡有", "應占", "應鐘", "硬麪", "硬嚥", "佣鈿", "佣金", "佣錢", "庸闇", "永曆", "永志", "永誌不忘", "詠歎", "泳氣鐘", "攸慼相關", "悠闇", "悠盪", "悠活麗緻", "悠悠盪盪", "悠遊", "悠遊錶", "尤里", "由余", "猶如錶", "猶如鐘", "猶太曆", "油麪", "油松", "莜麪", "游塵", "游出", "游到", "游芳來", "遊蜂浪蝶", "遊過", "游過來", "游過去", "游皓瑋", "游鴻明", "游鴻儒", "游回", "遊擊", "游擊隊", "游擊區", "游擊手", "游擊戰", "遊進", "游進來", "游進去", "游來", "游來游去", "遊樂器", "遊離電子", "游龍", "游履", "遊美洲", "游明金", "遊牧民族", "游乃海", "遊歐洲", "游泮", "游禽類", "游去", "游僧攢住持", "游上", "游水", "游完", "游文宏", "遊錫", "游錫坤", "游錫昆", "游錫堃", "遊戲", "遊戲機", "遊戲機檯", "游下", "遊亞洲", "游盈隆", "游泳", "游囿倫", "游魚", "游昭欽", "游志宏", "遊中國", "友于", "有采", "有仇不報非君子", "有齣好戲", "有髮頭陀寺", "有夠讚", "有儘有讓", "有進有出", "有云", "有徵", "有征無戰", "有隻", "羑里", "牖里", "又幹", "又乾又硬", "又云", "幼托", "誘姦", "迂迴", "紆迴", "于貝爾", "于斌", "于波", "于晨楠", "于成龍", "于成龍", "于從濂", "于從濂", "于大寶", "于丹", "于道泉", "于德海", "于爾岑", "于爾根", "于爾里克", "于爾里克", "于飛", "于風政", "于鳳桐", "于鳳至", "于格", "于根偉", "于光遠", "于廣洲", "于歸", "於國", "于國楨", "于海洋", "于漢超", "于浩威", "于衡", "于洪區", "于化虎", "于會泳", "于慧", "于吉", "于佳卉", "於家", "于家堡", "于堅", "于堅", "于江震", "于嗟", "于傑", "于禁", "于靖", "于娟", "于軍", "于軍", "于康震", "于克勒", "于孔兼", "于勒", "于里察", "于凌奎", "于冕", "于敏", "于明濤", "于默奧", "于娜", "于品海", "于奇庫杜克", "于謙", "于謙", "于晴", "于仁泰", "于若木", "于山", "于慎行", "于式枚", "于樹潔", "于帥", "于帥", "于雙戈", "于思", "於斯", "于斯達爾", "于斯達爾", "于斯納爾斯貝里", "于斯納爾斯貝里", "于斯塔德", "于素秋", "于臺煙", "于濤", "于濤", "于特森", "于天仁", "于田", "于闐", "于韋斯屈萊", "于韋斯屈萊", "于偉國", "于西翰", "于湘蘭", "于小彤", "于小偉", "于欣源", "于姓", "于秀敏", "于徐", "于學忠", "于蔭霖", "于永波", "于右任", "于幼軍", "于于", "于餘曲折", "于餘曲折", "于玉立", "于遠偉", "于越", "于樂", "于澤爾", "于贈", "于贈", "于占元", "于振", "于震環", "于震寰", "于正昌", "于正昇", "于志寧", "于寘", "于子千", "污衊", "余碧芬", "余秉諺", "余炳賢", "余燦榮", "余車", "余發揚", "餘干", "余歌滄", "餘光", "余光生", "余光中", "余蘭香", "餘里", "余男", "余珮琳", "餘三", "余三勝", "余三勝", "余上沅", "餘思", "余思敏", "余天", "餘威", "余威德", "余文", "余吾鎮", "余賢明", "余憲宗", "余筱萍", "余姓", "余秀菁", "余雪蘭", "余雪明", "余英時", "余余", "余苑綺", "余月", "余政憲", "餘隻", "餘子", "余子明", "魚乾", "魚鱗松", "魚游釜底", "魚游釜中", "魚冑", "逾閑蕩檢", "愚闇", "輿尸", "與國同休", "宇宙誌", "語彙", "語云", "玉斗", "玉里", "玉曆", "玉米鬚", "玉製", "郁達夫", "郁馥", "郁離子", "郁李", "郁烈", "郁穆", "郁樸", "鬱郁", "郁郁菲菲", "郁郁青青", "郁哉", "預製", "慾海", "慾壑難填", "慾火", "欲加之罪", "慾加之罪何患無詞", "慾令智昏", "慾念", "慾女", "欲求", "慾求不滿", "慾善其事必先利其器", "慾望", "慾障", "禦敵", "禦寒", "禦寇", "御史大夫", "禦侮", "御製", "寓禁於征", "癒合", "元后", "袁友范", "袁于令", "原鐘", "原子鐘", "遠縣纔至", "遠征", "愿而恭", "愿樸", "曰云", "約翰松", "月曆", "月相錶", "月岩", "岳飛", "岳墳", "岳父", "岳家", "岳珂", "岳廟", "岳母", "岳氏", "岳陽", "岳雲", "岳丈", "熅斗", "云城區", "云爾", "雲翻雨覆", "云何", "云乎", "云然", "雲吞", "雲吞麪", "云爲", "云爲", "云溪", "雲鬚", "云云", "蕓輝", "蕓薹", "蕓薹", "耘盪", "允准", "隕穫", "醞藉", "蘊藉", "熨斗", "雜合麪兒", "雜醬麪", "雜麪", "雜誌", "再製", "在覈", "讚唄", "讚不絕口", "讚詞", "讚辭", "讚道", "讚的", "讚歌", "讚個不", "讚口不", "讚樂", "讚了", "讚兩句", "讚美", "讚佩", "讚賞", "讚頌", "讚歎", "讚歎", "讚我", "讚羨", "讚許", "讚揚", "讚一句", "讚一聲", "讚一讚", "讚語", "讚譽", "讚自己", "髒髮", "臟腑", "臟器", "臧穀亡羊", "鑿巖", "鑿岩機", "早起的鳥兒有蟲吃", "早占勿藥", "皂化", "皂莢", "造麴", "造岩礦物", "造鐘", "譟動", "譟詐", "澤里可", "澤鹵", "澤滲灕而下降", "紮成", "紮帶", "紮囮", "扎爾達里", "紮根", "紮裹", "紮好", "紮腳", "紮結", "紮緊", "紮了", "紮馬剌丁", "紮馬魯丁", "紮歐紮翁", "紮起", "紮上", "紮實", "紮鐵", "紮下", "紮線帶", "紮營", "紮在", "扎扎", "紮紮實實", "紮詐", "紮寨", "札夸威", "軋製", "炸燬", "炸醬麪", "榨乾", "詹誌維", "薝蔔", "展采", "占拜", "占卜", "占城", "占斷", "占房", "占風使帆", "占鳳", "占卦", "占候", "占課", "佔了", "占了卜", "占夢", "占強", "占親", "占人", "占上", "占射", "占身", "佔世界", "占筮", "佔万", "占星", "占驗", "佔有", "占有五不驗", "佔有慾", "佔主", "占主導地位", "戰略伙伴", "站乾岸兒", "張斗輝", "張基郁", "張木松", "張琴松", "張三丰", "張文松", "張志", "張誌家", "長髮", "長干曲", "長干巷", "長鬍", "長几", "長曆", "長繩繫景", "長繩繫日", "長壽菸", "長吁", "長鬚", "長征", "丈母孃", "仗托", "趙坤郁", "照簽", "遮覆", "摺尺", "摺疊", "摺好", "摺合", "摺痕", "折節讀書", "折進", "摺進來", "摺進去", "摺篷", "摺裙", "摺扇", "折檯", "摺梯", "摺頁", "摺椅", "摺紙", "摺子", "摺奏", "這齣電影", "這齣好戲", "這齣劇", "這隻", "這鐘", "鍼砭", "鍼關", "鍼芥相投", "鍼灸", "鍼口", "針釦", "鍼芒", "針葉植物", "珍珠岩", "甄后", "枕藉", "枕蓆", "縝緻", "鬒髮", "振盪", "振杰", "賑饑", "震盪", "鎮榮里", "黮闇", "征塵", "征程", "征伐", "征帆", "征夫", "征服", "征駕", "征剿", "征斂", "征馬", "征旆", "徵辟", "征衫", "征戍", "征討", "征途", "征衣", "征戰", "征彸", "蒸乾", "蒸麪", "整齣劇", "整發", "整髮用品", "整隻", "整週", "正樑", "正凶", "鄭家鐘", "鄭凱云", "鄭榮松", "鄭易里", "鄭余豪", "癥結", "鄭凱云", "之慾", "之鐘", "支桿", "支努干", "支菸", "只查", "只吃", "隻雞絮酒", "隻立", "隻輪不反", "隻輪不返", "隻日", "隻身", "隻聲不出", "隻手", "隻言片語", "隻言片字", "隻眼", "隻影", "只佔", "只占吉", "只占神問卜", "只占算", "只准", "隻字", "枝不得大於榦", "織蓆", "直襬", "直髮", "直系", "直係軍閥", "植髮", "植物誌", "殖穀", "止謗莫如自脩", "紙菸", "紙紮", "紙製", "指揮台", "指親托故", "指水盟松", "誌哀", "志誠君子", "志沖斗牛", "誌悼", "志高氣揚", "誌慶", "誌喜", "誌異", "誌之不忘", "製版", "製備", "製表", "製冰", "製播", "製材", "製茶", "製成", "製程", "製出", "製得", "製毒", "製法", "製革", "製劑", "製假", "製件", "製漿", "製冷", "制面", "製面具", "製坯", "製片", "製品", "製取", "製售", "製酸性", "製糖", "製陶", "製圖", "製爲", "製爲", "製鞋", "製鹽", "製氧", "製藥", "製衣", "製造", "製紙", "制鐘", "製作", "製做", "治癒", "櫛髮工", "緻密", "窒慾", "躓仆", "薙髮", "中國製", "中環盃", "中崙", "中千世界", "中型鐘", "中游", "忠人之托", "終端檯", "終身有托", "鐘擺", "鐘被", "鐘壁", "鐘錶", "鐘表盤", "鐘不", "鐘差", "鐘楚紅", "鐘的", "鐘點", "鐘頂", "鐘鼎", "鐘發音", "鐘福松", "鐘鼓", "鐘關", "鐘行", "鐘好", "鐘匠", "鐘口", "鐘快", "鐘樂", "鐘樓", "鐘漏", "鐘螺", "鐘律", "鐘慢", "鐘沒", "鐘面", "鐘鳴", "鐘模", "鐘紐", "鐘盤", "鐘珮瑄", "鐘敲", "鐘琴", "鐘磬", "鐘乳洞", "鐘乳石", "鐘山", "鐘上", "鐘身", "鐘聲", "鐘速", "鐘塔", "鐘太", "鐘體", "鐘調", "鐘停", "鐘頭", "鐘王", "鐘下", "鐘相", "鐘響", "鐘形", "鐘腰", "鐘意", "鐘有", "鐘在寺裏", "鐘罩", "鐘左右", "鐘座", "种放", "種穀", "种師道", "种師中", "種子植物", "重複", "重見複出", "重羅麪", "重摺", "重製", "州里", "週報", "週而復始", "週二", "週迴", "週會", "週記", "賙濟", "周杰", "週刊", "週考", "周曆", "週六", "週末", "週年", "週期", "周全方便", "周人", "賙人之急", "週日", "週三", "週上", "週數", "週四", "週歲", "週五", "週薪", "週休", "週一", "周遊世界", "周召共和", "週中", "週週", "週轉", "週走秀", "洲際盃", "咒愿", "冑甲", "冑科", "晝伏夜游", "皺彆", "皺摺", "朱貝爾", "硃筆", "朱脣", "硃脣皓齒", "朱脣皓齒", "朱轓皂蓋", "朱干玉鏚", "硃紅", "硃卷", "朱儁", "朱口皓齒", "朱理安曆", "朱崙街", "硃批", "硃色", "硃砂", "朱天文", "朱郁信", "硃諭", "珠斗爛班", "豬八戒吃人參果", "豬肝麪", "豬腳麪", "豬舌麪", "豬隻", "竹苞松茂", "竹几", "竹蓆", "竹製", "主樑", "主鐘曲線", "煮麪", "煮粥焚鬚", "芧栗", "住紮", "註標", "註冊", "註冊", "註定", "註記", "註腳", "註解", "註名", "註明", "註批", "注入", "注入式教學法", "註上", "註生娘娘", "註失", "註釋", "註疏", "註文", "註銷", "註譯", "注云", "駐紮", "柱樑", "祝髮", "祝讚", "鑄鐘", "筑邦", "筑北", "筑波", "筑肥", "筑後", "筑後", "筑前", "筑西", "筑陽", "筑陽", "筑州", "筑紫", "抓姦", "專鑒", "專征", "轉速錶", "轉檯", "轉游", "轉註", "裝岩機", "裝摺", "壯麪", "撞府沖州", "撞球桿", "撞鐘", "準保", "准保護", "准保釋", "准不准你", "准不准誰", "准不准他", "准不准它", "准不准她", "准不准我", "准不准許", "准此", "準分子", "准伏", "准假", "准將", "准決鬥", "准考證", "准普爾", "准入", "准三后", "准算", "准尉", "准許", "准以", "准予", "准折", "准奏", "捉髮", "捉姦", "桌几", "桌曆", "濁積岩", "擢髮", "姿采", "髭鬍", "髭鬚", "子薑炒雞", "子游", "子云", "子之丰兮", "梓里", "紫薑", "紫雲", "紫云苗族布依族自治縣", "自動錶", "自然捲", "自制", "自製炸彈", "字彙", "漬已乾", "宗周", "宗周鐘", "綜覈", "總髮", "總桿賽", "總桿數", "總彙", "總檯", "總統盃", "縱橫交佈", "縱慾", "粽粑葉", "奏摺", "足協盃", "足總盃", "組合服裝", "祖沖之", "鑽桿", "最後一天", "左光斗", "左里克", "左鄰右里", "左手不托右手", "左右采之", "作倖", "座標", "坐如鐘", "坐檯", "坐鐘", "座鐘", "做出", "做齣好戲"];
    var _t2hk_t = ["僞", "兌", "冑", "冗", "勳", "叄", "告", "啓", "喫", "嘆", "囪", "妝", "媼", "嫋", "嫺", "嬀", "岩", "悅", "慍", "戶", "挩", "搵", "擡", "敓", "敘", "柺", "梲", "棱", "榲", "檯", "氳", "涌", "涗", "溫", "溼", "潙", "潨", "熅", "爲", "痹", "癡", "皁", "稅", "竈", "糉", "縕", "繮", "纔", "脫", "膃", "臥", "臺", "菸", "蒕", "蔥", "蔿", "蘊", "蛻", "衆", "衛", "覈", "說", "贗", "踊", "轀", "醞", "鉢", "鉤", "銳", "閱", "鰮", "鱉", "貝胄", "大欖涌", "大涌", "東涌", "冠胄", "韓侂胄", "蠔涌", "河涌", "甲胄", "介胄", "鎧胄", "葵涌", "黎涌", "麻涌", "免胄", "南涌", "泥涌", "沙魚涌", "深涌", "西涌", "溪涌", "懸胄", "壓胄子", "飮胄", "涌尾", "魚胄", "鰂魚涌", "胄甲", "胄科"];
    var _t2hk_hk = ["偽", "兑", "胄", "宂", "勛", "叁", "吿", "啟", "吃", "歎", "囱", "粧", "媪", "裊", "嫻", "媯", "巖", "悦", "愠", "户", "捝", "揾", "抬", "敚", "敍", "枴", "棁", "稜", "榅", "枱", "氲", "湧", "涚", "温", "濕", "溈", "潀", "煴", "為", "痺", "痴", "皂", "税", "灶", "粽", "緼", "韁", "才", "脱", "腽", "卧", "台", "煙", "蒀", "葱", "蒍", "藴", "蜕", "眾", "衞", "核", "説", "贋", "踴", "輼", "醖", "缽", "鈎", "鋭", "閲", "鰛", "鼈", "貝冑", "大欖涌", "大涌", "東涌", "冠冑", "韓侂冑", "蠔涌", "河涌", "甲冑", "介冑", "鎧冑", "葵涌", "黎涌", "麻涌", "免冑", "南涌", "泥涌", "沙魚涌", "深涌", "西涌", "溪涌", "懸冑", "壓冑子", "飮冑", "涌尾", "魚冑", "鰂魚涌", "冑甲", "冑科"];
    var _t2tw_t = ["僞", "兇", "啓", "喫", "嫺", "嬀", "峯", "幺", "擡", "曬", "棱", "檐", "污", "泄", "涌", "潙", "潨", "爲", "牀", "痹", "癡", "着", "睾", "竈", "糉", "繮", "纔", "羣", "蔿", "衆", "裏", "覈", "踊", "鉢", "鮎", "麪", "齶", "砹", "硅", "酰", "鈁", "鈈", "錇", "鍀", "鎄", "鎇", "鎿", "鐦", "鑥", "PN結", "SQL注入", "SQL注入攻擊", "阿拉伯聯合酋長國", "阿塞拜疆", "埃塞俄比亞", "安提瓜和巴布達", "巴巴多斯", "巴布亞新幾內亞", "半角", "綁定", "保存", "貝寧", "奔馳", "本地代碼", "蹦極", "比特", "筆記本電腦", "壁紙", "編程", "編程語言", "便攜式", "變量", "標識符", "表達式", "並行計算", "波分複用", "波斯尼亞黑塞哥維那", "伯利茲", "博茨瓦納", "博客", "布爾", "布基納法索", "布隆迪", "採樣", "菜單", "參數", "參數表", "操作數", "操作系統", "插件", "查看", "查找", "場效應管", "城域網", "程序", "程序員", "持久性", "出租車", "觸摸", "觸摸屏", "傳感", "串行", "串行端口", "串口", "窗口", "創建", "詞組", "磁道", "磁盤", "存儲", "存盤", "打開", "打印", "打印機", "代碼", "代碼頁", "帶寬", "單片機", "刀片服務器", "導出", "導入", "登錄", "低級", "地址", "地址欄", "遞歸", "點擊", "調度", "調色板", "調試", "調試器", "調用", "調制", "調製解調器", "端口", "短信", "堆棧", "隊列", "對話框", "對象", "多米尼加", "多任務", "多態", "厄瓜多爾", "厄立特里亞", "二極管", "發佈", "發送", "範式", "方便麪", "仿真", "飛行模式", "分辨率", "分佈式", "分區", "佛得角", "服務器", "輔音", "傅里葉", "複選按鈕", "複選框", "複印", "岡比亞", "高端", "高級", "高速緩存", "高性能計算", "哥斯達黎加", "格林納達", "格魯吉亞", "構造函數", "固件", "掛斷", "關係數據庫", "光標", "光盤", "光驅", "圭亞那", "過程式編程", "哈薩克斯坦", "哈希", "函數", "函數式編程", "航天飛機", "宏內核", "洪都拉斯", "紅心大戰", "後綴", "呼出", "呼叫轉移", "互聯網", "緩存", "回調", "彙編", "彙編語言", "基里巴斯", "激光", "激活", "吉布堤", "集成", "集成電路", "集羣", "幾內亞比紹", "計算機安全", "計算機科學", "寄存器", "加納", "加蓬", "加載", "兼容", "剪切", "剪貼板", "交互", "腳本", "接口", "截屏", "截取", "解釋器", "界面", "金屬氧化物半導體", "津巴布韋", "進程", "進制", "晶體管", "晶閘管", "鏡像", "局部", "局域網", "句柄", "捲積", "咖喱", "卡塔爾", "科摩羅", "科特迪瓦", "可視化", "克羅地亞", "刻錄", "肯尼亞", "空分多址", "空分複用", "控件", "寬帶", "擴展", "擴展名", "萊索托", "藍牙", "老撾", "類模板", "類型", "利比里亞", "連接", "連接器", "聯繫", "聯繫歷史", "鏈表", "鏈接", "涼菜", "列支敦士登", "盧旺達", "錄像", "邏輯門", "馬爾代夫", "馬里共和國", "碼分多址", "碼率", "毛里求斯", "毛里塔尼亞", "冒泡排序", "枚舉", "門電路", "門戶網站", "密鑰", "免提", "面向對象", "面向過程", "命令行", "命令式編程", "命名空間", "模塊", "模擬", "模擬電路", "模擬電子", "莫桑比克", "默認", "目標代碼", "納米", "奶酪", "瑙魯", "內存", "內核", "內聯函數", "內置", "尼日爾", "尼日利亞", "歐拉", "帕勞", "盤符", "盤片", "拋出", "皮膚", "頻分多址", "頻分複用", "屏蔽", "屏幕", "前綴", "嵌套", "全角", "全局", "權限", "缺省", "任務管理器", "任務欄", "軟件", "軟驅", "塞拉利昂", "塞浦路斯", "塞舌爾", "三極管", "散列", "掃描儀", "沙特阿拉伯", "殺毒", "閃存", "社區", "設備", "設置", "審覈", "聲卡", "聲明", "聖基茨和尼維斯", "聖盧西亞", "聖馬力諾", "聖文森特和格林納丁斯", "時分多址", "時分複用", "時鐘頻率", "實例", "實模式", "矢量", "視頻", "視圖", "適配器", "首席技術官", "首席信息官", "首席運營官", "首席執行官", "鼠標", "數據", "數據報", "數據倉庫", "數據挖掘", "數據源", "數字", "數字電路", "數字電子", "數字印刷", "數組", "刷新", "斯洛文尼亞", "斯威士蘭", "死機", "搜索", "蘇里南", "算法", "算子", "縮進", "縮略圖", "所羅門羣島", "索馬里", "塔吉克斯坦", "臺式機", "坦桑尼亞", "特立尼達和多巴哥", "添加", "通配符", "通信", "通訊卡", "頭文件", "突尼斯", "圖標", "圖庫", "圖瓦盧", "圖像", "土庫曼斯坦", "脫機", "瓦努阿圖", "外鍵", "外置", "萬維網", "萬象", "網吧", "網關", "網卡", "網絡", "網上鄰居", "危地馬拉", "僞代碼", "位圖", "溫納圖萬", "文本", "文檔", "文件", "文件夾", "文件擴展名", "文件名", "文萊", "文字處理", "烏茲別克斯坦", "無損壓縮", "物理地址", "物理內存", "析構函數", "下拉列表", "仙童半導體", "顯存", "顯卡", "顯像管", "線程", "相冊", "香農", "項目", "像素", "消息", "寫保護", "卸載", "芯片", "新西蘭", "信道", "信號", "信息", "信息安全", "信息技術", "信息論", "信噪比", "性價比", "性能", "虛函數", "虛擬機", "循環", "尋址", "演示文稿", "也門", "頁腳", "頁眉", "移動電話", "移動通信", "移動網絡", "移動硬盤", "移動資料", "以太網", "異步", "意大利", "溢出", "音頻", "引導程序", "映射", "硬件", "硬盤", "用戶", "用戶名", "優先級", "有損壓縮", "預處理器", "元編程", "元數據", "元音", "原代碼", "源代碼", "源碼", "源文件", "遠程", "雲存儲", "雲計算", "運行", "運算符", "在線", "贊比亞", "乍得", "粘貼", "正則表達式", "支持", "知識產權", "指針", "智能", "中間件", "重命名", "重載", "重裝", "主板", "主引導記錄", "註冊表", "註冊機", "註銷", "狀態欄", "桌面型", "自動轉屏", "自行車", "字段", "字符", "字符串", "字節", "字庫", "字體", "總線", "組件", "最終用戶"];
    var _t2tw_tw = ["偽", "凶", "啟", "吃", "嫻", "媯", "峰", "么", "抬", "晒", "稜", "簷", "汙", "洩", "湧", "溈", "潀", "為", "床", "痺", "痴", "著", "睪", "灶", "粽", "韁", "才", "群", "蒍", "眾", "裡", "核", "踴", "缽", "鯰", "麵", "顎", "砈", "矽", "醯", "鍅", "鈽", "鉳", "鎝", "鑀", "鋂", "錼", "鉲", "鎦", "PN接面", "SQL隱碼攻擊", "SQL隱碼攻擊", "阿拉伯聯合大公國", "亞塞拜然", "衣索比亞", "安地卡及巴布達", "巴貝多", "巴布亞紐幾內亞", "半形", "繫結", "儲存", "貝南", "賓士", "原生代碼", "高空彈跳", "位元", "膝上型電腦", "桌布", "程式設計", "程式語言", "行動式", "變數", "識別符號", "表示式", "平行計算", "波長分波多工", "波士尼亞赫塞哥維納", "貝里斯", "波札那", "部落格", "布林", "布吉納法索", "蒲隆地", "取樣", "選單", "引數", "參數列", "運算元", "作業系統", "外掛", "檢視", "查詢", "場效電晶體", "都會網路", "程式", "程式設計師", "永續性", "計程車", "觸控", "觸控式螢幕", "感測", "序列", "串列埠", "串列埠", "視窗", "建立", "片語", "磁軌", "磁碟", "儲存", "存檔", "開啟", "列印", "印表機", "程式碼", "內碼表", "頻寬", "微控制器", "刀鋒伺服器", "匯出", "匯入", "登入", "低階", "地址", "位址列", "遞迴", "點選", "排程", "調色盤", "除錯", "偵錯程式", "呼叫", "調變", "數據機", "埠", "簡訊", "堆疊", "佇列", "對話方塊", "物件", "多明尼加", "多工", "多型", "厄瓜多", "厄利垂亞", "二極體", "釋出", "傳送", "正規化", "泡麵", "模擬", "飛航模式", "解析度", "分散式", "分割槽", "維德角", "伺服器", "子音", "傅立葉", "覈取按鈕", "覈取方塊", "影印", "甘比亞", "高階", "高階", "快取記憶體", "高效能運算", "哥斯大黎加", "格瑞那達", "喬治亞", "建構函式", "韌體", "結束通話", "關聯式資料庫", "游標", "光碟", "光碟機", "蓋亞那", "程序式程式設計", "哈薩克", "雜湊", "函式", "函數語言程式設計", "太空梭", "單核心", "宏都拉斯", "傷心小棧", "字尾", "撥出", "來電轉駁", "網際網路", "快取", "回撥", "彙編", "組合語言", "吉里巴斯", "鐳射", "啟用", "吉布地", "整合", "積體電路", "叢集", "幾內亞比索", "電腦保安", "電腦科學", "暫存器", "迦納", "加彭", "載入", "相容", "剪下", "剪貼簿", "互動", "指令碼", "介面", "截圖", "擷取", "直譯器", "介面", "金氧半導體", "辛巴威", "程序", "進位制", "電晶體", "閘流體", "映象", "區域性", "區域網", "控制代碼", "摺積", "咖哩", "卡達", "葛摩", "象牙海岸", "視覺化", "克羅埃西亞", "燒錄", "肯亞", "分空間多重進接", "空間多工", "控制元件", "寬頻", "擴充套件", "副檔名", "賴索托", "藍芽", "寮國", "類别範本", "型別", "賴比瑞亞", "連線", "聯結器", "聯絡", "通話記錄", "連結串列", "連結", "冷盤", "列支敦斯登", "盧安達", "錄影", "邏輯閘", "馬爾地夫", "馬利共和國", "分碼多重進接", "位元速率", "模里西斯", "茅利塔尼亞", "氣泡排序", "列舉", "閘電路", "入口網站", "金鑰", "擴音", "物件導向", "程序導向", "命令列", "指令式程式設計", "名稱空間", "模組", "模擬", "類比電路", "類比電子", "莫三比克", "預設", "目的碼", "奈米", "乳酪", "諾魯", "記憶體", "核心", "行內函數", "內建", "尼日", "奈及利亞", "尤拉", "帛琉", "碟符", "碟片", "丟擲", "面板", "分頻多重進接", "分頻多工", "遮蔽", "螢幕", "字首", "巢狀", "全形", "全域性", "許可權", "預設", "工作管理員", "工作列", "軟體", "軟碟機", "獅子山", "塞普勒斯", "塞席爾", "三極體", "雜湊", "掃描器", "沙烏地阿拉伯", "防毒", "快閃記憶體", "社羣", "裝置", "設定", "稽覈", "音效卡", "宣告", "聖克里斯多福及尼維斯", "聖露西亞", "聖馬利諾", "聖文森及格瑞那丁", "分時多重進接", "分時多工", "時脈頻率", "例項", "真實模式", "向量", "視訊", "檢視", "介面卡", "技術長", "資訊長", "營運長", "執行長", "滑鼠", "資料", "資料包", "資料倉儲", "資料探勘", "資料來源", "數字", "數位電路", "數位電子", "數位印刷", "陣列", "重新整理", "斯洛維尼亞", "史瓦濟蘭", "宕機", "搜尋", "蘇利南", "演算法", "運算元", "縮排", "縮圖", "索羅門羣島", "索馬利亞", "塔吉克", "桌上型電腦", "坦尚尼亞", "千里達及托巴哥", "新增", "萬用字元", "通訊", "通話卡", "標頭檔案", "突尼西亞", "圖示", "相簿", "吐瓦魯", "影象", "土庫曼", "離線", "萬那杜", "外來鍵", "外接", "全球資訊網", "永珍", "網咖", "閘道器", "網絡卡", "網路", "網路上的芳鄰", "瓜地馬拉", "虛擬碼", "點陣圖", "那杜", "文字", "文件", "檔案", "資料夾", "副檔名", "檔名", "汶萊", "文書處理", "烏茲別克", "無失真壓縮", "實體地址", "實體記憶體", "解構函式", "下拉選單", "快捷半導體", "視訊記憶體", "顯示卡", "映象管", "執行緒", "相簿", "夏農", "專案", "畫素", "訊息", "防寫", "解除安裝", "晶片", "紐西蘭", "通道", "訊號", "資訊", "資訊保安", "資訊科技", "資訊理論", "訊雜比", "價效比", "效能", "虛擬函式", "虛擬機器", "迴圈", "定址", "簡報", "葉門", "頁尾", "頁首", "行動電話", "行動通訊", "行動網路", "行動硬碟", "行動資料", "乙太網", "非同步", "義大利", "溢位", "音訊", "載入程式", "對映", "硬體", "硬碟", "使用者", "使用者名稱", "優先順序", "有失真壓縮", "前處理器", "超程式設計", "後設資料", "母音", "原始碼", "原始碼", "原始碼", "原始檔", "遠端", "雲端儲存", "雲端計算", "執行", "運算子", "線上", "尚比亞", "查德", "貼上", "正規表示式", "支援", "智慧財產權", "指標", "智慧", "中介軟體", "重新命名", "過載", "重灌", "主機板", "主開機記錄", "登錄檔", "序號產生器", "登出", "狀態列", "桌上型", "自動旋轉螢幕", "腳踏車", "欄位", "字元", "字串", "位元組", "字型檔", "字型", "匯流排", "元件", "終端使用者"];
    var _t2s_t = ["㑮", "㑯", "㑳", "㑶", "㒓", "㓄", "㓨", "㔋", "㖮", "㗲", "㗿", "㘉", "㘓", "㘔", "㘚", "㛝", "㜄", "㜏", "㜐", "㜗", "㜢", "㜷", "㞞", "㟺", "㠏", "㢗", "㢝", "㥮", "㦎", "㦛", "㦞", "㨻", "㩋", "㩜", "㩳", "㩵", "㪎", "㯤", "㰙", "㵗", "㵾", "㶆", "㷍", "㷿", "㸇", "㹽", "㺏", "㺜", "㻶", "㿖", "㿗", "㿧", "䀉", "䀹", "䁪", "䁻", "䂎", "䃮", "䅐", "䅳", "䆉", "䉑", "䉙", "䉬", "䉲", "䉶", "䊭", "䊷", "䊺", "䋃", "䋔", "䋙", "䋚", "䋦", "䋹", "䋻", "䋼", "䋿", "䌈", "䌋", "䌖", "䌝", "䌟", "䌥", "䌰", "䍤", "䍦", "䍽", "䎙", "䎱", "䕤", "䕳", "䖅", "䗅", "䗿", "䙔", "䙡", "䙱", "䚩", "䛄", "䛳", "䜀", "䜖", "䝭", "䝻", "䝼", "䞈", "䞋", "䞓", "䟃", "䟆", "䟐", "䠆", "䠱", "䡐", "䡩", "䡵", "䢨", "䤤", "䥄", "䥇", "䥑", "䥗", "䥩", "䥯", "䥱", "䦘", "䦛", "䦟", "䦯", "䦳", "䧢", "䪊", "䪏", "䪗", "䪘", "䪴", "䪾", "䫀", "䫂", "䫟", "䫴", "䫶", "䫻", "䫾", "䬓", "䬘", "䬝", "䬞", "䬧", "䭀", "䭃", "䭑", "䭔", "䭿", "䮄", "䮝", "䮞", "䮠", "䮫", "䮰", "䮳", "䮾", "䯀", "䯤", "䰾", "䱀", "䱁", "䱙", "䱧", "䱬", "䱰", "䱷", "䱸", "䱽", "䲁", "䲅", "䲖", "䲘", "䲰", "䳜", "䳢", "䳤", "䳧", "䳫", "䴉", "䴋", "䴬", "䴱", "䴴", "䴽", "䵳", "䵴", "䶕", "䶲", "丟", "並", "乾", "亂", "亙", "亞", "佇", "佈", "佔", "併", "來", "侖", "侶", "侷", "俁", "係", "俓", "俔", "俠", "俥", "俬", "倀", "倆", "倈", "倉", "個", "們", "倖", "倫", "倲", "偉", "偑", "側", "偵", "偽", "傌", "傑", "傖", "傘", "備", "傢", "傭", "傯", "傳", "傴", "債", "傷", "傾", "僂", "僅", "僉", "僑", "僕", "僞", "僥", "僨", "僱", "價", "儀", "儁", "儂", "億", "儈", "儉", "儎", "儐", "儔", "儕", "儘", "償", "儣", "優", "儭", "儲", "儷", "儸", "儺", "儻", "儼", "兇", "兌", "兒", "兗", "內", "兩", "冊", "冑", "冪", "凈", "凍", "凙", "凜", "凱", "別", "刪", "剄", "則", "剋", "剎", "剗", "剛", "剝", "剮", "剴", "創", "剷", "剾", "劃", "劇", "劉", "劊", "劌", "劍", "劏", "劑", "劚", "勁", "勑", "動", "務", "勛", "勝", "勞", "勢", "勣", "勩", "勱", "勳", "勵", "勸", "勻", "匭", "匯", "匱", "區", "協", "卹", "卻", "卽", "厙", "厠", "厤", "厭", "厲", "厴", "參", "叄", "叢", "吒", "吳", "吶", "呂", "咼", "員", "哯", "唄", "唓", "唚", "唸", "問", "啓", "啞", "啟", "啢", "喎", "喚", "喪", "喫", "喬", "單", "喲", "嗆", "嗇", "嗊", "嗎", "嗚", "嗩", "嗰", "嗶", "嗹", "嘆", "嘍", "嘓", "嘔", "嘖", "嘗", "嘜", "嘩", "嘪", "嘮", "嘯", "嘰", "嘳", "嘵", "嘸", "嘺", "嘽", "噁", "噅", "噓", "噚", "噝", "噞", "噠", "噥", "噦", "噯", "噲", "噴", "噸", "噹", "嚀", "嚇", "嚌", "嚐", "嚕", "嚙", "嚛", "嚥", "嚦", "嚧", "嚨", "嚮", "嚲", "嚳", "嚴", "嚶", "嚽", "囀", "囁", "囂", "囃", "囅", "囈", "囉", "囌", "囑", "囒", "囪", "圇", "國", "圍", "園", "圓", "圖", "團", "圞", "垵", "埡", "埬", "埰", "執", "堅", "堊", "堖", "堚", "堝", "堯", "報", "場", "塊", "塋", "塏", "塒", "塗", "塚", "塢", "塤", "塵", "塹", "塿", "墊", "墜", "墮", "墰", "墲", "墳", "墶", "墻", "墾", "壇", "壈", "壋", "壎", "壓", "壗", "壘", "壙", "壚", "壜", "壞", "壟", "壠", "壢", "壣", "壩", "壪", "壯", "壺", "壼", "壽", "夠", "夢", "夥", "夾", "奐", "奧", "奩", "奪", "奬", "奮", "奼", "妝", "姍", "姦", "娛", "婁", "婡", "婦", "婭", "媈", "媧", "媯", "媰", "媼", "媽", "嫋", "嫗", "嫵", "嫺", "嫻", "嫿", "嬀", "嬃", "嬇", "嬈", "嬋", "嬌", "嬙", "嬡", "嬣", "嬤", "嬦", "嬪", "嬰", "嬸", "嬻", "孃", "孄", "孆", "孇", "孋", "孌", "孎", "孫", "學", "孻", "孾", "孿", "宮", "寀", "寠", "寢", "實", "寧", "審", "寫", "寬", "寵", "寶", "將", "專", "尋", "對", "導", "尷", "屆", "屍", "屓", "屜", "屢", "層", "屨", "屩", "屬", "岡", "峯", "峴", "島", "峽", "崍", "崑", "崗", "崙", "崢", "崬", "嵐", "嵗", "嵼", "嵾", "嶁", "嶄", "嶇", "嶈", "嶔", "嶗", "嶘", "嶠", "嶢", "嶧", "嶨", "嶮", "嶴", "嶸", "嶹", "嶺", "嶼", "嶽", "巊", "巋", "巒", "巔", "巖", "巗", "巘", "巰", "巹", "帥", "師", "帳", "帶", "幀", "幃", "幓", "幗", "幘", "幝", "幟", "幣", "幩", "幫", "幬", "幹", "幾", "庫", "廁", "廂", "廄", "廈", "廎", "廕", "廚", "廝", "廟", "廠", "廡", "廢", "廣", "廧", "廩", "廬", "廳", "弒", "弔", "弳", "張", "強", "彃", "彆", "彈", "彌", "彎", "彔", "彙", "彞", "彠", "彥", "彫", "彲", "彷", "彿", "後", "徑", "從", "徠", "復", "徵", "徹", "徿", "恆", "恥", "悅", "悞", "悵", "悶", "悽", "惡", "惱", "惲", "惻", "愛", "愜", "愨", "愴", "愷", "愻", "愾", "慄", "態", "慍", "慘", "慚", "慟", "慣", "慤", "慪", "慫", "慮", "慳", "慶", "慺", "慼", "慾", "憂", "憊", "憐", "憑", "憒", "憖", "憚", "憢", "憤", "憫", "憮", "憲", "憶", "憸", "憹", "懀", "懇", "應", "懌", "懍", "懎", "懞", "懟", "懣", "懤", "懨", "懲", "懶", "懷", "懸", "懺", "懼", "懾", "戀", "戇", "戔", "戧", "戩", "戰", "戱", "戲", "戶", "拋", "挩", "挱", "挾", "捨", "捫", "捱", "捲", "掃", "掄", "掆", "掗", "掙", "掚", "掛", "採", "揀", "揚", "換", "揮", "揯", "損", "搖", "搗", "搵", "搶", "摋", "摐", "摑", "摜", "摟", "摯", "摳", "摶", "摺", "摻", "撈", "撊", "撏", "撐", "撓", "撝", "撟", "撣", "撥", "撧", "撫", "撲", "撳", "撻", "撾", "撿", "擁", "擄", "擇", "擊", "擋", "擓", "擔", "據", "擟", "擠", "擡", "擣", "擫", "擬", "擯", "擰", "擱", "擲", "擴", "擷", "擺", "擻", "擼", "擽", "擾", "攄", "攆", "攋", "攏", "攔", "攖", "攙", "攛", "攜", "攝", "攢", "攣", "攤", "攪", "攬", "敎", "敓", "敗", "敘", "敵", "數", "斂", "斃", "斅", "斆", "斕", "斬", "斷", "斸", "於", "旂", "旣", "昇", "時", "晉", "晝", "暈", "暉", "暘", "暢", "暫", "曄", "曆", "曇", "曉", "曊", "曏", "曖", "曠", "曥", "曨", "曬", "書", "會", "朥", "朧", "朮", "東", "杴", "枴", "柵", "柺", "査", "桱", "桿", "梔", "梖", "梘", "條", "梟", "梲", "棄", "棊", "棖", "棗", "棟", "棡", "棧", "棲", "棶", "椏", "椲", "楇", "楊", "楓", "楨", "業", "極", "榘", "榦", "榪", "榮", "榲", "榿", "構", "槍", "槓", "槤", "槧", "槨", "槫", "槮", "槳", "槶", "槼", "樁", "樂", "樅", "樑", "樓", "標", "樞", "樠", "樢", "樣", "樤", "樧", "樫", "樳", "樸", "樹", "樺", "樿", "橈", "橋", "機", "橢", "橫", "橯", "檁", "檉", "檔", "檜", "檟", "檢", "檣", "檭", "檮", "檯", "檳", "檵", "檸", "檻", "檾", "櫃", "櫅", "櫓", "櫚", "櫛", "櫝", "櫞", "櫟", "櫠", "櫥", "櫧", "櫨", "櫪", "櫫", "櫬", "櫱", "櫳", "櫸", "櫺", "櫻", "欄", "欅", "欇", "權", "欍", "欏", "欐", "欑", "欒", "欓", "欖", "欘", "欞", "欽", "歎", "歐", "歟", "歡", "歲", "歷", "歸", "歿", "殘", "殞", "殢", "殤", "殨", "殫", "殭", "殮", "殯", "殰", "殲", "殺", "殻", "殼", "毀", "毆", "毊", "毿", "氂", "氈", "氌", "氣", "氫", "氬", "氭", "氳", "氾", "汎", "汙", "決", "沈", "沒", "沖", "況", "泝", "洩", "洶", "浹", "涇", "涗", "涼", "淒", "淚", "淥", "淨", "淩", "淪", "淵", "淶", "淺", "渙", "減", "渢", "渦", "測", "渾", "湊", "湋", "湞", "湧", "湯", "溈", "準", "溝", "溡", "溫", "溮", "溳", "溼", "滄", "滅", "滌", "滎", "滙", "滬", "滯", "滲", "滷", "滸", "滻", "滾", "滿", "漁", "漊", "漚", "漢", "漣", "漬", "漲", "漵", "漸", "漿", "潁", "潑", "潔", "潕", "潙", "潚", "潛", "潣", "潤", "潯", "潰", "潷", "潿", "澀", "澅", "澆", "澇", "澐", "澗", "澠", "澤", "澦", "澩", "澬", "澮", "澱", "澾", "濁", "濃", "濄", "濆", "濕", "濘", "濚", "濛", "濜", "濟", "濤", "濧", "濫", "濰", "濱", "濺", "濼", "濾", "濿", "瀂", "瀃", "瀅", "瀆", "瀇", "瀉", "瀋", "瀏", "瀕", "瀘", "瀝", "瀟", "瀠", "瀦", "瀧", "瀨", "瀰", "瀲", "瀾", "灃", "灄", "灍", "灑", "灒", "灕", "灘", "灙", "灝", "灡", "灣", "灤", "灧", "灩", "災", "為", "烏", "烴", "無", "煇", "煉", "煒", "煙", "煢", "煥", "煩", "煬", "煱", "熂", "熅", "熉", "熌", "熒", "熓", "熗", "熚", "熡", "熱", "熲", "熾", "燁", "燈", "燉", "燒", "燙", "燜", "營", "燦", "燬", "燭", "燴", "燶", "燻", "燼", "燾", "爃", "爄", "爇", "爍", "爐", "爖", "爛", "爥", "爧", "爭", "爲", "爺", "爾", "牀", "牆", "牘", "牴", "牽", "犖", "犛", "犞", "犢", "犧", "狀", "狹", "狽", "猌", "猙", "猶", "猻", "獁", "獃", "獄", "獅", "獊", "獎", "獨", "獩", "獪", "獫", "獮", "獰", "獱", "獲", "獵", "獷", "獸", "獺", "獻", "獼", "玀", "玁", "珼", "現", "琱", "琺", "琿", "瑋", "瑒", "瑣", "瑤", "瑩", "瑪", "瑲", "瑻", "瑽", "璉", "璊", "璝", "璡", "璣", "璦", "璫", "璯", "環", "璵", "璸", "璼", "璽", "璾", "瓄", "瓊", "瓏", "瓔", "瓕", "瓚", "瓛", "甌", "甕", "產", "産", "甦", "甯", "畝", "畢", "畫", "異", "畵", "當", "畼", "疇", "疊", "痙", "痠", "痮", "痾", "瘂", "瘋", "瘍", "瘓", "瘞", "瘡", "瘧", "瘮", "瘱", "瘲", "瘺", "瘻", "療", "癆", "癇", "癉", "癐", "癒", "癘", "癟", "癡", "癢", "癤", "癥", "癧", "癩", "癬", "癭", "癮", "癰", "癱", "癲", "發", "皁", "皚", "皟", "皰", "皸", "皺", "盃", "盜", "盞", "盡", "監", "盤", "盧", "盨", "盪", "眝", "眞", "眥", "眾", "睍", "睏", "睜", "睞", "睪", "瞘", "瞜", "瞞", "瞤", "瞭", "瞶", "瞼", "矇", "矉", "矑", "矓", "矚", "矯", "硃", "硜", "硤", "硨", "硯", "碕", "碙", "碩", "碭", "碸", "確", "碼", "碽", "磑", "磚", "磠", "磣", "磧", "磯", "磽", "磾", "礄", "礆", "礎", "礒", "礙", "礦", "礪", "礫", "礬", "礮", "礱", "祇", "祕", "祿", "禍", "禎", "禕", "禡", "禦", "禪", "禮", "禰", "禱", "禿", "秈", "稅", "稈", "稏", "稜", "稟", "種", "稱", "穀", "穇", "穌", "積", "穎", "穠", "穡", "穢", "穩", "穫", "穭", "窩", "窪", "窮", "窯", "窵", "窶", "窺", "竄", "竅", "竇", "竈", "竊", "竚", "竪", "竱", "競", "筆", "筍", "筧", "筴", "箇", "箋", "箏", "節", "範", "築", "篋", "篔", "篘", "篠", "篤", "篩", "篳", "篸", "簀", "簂", "簍", "簑", "簞", "簡", "簢", "簣", "簫", "簹", "簽", "簾", "籃", "籅", "籋", "籌", "籔", "籙", "籛", "籜", "籟", "籠", "籤", "籩", "籪", "籬", "籮", "籲", "粵", "糉", "糝", "糞", "糧", "糰", "糲", "糴", "糶", "糹", "糺", "糾", "紀", "紂", "約", "紅", "紆", "紇", "紈", "紉", "紋", "納", "紐", "紓", "純", "紕", "紖", "紗", "紘", "紙", "級", "紛", "紜", "紝", "紟", "紡", "紬", "紮", "細", "紱", "紲", "紳", "紵", "紹", "紺", "紼", "紿", "絀", "絁", "終", "絃", "組", "絅", "絆", "絍", "絎", "結", "絕", "絙", "絛", "絝", "絞", "絡", "絢", "絥", "給", "絧", "絨", "絰", "統", "絲", "絳", "絶", "絹", "絺", "綀", "綁", "綃", "綆", "綇", "綈", "綉", "綋", "綌", "綏", "綐", "綑", "經", "綖", "綜", "綞", "綟", "綠", "綡", "綢", "綣", "綫", "綬", "維", "綯", "綰", "綱", "網", "綳", "綴", "綵", "綸", "綹", "綺", "綻", "綽", "綾", "綿", "緄", "緇", "緊", "緋", "緍", "緑", "緒", "緓", "緔", "緗", "緘", "緙", "線", "緝", "緞", "緟", "締", "緡", "緣", "緤", "緦", "編", "緩", "緬", "緮", "緯", "緰", "緱", "緲", "練", "緶", "緷", "緸", "緹", "緻", "緼", "縈", "縉", "縊", "縋", "縍", "縎", "縐", "縑", "縕", "縗", "縛", "縝", "縞", "縟", "縣", "縧", "縫", "縬", "縭", "縮", "縰", "縱", "縲", "縳", "縴", "縵", "縶", "縷", "縸", "縹", "縺", "總", "績", "繂", "繃", "繅", "繆", "繈", "繏", "繐", "繒", "繓", "織", "繕", "繚", "繞", "繟", "繡", "繢", "繨", "繩", "繪", "繫", "繬", "繭", "繮", "繯", "繰", "繳", "繶", "繷", "繸", "繹", "繻", "繼", "繽", "繾", "繿", "纁", "纇", "纈", "纊", "續", "纍", "纏", "纓", "纔", "纖", "纗", "纘", "纚", "纜", "缽", "罃", "罈", "罌", "罎", "罰", "罵", "罷", "羅", "羆", "羈", "羋", "羣", "羥", "羨", "義", "羵", "羶", "習", "翬", "翹", "翽", "耬", "耮", "聖", "聞", "聯", "聰", "聲", "聳", "聵", "聶", "職", "聹", "聻", "聽", "聾", "肅", "脅", "脈", "脛", "脣", "脥", "脩", "脫", "脹", "腎", "腖", "腡", "腦", "腪", "腫", "腳", "腸", "膃", "膕", "膚", "膞", "膠", "膢", "膩", "膹", "膽", "膾", "膿", "臉", "臍", "臏", "臗", "臘", "臚", "臟", "臠", "臢", "臥", "臨", "臺", "與", "興", "舉", "舊", "舘", "艙", "艣", "艤", "艦", "艫", "艱", "艷", "芻", "苧", "茲", "荊", "莊", "莖", "莢", "莧", "菕", "華", "菴", "菸", "萇", "萊", "萬", "萴", "萵", "葉", "葒", "葝", "葤", "葦", "葯", "葷", "蒍", "蒐", "蒓", "蒔", "蒕", "蒞", "蒭", "蒼", "蓀", "蓆", "蓋", "蓧", "蓮", "蓯", "蓴", "蓽", "蔔", "蔘", "蔞", "蔣", "蔥", "蔦", "蔭", "蔯", "蔿", "蕁", "蕆", "蕎", "蕒", "蕓", "蕕", "蕘", "蕝", "蕢", "蕩", "蕪", "蕭", "蕳", "蕷", "蕽", "薀", "薆", "薈", "薊", "薌", "薑", "薔", "薘", "薟", "薦", "薩", "薰", "薳", "薴", "薵", "薹", "薺", "藉", "藍", "藎", "藝", "藥", "藪", "藭", "藴", "藶", "藷", "藹", "藺", "蘀", "蘄", "蘆", "蘇", "蘊", "蘋", "蘚", "蘞", "蘟", "蘢", "蘭", "蘺", "蘿", "虆", "處", "虛", "虜", "號", "虧", "虯", "蛺", "蛻", "蜆", "蝕", "蝟", "蝦", "蝨", "蝸", "螄", "螞", "螢", "螮", "螻", "螿", "蟂", "蟄", "蟈", "蟎", "蟘", "蟜", "蟣", "蟬", "蟯", "蟲", "蟳", "蟶", "蟻", "蠀", "蠁", "蠅", "蠆", "蠍", "蠐", "蠑", "蠔", "蠙", "蠟", "蠣", "蠦", "蠨", "蠱", "蠶", "蠻", "蠾", "衆", "衊", "術", "衕", "衚", "衛", "衝", "衹", "袞", "裊", "裏", "補", "裝", "裡", "製", "複", "褌", "褘", "褲", "褳", "褸", "褻", "襀", "襆", "襇", "襉", "襏", "襓", "襖", "襗", "襘", "襝", "襠", "襤", "襪", "襬", "襯", "襰", "襲", "襴", "襵", "覆", "覈", "見", "覎", "規", "覓", "視", "覘", "覛", "覡", "覥", "覦", "親", "覬", "覯", "覲", "覷", "覹", "覺", "覼", "覽", "覿", "觀", "觴", "觶", "觸", "訁", "訂", "訃", "計", "訊", "訌", "討", "訐", "訑", "訒", "訓", "訕", "訖", "託", "記", "訛", "訜", "訝", "訞", "訟", "訢", "訣", "訥", "訨", "訩", "訪", "設", "許", "訴", "訶", "診", "註", "証", "詀", "詁", "詆", "詊", "詎", "詐", "詑", "詒", "詓", "詔", "評", "詖", "詗", "詘", "詛", "詞", "詠", "詡", "詢", "詣", "試", "詩", "詫", "詬", "詭", "詮", "詰", "話", "該", "詳", "詵", "詷", "詼", "詿", "誂", "誄", "誅", "誆", "誇", "誋", "誌", "認", "誑", "誒", "誕", "誘", "誚", "語", "誠", "誡", "誣", "誤", "誥", "誦", "誨", "說", "誫", "説", "誰", "課", "誳", "誴", "誶", "誷", "誹", "誺", "誼", "誾", "調", "諂", "諄", "談", "諉", "請", "諍", "諏", "諑", "諒", "論", "諗", "諛", "諜", "諝", "諞", "諡", "諢", "諣", "諤", "諥", "諦", "諧", "諫", "諭", "諮", "諯", "諰", "諱", "諳", "諴", "諶", "諷", "諸", "諺", "諼", "諾", "謀", "謁", "謂", "謄", "謅", "謆", "謉", "謊", "謎", "謏", "謐", "謔", "謖", "謗", "謙", "謚", "講", "謝", "謠", "謡", "謨", "謫", "謬", "謭", "謯", "謱", "謳", "謸", "謹", "謾", "譁", "譂", "譅", "譆", "證", "譊", "譎", "譏", "譑", "譖", "識", "譙", "譚", "譜", "譞", "譟", "譨", "譫", "譭", "譯", "議", "譴", "護", "譸", "譽", "譾", "讀", "讅", "變", "讋", "讌", "讎", "讒", "讓", "讕", "讖", "讚", "讜", "讞", "豈", "豎", "豐", "豔", "豬", "豵", "豶", "貓", "貗", "貙", "貝", "貞", "貟", "負", "財", "貢", "貧", "貨", "販", "貪", "貫", "責", "貯", "貰", "貲", "貳", "貴", "貶", "買", "貸", "貺", "費", "貼", "貽", "貿", "賀", "賁", "賂", "賃", "賄", "賅", "資", "賈", "賊", "賑", "賒", "賓", "賕", "賙", "賚", "賜", "賝", "賞", "賟", "賠", "賡", "賢", "賣", "賤", "賦", "賧", "質", "賫", "賬", "賭", "賰", "賴", "賵", "賺", "賻", "購", "賽", "賾", "贃", "贄", "贅", "贇", "贈", "贉", "贊", "贋", "贍", "贏", "贐", "贑", "贓", "贔", "贖", "贗", "贚", "贛", "贜", "赬", "趕", "趙", "趨", "趲", "跡", "踐", "踰", "踴", "蹌", "蹔", "蹕", "蹟", "蹣", "蹤", "蹳", "蹺", "蹻", "躂", "躉", "躊", "躋", "躍", "躎", "躑", "躒", "躓", "躕", "躘", "躚", "躝", "躡", "躥", "躦", "躪", "軀", "軉", "車", "軋", "軌", "軍", "軏", "軑", "軒", "軔", "軕", "軗", "軛", "軜", "軟", "軤", "軨", "軫", "軬", "軲", "軷", "軸", "軹", "軺", "軻", "軼", "軾", "軿", "較", "輄", "輅", "輇", "輈", "載", "輊", "輋", "輒", "輓", "輔", "輕", "輖", "輗", "輛", "輜", "輝", "輞", "輟", "輢", "輥", "輦", "輨", "輩", "輪", "輬", "輮", "輯", "輳", "輷", "輸", "輻", "輼", "輾", "輿", "轀", "轂", "轄", "轅", "轆", "轇", "轉", "轊", "轍", "轎", "轐", "轔", "轗", "轟", "轠", "轡", "轢", "轣", "轤", "辦", "辭", "辮", "辯", "農", "迴", "逕", "這", "連", "週", "進", "遊", "運", "過", "達", "違", "遙", "遜", "遞", "遠", "遡", "適", "遱", "遲", "遷", "選", "遺", "遼", "邁", "還", "邇", "邊", "邏", "邐", "郟", "郵", "鄆", "鄉", "鄒", "鄔", "鄖", "鄟", "鄧", "鄭", "鄰", "鄲", "鄳", "鄴", "鄶", "鄺", "酇", "酈", "醃", "醖", "醜", "醞", "醟", "醣", "醫", "醬", "醱", "醶", "釀", "釁", "釃", "釅", "釋", "釐", "釒", "釓", "釔", "釕", "釗", "釘", "釙", "釚", "針", "釟", "釣", "釤", "釦", "釧", "釨", "釩", "釲", "釳", "釵", "釷", "釹", "釺", "釾", "鈀", "鈁", "鈃", "鈄", "鈅", "鈆", "鈇", "鈈", "鈉", "鈋", "鈍", "鈎", "鈐", "鈑", "鈒", "鈔", "鈕", "鈖", "鈗", "鈛", "鈞", "鈠", "鈡", "鈣", "鈥", "鈦", "鈧", "鈮", "鈯", "鈰", "鈲", "鈳", "鈴", "鈷", "鈸", "鈹", "鈺", "鈽", "鈾", "鈿", "鉀", "鉁", "鉅", "鉆", "鉈", "鉉", "鉋", "鉍", "鉑", "鉔", "鉕", "鉗", "鉚", "鉛", "鉝", "鉞", "鉠", "鉢", "鉤", "鉦", "鉬", "鉭", "鉳", "鉶", "鉷", "鉸", "鉺", "鉻", "鉽", "鉾", "鉿", "銀", "銁", "銂", "銃", "銅", "銈", "銊", "銍", "銏", "銑", "銓", "銖", "銘", "銚", "銛", "銜", "銠", "銣", "銥", "銦", "銨", "銩", "銪", "銫", "銬", "銱", "銳", "銶", "銷", "銹", "銻", "銼", "鋁", "鋂", "鋃", "鋅", "鋇", "鋉", "鋌", "鋏", "鋒", "鋗", "鋙", "鋝", "鋟", "鋠", "鋣", "鋤", "鋥", "鋦", "鋨", "鋩", "鋪", "鋭", "鋮", "鋯", "鋰", "鋱", "鋶", "鋸", "鋼", "錀", "錁", "錂", "錄", "錆", "錇", "錈", "錏", "錐", "錒", "錕", "錘", "錙", "錚", "錛", "錜", "錝", "錟", "錠", "錡", "錢", "錤", "錥", "錦", "錨", "錩", "錫", "錮", "錯", "録", "錳", "錶", "錸", "錼", "錽", "鍀", "鍁", "鍃", "鍄", "鍅", "鍆", "鍇", "鍈", "鍉", "鍊", "鍋", "鍍", "鍒", "鍔", "鍘", "鍚", "鍛", "鍠", "鍤", "鍥", "鍩", "鍬", "鍮", "鍰", "鍵", "鍶", "鍺", "鍼", "鍾", "鎂", "鎄", "鎇", "鎈", "鎊", "鎌", "鎍", "鎔", "鎖", "鎘", "鎙", "鎚", "鎛", "鎝", "鎞", "鎡", "鎢", "鎣", "鎦", "鎧", "鎩", "鎪", "鎬", "鎭", "鎮", "鎯", "鎰", "鎲", "鎳", "鎵", "鎶", "鎷", "鎸", "鎿", "鏃", "鏆", "鏇", "鏈", "鏉", "鏌", "鏍", "鏐", "鏑", "鏗", "鏘", "鏚", "鏜", "鏝", "鏞", "鏟", "鏡", "鏢", "鏤", "鏥", "鏦", "鏨", "鏰", "鏵", "鏷", "鏹", "鏺", "鏽", "鏾", "鐃", "鐄", "鐇", "鐈", "鐋", "鐍", "鐎", "鐏", "鐐", "鐒", "鐓", "鐔", "鐗", "鐘", "鐙", "鐝", "鐠", "鐥", "鐦", "鐧", "鐨", "鐪", "鐫", "鐮", "鐯", "鐲", "鐳", "鐵", "鐶", "鐸", "鐺", "鐼", "鐽", "鐿", "鑀", "鑄", "鑉", "鑊", "鑌", "鑑", "鑒", "鑔", "鑕", "鑞", "鑠", "鑣", "鑥", "鑪", "鑭", "鑰", "鑱", "鑲", "鑴", "鑷", "鑹", "鑼", "鑽", "鑾", "鑿", "钁", "钂", "镟", "長", "門", "閂", "閃", "閆", "閈", "閉", "開", "閌", "閍", "閎", "閏", "閐", "閑", "閒", "間", "閔", "閗", "閘", "閝", "閞", "閡", "閣", "閤", "閥", "閨", "閩", "閫", "閬", "閭", "閱", "閲", "閵", "閶", "閹", "閻", "閼", "閽", "閾", "閿", "闃", "闆", "闇", "闈", "闊", "闋", "闌", "闍", "闐", "闑", "闒", "闓", "闔", "闕", "闖", "關", "闞", "闠", "闡", "闢", "闤", "闥", "阪", "陘", "陝", "陞", "陣", "陰", "陳", "陸", "陽", "隉", "隊", "階", "隕", "際", "隨", "險", "隯", "隱", "隴", "隸", "隻", "雋", "雖", "雙", "雛", "雜", "雞", "離", "難", "雲", "電", "霢", "霣", "霧", "霼", "霽", "靂", "靄", "靆", "靈", "靉", "靚", "靜", "靝", "靦", "靧", "靨", "鞀", "鞏", "鞝", "鞦", "鞽", "鞾", "韁", "韃", "韆", "韉", "韋", "韌", "韍", "韓", "韙", "韚", "韛", "韜", "韝", "韞", "韠", "韻", "響", "頁", "頂", "頃", "項", "順", "頇", "須", "頊", "頌", "頍", "頎", "頏", "預", "頑", "頒", "頓", "頗", "領", "頜", "頡", "頤", "頦", "頫", "頭", "頮", "頰", "頲", "頴", "頵", "頷", "頸", "頹", "頻", "頽", "顂", "顃", "顅", "顆", "題", "額", "顎", "顏", "顒", "顓", "顔", "顗", "願", "顙", "顛", "類", "顢", "顣", "顥", "顧", "顫", "顬", "顯", "顰", "顱", "顳", "顴", "風", "颭", "颮", "颯", "颰", "颱", "颳", "颶", "颷", "颸", "颺", "颻", "颼", "颾", "飀", "飄", "飆", "飈", "飋", "飛", "飠", "飢", "飣", "飥", "飦", "飩", "飪", "飫", "飭", "飯", "飱", "飲", "飴", "飵", "飶", "飼", "飽", "飾", "飿", "餃", "餄", "餅", "餉", "養", "餌", "餎", "餏", "餑", "餒", "餓", "餔", "餕", "餖", "餗", "餘", "餚", "餛", "餜", "餞", "餡", "餦", "餧", "館", "餪", "餫", "餬", "餭", "餱", "餳", "餵", "餶", "餷", "餸", "餺", "餼", "餾", "餿", "饁", "饃", "饅", "饈", "饉", "饊", "饋", "饌", "饑", "饒", "饗", "饘", "饜", "饞", "饟", "饠", "饢", "馬", "馭", "馮", "馯", "馱", "馳", "馴", "馹", "馼", "駁", "駃", "駊", "駎", "駐", "駑", "駒", "駔", "駕", "駘", "駙", "駚", "駛", "駝", "駞", "駟", "駡", "駢", "駤", "駧", "駩", "駫", "駭", "駰", "駱", "駶", "駸", "駻", "駿", "騁", "騂", "騃", "騄", "騅", "騉", "騊", "騌", "騍", "騎", "騏", "騔", "騖", "騙", "騚", "騜", "騝", "騟", "騠", "騤", "騧", "騪", "騫", "騭", "騮", "騰", "騱", "騴", "騵", "騶", "騷", "騸", "騻", "騼", "騾", "驀", "驁", "驂", "驃", "驄", "驅", "驊", "驋", "驌", "驍", "驏", "驓", "驕", "驗", "驙", "驚", "驛", "驟", "驢", "驤", "驥", "驦", "驨", "驪", "驫", "骯", "髏", "髒", "體", "髕", "髖", "髮", "鬆", "鬍", "鬖", "鬚", "鬠", "鬢", "鬥", "鬧", "鬨", "鬩", "鬮", "鬱", "鬹", "魎", "魘", "魚", "魛", "魟", "魢", "魥", "魦", "魨", "魯", "魴", "魵", "魷", "魺", "魽", "鮁", "鮃", "鮄", "鮅", "鮆", "鮊", "鮋", "鮍", "鮎", "鮐", "鮑", "鮒", "鮓", "鮕", "鮚", "鮜", "鮝", "鮞", "鮟", "鮣", "鮤", "鮦", "鮪", "鮫", "鮭", "鮮", "鮯", "鮰", "鮳", "鮵", "鮶", "鮸", "鮺", "鮿", "鯀", "鯁", "鯄", "鯆", "鯇", "鯉", "鯊", "鯒", "鯔", "鯕", "鯖", "鯗", "鯛", "鯝", "鯞", "鯡", "鯢", "鯤", "鯧", "鯨", "鯪", "鯫", "鯬", "鯰", "鯱", "鯴", "鯶", "鯷", "鯽", "鯾", "鯿", "鰁", "鰂", "鰃", "鰆", "鰈", "鰉", "鰋", "鰌", "鰍", "鰏", "鰐", "鰑", "鰒", "鰓", "鰕", "鰛", "鰜", "鰟", "鰠", "鰣", "鰤", "鰥", "鰦", "鰧", "鰨", "鰩", "鰫", "鰭", "鰮", "鰱", "鰲", "鰳", "鰵", "鰷", "鰹", "鰺", "鰻", "鰼", "鰽", "鰾", "鱂", "鱄", "鱅", "鱆", "鱇", "鱈", "鱉", "鱊", "鱒", "鱔", "鱖", "鱗", "鱘", "鱝", "鱟", "鱠", "鱢", "鱣", "鱤", "鱧", "鱨", "鱭", "鱮", "鱯", "鱲", "鱷", "鱸", "鱺", "鳥", "鳧", "鳩", "鳬", "鳲", "鳳", "鳴", "鳶", "鳷", "鳼", "鳽", "鳾", "鴀", "鴃", "鴅", "鴆", "鴇", "鴉", "鴐", "鴒", "鴔", "鴕", "鴗", "鴛", "鴜", "鴝", "鴞", "鴟", "鴣", "鴥", "鴦", "鴨", "鴮", "鴯", "鴰", "鴲", "鴳", "鴴", "鴷", "鴻", "鴽", "鴿", "鵁", "鵂", "鵃", "鵊", "鵐", "鵑", "鵒", "鵓", "鵚", "鵜", "鵝", "鵟", "鵠", "鵡", "鵧", "鵩", "鵪", "鵫", "鵬", "鵮", "鵯", "鵰", "鵲", "鵷", "鵾", "鶄", "鶇", "鶉", "鶊", "鶌", "鶒", "鶓", "鶖", "鶗", "鶘", "鶚", "鶡", "鶥", "鶦", "鶩", "鶪", "鶬", "鶭", "鶯", "鶰", "鶲", "鶴", "鶹", "鶺", "鶻", "鶼", "鶿", "鷀", "鷁", "鷂", "鷄", "鷅", "鷈", "鷉", "鷊", "鷐", "鷓", "鷔", "鷖", "鷗", "鷙", "鷚", "鷣", "鷤", "鷥", "鷦", "鷨", "鷩", "鷫", "鷯", "鷲", "鷳", "鷴", "鷷", "鷸", "鷹", "鷺", "鷽", "鷿", "鸂", "鸇", "鸊", "鸋", "鸌", "鸏", "鸕", "鸗", "鸘", "鸚", "鸛", "鸝", "鸞", "鹵", "鹹", "鹺", "鹼", "鹽", "麗", "麥", "麨", "麩", "麪", "麫", "麬", "麯", "麲", "麳", "麴", "麵", "麷", "麼", "麽", "黃", "黌", "點", "黨", "黲", "黴", "黶", "黷", "黽", "黿", "鼂", "鼉", "鼕", "鼴", "齇", "齊", "齋", "齎", "齏", "齒", "齔", "齕", "齗", "齙", "齜", "齟", "齠", "齡", "齣", "齦", "齧", "齩", "齪", "齬", "齭", "齯", "齰", "齲", "齴", "齶", "齷", "齾", "龍", "龎", "龐", "龑", "龓", "龔", "龕", "龜", "龭", "龯", "鿁", "鿓", "𠁞", "𠌥", "𠏢", "𠐊", "𠗣", "𠞆", "𠠎", "𠬙", "𠽃", "𠿕", "𡂡", "𡃄", "𡃕", "𡃤", "𡄔", "𡄣", "𡅏", "𡅯", "𡑭", "𡓁", "𡓾", "𡔖", "𡞵", "𡟫", "𡠹", "𡡎", "𡢃", "𡮉", "𡮣", "𡳳", "𡸗", "𡹬", "𡻕", "𡽗", "𡾱", "𡿖", "𢍰", "𢠼", "𢣐", "𢣚", "𢣭", "𢤩", "𢤱", "𢤿", "𢯷", "𢶒", "𢶫", "𢷬", "𢷮", "𢹿", "𢺳", "𣈶", "𣋋", "𣍐", "𣙎", "𣜬", "𣝕", "𣞻", "𣠩", "𣠲", "𣯩", "𣯴", "𣯶", "𣽏", "𣾷", "𣿉", "𤁣", "𤄷", "𤅶", "𤑳", "𤑹", "𤒎", "𤒻", "𤓌", "𤓩", "𤘀", "𤛮", "𤛱", "𤜆", "𤠮", "𤢟", "𤢻", "𤩂", "𤪺", "𤫩", "𤬅", "𤳷", "𤳸", "𤷃", "𤸫", "𤺔", "𥊝", "𥌃", "𥏝", "𥕥", "𥖅", "𥖲", "𥗇", "𥜐", "𥜰", "𥞵", "𥢢", "𥢶", "𥢷", "𥨐", "𥪂", "𥯤", "𥴨", "𥴼", "𥵃", "𥵊", "𥶽", "𥸠", "𥻦", "𥼽", "𥽖", "𥾯", "𥿊", "𦀖", "𦂅", "𦃄", "𦃩", "𦅇", "𦅈", "𦆲", "𦒀", "𦔖", "𦘧", "𦟼", "𦠅", "𦡝", "𦢈", "𦣎", "𦧺", "𦪙", "𦪽", "𦱌", "𦾟", "𧎈", "𧒯", "𧔥", "𧕟", "𧜗", "𧜵", "𧝞", "𧞫", "𧟀", "𧡴", "𧢄", "𧦝", "𧦧", "𧩕", "𧩙", "𧩼", "𧫝", "𧬤", "𧭈", "𧭹", "𧳟", "𧵳", "𧶔", "𧶧", "𧷎", "𧸘", "𧹈", "𧽯", "𨂐", "𨄣", "𨅍", "𨆪", "𨇁", "𨇞", "𨇤", "𨇰", "𨇽", "𨈊", "𨈌", "𨊰", "𨊸", "𨊻", "𨋢", "𨌈", "𨍰", "𨎌", "𨎮", "𨏠", "𨏥", "𨞺", "𨟊", "𨢿", "𨣈", "𨣞", "𨣧", "𨤻", "𨥛", "𨥟", "𨦫", "𨧀", "𨧜", "𨧰", "𨧱", "𨨏", "𨨛", "𨨢", "𨩰", "𨪕", "𨫒", "𨬖", "𨭆", "𨭎", "𨭖", "𨭸", "𨮂", "𨮳", "𨯅", "𨯟", "𨰃", "𨰋", "𨰥", "𨰲", "𨲳", "𨳑", "𨳕", "𨴗", "𨴹", "𨵩", "𨵸", "𨶀", "𨶏", "𨶮", "𨶲", "𨷲", "𨼳", "𨽏", "𩀨", "𩅙", "𩎖", "𩎢", "𩏂", "𩏠", "𩏪", "𩏷", "𩑔", "𩒎", "𩓣", "𩓥", "𩔑", "𩔳", "𩖰", "𩗀", "𩗓", "𩗴", "𩘀", "𩘝", "𩘹", "𩘺", "𩙈", "𩚛", "𩚥", "𩚩", "𩚵", "𩛆", "𩛌", "𩛡", "𩛩", "𩜇", "𩜦", "𩜵", "𩝔", "𩝽", "𩞄", "𩞦", "𩞯", "𩟐", "𩟗", "𩠴", "𩡣", "𩡺", "𩢡", "𩢴", "𩢸", "𩢾", "𩣏", "𩣑", "𩣫", "𩣵", "𩣺", "𩤊", "𩤙", "𩤲", "𩤸", "𩥄", "𩥇", "𩥉", "𩥑", "𩦠", "𩧆", "𩭙", "𩯁", "𩯳", "𩰀", "𩰹", "𩳤", "𩴵", "𩵦", "𩵩", "𩵹", "𩶁", "𩶘", "𩶰", "𩶱", "𩷰", "𩸃", "𩸄", "𩸡", "𩸦", "𩻗", "𩻬", "𩻮", "𩼶", "𩽇", "𩿅", "𩿤", "𩿪", "𪀖", "𪀦", "𪀾", "𪁈", "𪁖", "𪂆", "𪃍", "𪃏", "𪃒", "𪃧", "𪄆", "𪄕", "𪅂", "𪆷", "𪇳", "𪈼", "𪉸", "𪋿", "𪌭", "𪍠", "𪓰", "𪔵", "𪘀", "𪘯", "𪙏", "𪟖", "𪷓", "𫒡", "𫜦", "八濛山", "拜覆", "畢昇", "變徵", "稟覆", "朝乾夕惕", "籌畫", "出醜狼藉", "答覆", "大目乾連冥間救母變文", "蕩覆", "牴觸", "牴牾", "顛乾倒坤", "電覆", "貂覆額", "篤麼", "鍛鍾", "發覆", "樊於期", "反反覆覆", "反覆", "覆按", "覆杯", "覆瓿", "覆成", "覆呈", "覆幬", "覆電", "覆鼎", "覆醢", "覆醬瓿", "覆蕉尋鹿", "覆鹿尋蕉", "覆鹿遺蕉", "覆露", "覆冒", "覆命", "覆墓", "覆逆", "覆上", "覆文", "覆校", "覆信", "覆盂", "覆育", "覆帳", "覆住", "覆宗", "顧藉", "郭子乾", "函覆", "狐藉虎威", "滑藉", "黄鍾公", "回覆", "藉草枕塊", "藉詞", "藉此", "藉代", "藉讀", "藉端", "藉方", "藉槁", "藉故", "藉卉", "藉機", "藉藉", "藉據", "藉口", "藉寇兵", "藉甚", "藉手", "藉條", "藉以", "藉由", "藉喻", "藉着", "藉助", "藉箸代籌", "藉資", "計畫", "傢俱", "見覆", "角徵", "鉸鍊", "金鍊", "金吒", "頸鍊", "酒逢知己千鍾少", "鉅防", "鉅萬", "鉅子", "鈞覆", "康乾", "拉鍊", "李鍊福", "李乾德", "李澤鉅", "李鍾郁", "鍊錘", "鍊鎖", "鍊條", "鍊形", "鍊墜", "鍊子", "瞭解", "瞭然", "瞭如", "瞭若指掌", "躪藉", "凌藉", "流徵", "露覆", "買臣覆水", "么麽", "么麼", "麼些族", "名覆金甌", "明覆", "明瞭", "侔德覆載", "木吒", "哪吒", "尼乾陀", "鈕釦", "彷彿", "憑藉", "頗覆", "千鍾粟", "乾旦", "乾道", "乾斷", "乾綱", "乾卦", "乾紅", "乾嘉", "乾坤", "乾陵", "乾隆", "乾清宮", "乾盛世", "乾圖", "乾縣", "乾象", "乾曜", "乾元", "乾造", "乾宅", "錢鍾書", "傷亡枕藉", "上鍊", "射覆", "申覆", "沈船", "沈積", "沈沒", "沈默", "甚鉅", "示覆", "手鍊", "素藉", "蒜薹", "鎖鍊", "鐵鍊", "萬鍾", "文錦覆阱", "項鍊", "蕭乾", "旋乾轉坤", "踅門瞭戶", "雁杳魚沈", "幺麼", "以功覆過", "茵藉", "銀鍊", "於敖", "於崇文", "於單", "於乎", "於呼哀哉", "於梨華", "於琳", "於陵子", "於倫", "於穆", "於其一", "於清言", "於世成", "於坦", "於菟", "於惟一", "於戲", "於邑", "於勇明", "於則", "於忠祥", "於仲完", "於竹屋", "蕓薹", "醞藉", "張法乾", "衹得", "衹見樹木", "徵調", "徵聲", "徵弦", "徵絃", "徵音", "鍾鍛", "鍾馗", "鍾萬梅", "鍾重發", "重覆"];
    var _t2s_s = ["𫝈", "㑔", "㑇", "㐹", "𠉂", "𪠟", "刾", "𪟎", "𪠵", "𠵾", "𪡛", "𠰱", "𪢌", "㗷", "㘎", "𫝦", "㚯", "㛣", "𫝧", "𡞋", "𡞱", "𡝠", "𪨊", "𪩇", "㟆", "𪪑", "𢋈", "㤘", "𢛯", "𢗓", "𪫷", "𪮃", "𪮋", "㨫", "㧐", "擜", "𪯋", "𣘐", "𣗙", "𣳆", "𪷍", "𫞛", "𤆢", "𤈷", "𤎺", "𫞣", "𤠋", "𪺻", "𪼋", "𪽮", "𤻊", "𤽯", "𥁢", "𥅴", "𥇢", "䀥", "𥎝", "鿎", "𫀨", "𫀬", "𫁂", "𫁲", "𥬀", "𫂈", "𥮜", "𫁷", "𥺅", "䌶", "𫄚", "𫄜", "𫄞", "䌺", "䌻", "𫄩", "䌿", "䌾", "𫄮", "𦈓", "𦈖", "𦈘", "𦈜", "𦈟", "𦈞", "𦈠", "𦈙", "𫅅", "䍠", "𦍠", "𫅭", "䎬", "𫟕", "𦰴", "𫟑", "𫊪", "𧉞", "𫋲", "䙌", "𧜭", "𫌯", "𫍠", "𫍫", "䜧", "𫟢", "𫎧", "𧹕", "䞍", "𧹑", "𫎪", "𫎭", "𫎺", "𫎳", "𫎱", "𫏃", "𨅛", "𫟤", "𫟥", "𫟦", "𨑹", "𫟺", "𫠀", "䦂", "鿏", "𫔋", "𨱖", "𫔆", "䥾", "𨸄", "䦶", "䦷", "𫔵", "𨷿", "𨸟", "𫖅", "𩏼", "𩐀", "𩏿", "𫖫", "𫖬", "𫖱", "𫖰", "𫖲", "𩖗", "𫖺", "𫗇", "𫠈", "𫗊", "𩙮", "𩙯", "𩙧", "𫗟", "𩠇", "𩠈", "𫗱", "𫗰", "𩧭", "𫠊", "𩧰", "𩨁", "𩧿", "𩨇", "𫘮", "𩨏", "𩧪", "䯅", "𩩈", "鲃", "𫚐", "𫚏", "𩾈", "𫚠", "𩾊", "𩾋", "䲣", "𫠑", "䲝", "鳚", "𫚜", "𩾂", "鳤", "𪉂", "𫛬", "𫛰", "𫛮", "𫛺", "𫛼", "鹮", "𫜅", "𪎈", "𫜒", "𪎋", "𫜔", "𪑅", "𫜙", "𫜨", "𫜳", "丢", "并", "干", "乱", "亘", "亚", "伫", "布", "占", "并", "来", "仑", "侣", "局", "俣", "系", "𠇹", "伣", "侠", "伡", "私", "伥", "俩", "俫", "仓", "个", "们", "幸", "伦", "㑈", "伟", "㐽", "侧", "侦", "伪", "㐷", "杰", "伧", "伞", "备", "家", "佣", "偬", "传", "伛", "债", "伤", "倾", "偻", "仅", "佥", "侨", "仆", "伪", "侥", "偾", "雇", "价", "仪", "俊", "侬", "亿", "侩", "俭", "傤", "傧", "俦", "侪", "尽", "偿", "𠆲", "优", "𠋆", "储", "俪", "㑩", "傩", "傥", "俨", "凶", "兑", "儿", "兖", "内", "两", "册", "胄", "幂", "净", "冻", "𪞝", "凛", "凯", "别", "删", "刭", "则", "克", "刹", "刬", "刚", "剥", "剐", "剀", "创", "铲", "𠛅", "划", "剧", "刘", "刽", "刿", "剑", "㓥", "剂", "㔉", "劲", "𠡠", "动", "务", "勋", "胜", "劳", "势", "𪟝", "勚", "劢", "勋", "励", "劝", "匀", "匦", "汇", "匮", "区", "协", "恤", "却", "即", "厍", "厕", "历", "厌", "厉", "厣", "参", "叁", "丛", "咤", "吴", "呐", "吕", "呙", "员", "𠯟", "呗", "𪠳", "吣", "念", "问", "启", "哑", "启", "唡", "㖞", "唤", "丧", "吃", "乔", "单", "哟", "呛", "啬", "唝", "吗", "呜", "唢", "𠮶", "哔", "𪡏", "叹", "喽", "啯", "呕", "啧", "尝", "唛", "哗", "𪡃", "唠", "啸", "叽", "𪡞", "哓", "呒", "𪡀", "啴", "恶", "𠯠", "嘘", "㖊", "咝", "𪡋", "哒", "哝", "哕", "嗳", "哙", "喷", "吨", "当", "咛", "吓", "哜", "尝", "噜", "啮", "𪠸", "咽", "呖", "𠰷", "咙", "向", "亸", "喾", "严", "嘤", "𪢕", "啭", "嗫", "嚣", "𠱞", "冁", "呓", "啰", "苏", "嘱", "𪢠", "囱", "囵", "国", "围", "园", "圆", "图", "团", "𪢮", "埯", "垭", "𪣆", "采", "执", "坚", "垩", "垴", "𪣒", "埚", "尧", "报", "场", "块", "茔", "垲", "埘", "涂", "冢", "坞", "埙", "尘", "堑", "𪣻", "垫", "坠", "堕", "坛", "𪢸", "坟", "垯", "墙", "垦", "坛", "𡒄", "垱", "埙", "压", "𡋤", "垒", "圹", "垆", "坛", "坏", "垄", "垅", "坜", "𪤚", "坝", "塆", "壮", "壶", "壸", "寿", "够", "梦", "伙", "夹", "奂", "奥", "奁", "夺", "奖", "奋", "姹", "妆", "姗", "奸", "娱", "娄", "𫝫", "妇", "娅", "𫝨", "娲", "妫", "㛀", "媪", "妈", "袅", "妪", "妩", "娴", "娴", "婳", "妫", "媭", "𫝬", "娆", "婵", "娇", "嫱", "嫒", "𪥰", "嬷", "𫝩", "嫔", "婴", "婶", "𪥿", "娘", "𫝮", "𫝭", "𪥫", "㛤", "娈", "𡠟", "孙", "学", "𡥧", "𪧀", "孪", "宫", "采", "𪧘", "寝", "实", "宁", "审", "写", "宽", "宠", "宝", "将", "专", "寻", "对", "导", "尴", "届", "尸", "屃", "屉", "屡", "层", "屦", "𪨗", "属", "冈", "峰", "岘", "岛", "峡", "崃", "昆", "岗", "仑", "峥", "岽", "岚", "岁", "𡶴", "㟥", "嵝", "崭", "岖", "𡺃", "嵚", "崂", "𡺄", "峤", "峣", "峄", "峃", "崄", "岙", "嵘", "𫝵", "岭", "屿", "岳", "𪩎", "岿", "峦", "巅", "岩", "𪨷", "𪩘", "巯", "卺", "帅", "师", "帐", "带", "帧", "帏", "㡎", "帼", "帻", "𪩷", "帜", "币", "𪩸", "帮", "帱", "干", "几", "库", "厕", "厢", "厩", "厦", "庼", "荫", "厨", "厮", "庙", "厂", "庑", "废", "广", "𪪞", "廪", "庐", "厅", "弑", "吊", "弪", "张", "强", "𪪼", "别", "弹", "弥", "弯", "录", "汇", "彝", "彟", "彦", "雕", "彨", "彷", "佛", "后", "径", "从", "徕", "复", "征", "彻", "𪫌", "恒", "耻", "悦", "悮", "怅", "闷", "凄", "恶", "恼", "恽", "恻", "爱", "惬", "悫", "怆", "恺", "𢙏", "忾", "栗", "态", "愠", "惨", "惭", "恸", "惯", "悫", "怄", "怂", "虑", "悭", "庆", "㥪", "戚", "欲", "忧", "惫", "怜", "凭", "愦", "慭", "惮", "𢙒", "愤", "悯", "怃", "宪", "忆", "𪫺", "𢙐", "𢙓", "恳", "应", "怿", "懔", "𢠁", "蒙", "怼", "懑", "㤽", "恹", "惩", "懒", "怀", "悬", "忏", "惧", "慑", "恋", "戆", "戋", "戗", "戬", "战", "戯", "戏", "户", "抛", "捝", "挲", "挟", "舍", "扪", "挨", "卷", "扫", "抡", "㧏", "挜", "挣", "𪭵", "挂", "采", "拣", "扬", "换", "挥", "搄", "损", "摇", "捣", "揾", "抢", "𢫬", "𪭢", "掴", "掼", "搂", "挚", "抠", "抟", "折", "掺", "捞", "𪭾", "挦", "撑", "挠", "㧑", "挢", "掸", "拨", "𪮖", "抚", "扑", "揿", "挞", "挝", "捡", "拥", "掳", "择", "击", "挡", "㧟", "担", "据", "𪭧", "挤", "抬", "捣", "𢬍", "拟", "摈", "拧", "搁", "掷", "扩", "撷", "摆", "擞", "撸", "㧰", "扰", "摅", "撵", "𪮶", "拢", "拦", "撄", "搀", "撺", "携", "摄", "攒", "挛", "摊", "搅", "揽", "教", "敚", "败", "叙", "敌", "数", "敛", "毙", "𢽾", "敩", "斓", "斩", "断", "𣃁", "于", "旗", "既", "升", "时", "晋", "昼", "晕", "晖", "旸", "畅", "暂", "晔", "历", "昙", "晓", "𪰶", "向", "暧", "旷", "𣆐", "昽", "晒", "书", "会", "𦛨", "胧", "术", "东", "锨", "拐", "栅", "拐", "查", "𣐕", "杆", "栀", "𪱷", "枧", "条", "枭", "棁", "弃", "棋", "枨", "枣", "栋", "㭎", "栈", "栖", "梾", "桠", "㭏", "𣒌", "杨", "枫", "桢", "业", "极", "矩", "干", "杩", "荣", "榅", "桤", "构", "枪", "杠", "梿", "椠", "椁", "𣏢", "椮", "桨", "椢", "椝", "桩", "乐", "枞", "梁", "楼", "标", "枢", "𣗊", "㭤", "样", "𣔌", "榝", "㭴", "桪", "朴", "树", "桦", "椫", "桡", "桥", "机", "椭", "横", "𣓿", "檩", "柽", "档", "桧", "槚", "检", "樯", "𣘴", "梼", "台", "槟", "𪲛", "柠", "槛", "𦼖", "柜", "𪲎", "橹", "榈", "栉", "椟", "橼", "栎", "𪲮", "橱", "槠", "栌", "枥", "橥", "榇", "蘖", "栊", "榉", "棂", "樱", "栏", "榉", "𪳍", "权", "𣐤", "椤", "𪲔", "𪴙", "栾", "𣗋", "榄", "𣚚", "棂", "钦", "叹", "欧", "欤", "欢", "岁", "历", "归", "殁", "残", "殒", "𣨼", "殇", "㱮", "殚", "僵", "殓", "殡", "㱩", "歼", "杀", "壳", "壳", "毁", "殴", "𪵑", "毵", "牦", "毡", "氇", "气", "氢", "氩", "𣱝", "氲", "泛", "泛", "污", "决", "沈", "没", "冲", "况", "溯", "泄", "汹", "浃", "泾", "涚", "凉", "凄", "泪", "渌", "净", "凌", "沦", "渊", "涞", "浅", "涣", "减", "沨", "涡", "测", "浑", "凑", "𣲗", "浈", "涌", "汤", "沩", "准", "沟", "𪶄", "温", "浉", "涢", "湿", "沧", "灭", "涤", "荥", "汇", "沪", "滞", "渗", "卤", "浒", "浐", "滚", "满", "渔", "溇", "沤", "汉", "涟", "渍", "涨", "溆", "渐", "浆", "颍", "泼", "洁", "𣲘", "沩", "㴋", "潜", "𫞗", "润", "浔", "溃", "滗", "涠", "涩", "𣶩", "浇", "涝", "沄", "涧", "渑", "泽", "滪", "泶", "𫞚", "浍", "淀", "㳠", "浊", "浓", "㳡", "𣸣", "湿", "泞", "溁", "蒙", "浕", "济", "涛", "㳔", "滥", "潍", "滨", "溅", "泺", "滤", "𪵱", "澛", "𣽷", "滢", "渎", "㲿", "泻", "沈", "浏", "濒", "泸", "沥", "潇", "潆", "潴", "泷", "濑", "弥", "潋", "澜", "沣", "滠", "𫞝", "洒", "𪷽", "漓", "滩", "𣺼", "灏", "㳕", "湾", "滦", "滟", "滟", "灾", "为", "乌", "烃", "无", "𪸩", "炼", "炜", "烟", "茕", "焕", "烦", "炀", "㶽", "𪸕", "煴", "𤈶", "𤇄", "荧", "𤆡", "炝", "𤇹", "𤋏", "热", "颎", "炽", "烨", "灯", "炖", "烧", "烫", "焖", "营", "灿", "毁", "烛", "烩", "㶶", "熏", "烬", "焘", "𫞡", "𤇃", "𦶟", "烁", "炉", "𤇭", "烂", "𪹳", "𫞠", "争", "为", "爷", "尔", "床", "墙", "牍", "牴", "牵", "荦", "牦", "𪺭", "犊", "牺", "状", "狭", "狈", "𪺽", "狰", "犹", "狲", "犸", "呆", "狱", "狮", "𪺷", "奖", "独", "𤞃", "狯", "猃", "狝", "狞", "㺍", "获", "猎", "犷", "兽", "獭", "献", "猕", "猡", "𤞤", "𫞥", "现", "雕", "珐", "珲", "玮", "玚", "琐", "瑶", "莹", "玛", "玱", "𪻲", "𪻐", "琏", "𫞩", "𪻺", "琎", "玑", "瑷", "珰", "㻅", "环", "玙", "瑸", "𫞨", "玺", "𫞦", "𪻨", "琼", "珑", "璎", "𤦀", "瓒", "𤩽", "瓯", "瓮", "产", "产", "苏", "宁", "亩", "毕", "画", "异", "画", "当", "𪽈", "畴", "叠", "痉", "酸", "𪽪", "疴", "痖", "疯", "疡", "痪", "瘗", "疮", "疟", "瘆", "𪽷", "疭", "瘘", "瘘", "疗", "痨", "痫", "瘅", "𤶊", "愈", "疠", "瘪", "痴", "痒", "疖", "症", "疬", "癞", "癣", "瘿", "瘾", "痈", "瘫", "癫", "发", "皂", "皑", "𤾀", "疱", "皲", "皱", "杯", "盗", "盏", "尽", "监", "盘", "卢", "𪾔", "荡", "𪾣", "真", "眦", "众", "𪾢", "困", "睁", "睐", "睾", "眍", "䁖", "瞒", "𥆧", "瞭", "瞆", "睑", "蒙", "𪾸", "𪾦", "眬", "瞩", "矫", "朱", "硁", "硖", "砗", "砚", "埼", "𥐻", "硕", "砀", "砜", "确", "码", "䂵", "硙", "砖", "硵", "碜", "碛", "矶", "硗", "䃅", "硚", "硷", "础", "𥐟", "碍", "矿", "砺", "砾", "矾", "𪿫", "砻", "祇", "秘", "禄", "祸", "祯", "祎", "祃", "御", "禅", "礼", "祢", "祷", "秃", "籼", "税", "秆", "䅉", "棱", "禀", "种", "称", "谷", "䅟", "稣", "积", "颖", "秾", "穑", "秽", "稳", "获", "穞", "窝", "洼", "穷", "窑", "窎", "窭", "窥", "窜", "窍", "窦", "灶", "窃", "𥩟", "竖", "𫁟", "竞", "笔", "笋", "笕", "䇲", "个", "笺", "筝", "节", "范", "筑", "箧", "筼", "𥬠", "筿", "笃", "筛", "筚", "𥮾", "箦", "𫂆", "篓", "蓑", "箪", "简", "𫂃", "篑", "箫", "筜", "签", "帘", "篮", "𥫣", "𥬞", "筹", "䉤", "箓", "篯", "箨", "籁", "笼", "签", "笾", "簖", "篱", "箩", "吁", "粤", "粽", "糁", "粪", "粮", "团", "粝", "籴", "粜", "纟", "𫄙", "纠", "纪", "纣", "约", "红", "纡", "纥", "纨", "纫", "纹", "纳", "纽", "纾", "纯", "纰", "纼", "纱", "纮", "纸", "级", "纷", "纭", "纴", "𫄛", "纺", "䌷", "扎", "细", "绂", "绁", "绅", "纻", "绍", "绀", "绋", "绐", "绌", "𫄟", "终", "弦", "组", "䌹", "绊", "𫟃", "绗", "结", "绝", "𫄠", "绦", "绔", "绞", "络", "绚", "𫄢", "给", "𫄡", "绒", "绖", "统", "丝", "绛", "绝", "绢", "𫄨", "𦈌", "绑", "绡", "绠", "𦈋", "绨", "绣", "𫟄", "绤", "绥", "䌼", "捆", "经", "𫄧", "综", "缍", "𫄫", "绿", "𫟅", "绸", "绻", "线", "绶", "维", "绹", "绾", "纲", "网", "绷", "缀", "彩", "纶", "绺", "绮", "绽", "绰", "绫", "绵", "绲", "缁", "紧", "绯", "𦈏", "绿", "绪", "绬", "绱", "缃", "缄", "缂", "线", "缉", "缎", "𫟆", "缔", "缗", "缘", "𫄬", "缌", "编", "缓", "缅", "𫄭", "纬", "𦈕", "缑", "缈", "练", "缏", "𦈉", "𦈑", "缇", "致", "缊", "萦", "缙", "缢", "缒", "𫄰", "𦈔", "绉", "缣", "缊", "缞", "缚", "缜", "缟", "缛", "县", "绦", "缝", "𦈚", "缡", "缩", "𫄳", "纵", "缧", "䌸", "纤", "缦", "絷", "缕", "𫄲", "缥", "𦈐", "总", "绩", "𫄴", "绷", "缫", "缪", "𫄶", "𦈝", "穗", "缯", "𦈛", "织", "缮", "缭", "绕", "𦈎", "绣", "缋", "𫄤", "绳", "绘", "系", "𫄱", "茧", "缰", "缳", "缲", "缴", "𫄷", "𫄣", "䍁", "绎", "𦈡", "继", "缤", "缱", "䍀", "𫄸", "颣", "缬", "纩", "续", "累", "缠", "缨", "才", "纤", "𫄹", "缵", "𫄥", "缆", "钵", "䓨", "坛", "罂", "坛", "罚", "骂", "罢", "罗", "罴", "羁", "芈", "群", "羟", "羡", "义", "𫅗", "膻", "习", "翚", "翘", "翙", "耧", "耢", "圣", "闻", "联", "聪", "声", "耸", "聩", "聂", "职", "聍", "𫆏", "听", "聋", "肃", "胁", "脉", "胫", "唇", "𣍰", "修", "脱", "胀", "肾", "胨", "脶", "脑", "𣍯", "肿", "脚", "肠", "腽", "腘", "肤", "䏝", "胶", "𦝼", "腻", "𪱥", "胆", "脍", "脓", "脸", "脐", "膑", "𣎑", "腊", "胪", "脏", "脔", "臜", "卧", "临", "台", "与", "兴", "举", "旧", "馆", "舱", "𫇛", "舣", "舰", "舻", "艰", "艳", "刍", "苎", "兹", "荆", "庄", "茎", "荚", "苋", "芲", "华", "庵", "烟", "苌", "莱", "万", "荝", "莴", "叶", "荭", "𫈎", "荮", "苇", "药", "荤", "𫇭", "搜", "莼", "莳", "蒀", "莅", "𫇴", "苍", "荪", "席", "盖", "𦰏", "莲", "苁", "莼", "荜", "卜", "参", "蒌", "蒋", "葱", "茑", "荫", "𫈟", "𫇭", "荨", "蒇", "荞", "荬", "芸", "莸", "荛", "𫈵", "蒉", "荡", "芜", "萧", "𫈉", "蓣", "𫇽", "蕰", "𫉁", "荟", "蓟", "芗", "姜", "蔷", "荙", "莶", "荐", "萨", "薰", "䓕", "苧", "䓓", "苔", "荠", "藉", "蓝", "荩", "艺", "药", "薮", "䓖", "蕴", "苈", "𫉄", "蔼", "蔺", "萚", "蕲", "芦", "苏", "蕴", "苹", "藓", "蔹", "𦻕", "茏", "兰", "蓠", "萝", "蔂", "处", "虚", "虏", "号", "亏", "虬", "蛱", "蜕", "蚬", "蚀", "猬", "虾", "虱", "蜗", "蛳", "蚂", "萤", "䗖", "蝼", "螀", "𫋇", "蛰", "蝈", "螨", "𫋌", "𫊸", "虮", "蝉", "蛲", "虫", "𫊻", "蛏", "蚁", "𧏗", "蚃", "蝇", "虿", "蝎", "蛴", "蝾", "蚝", "𧏖", "蜡", "蛎", "𫊮", "蟏", "蛊", "蚕", "蛮", "𧑏", "众", "蔑", "术", "同", "胡", "卫", "冲", "衹", "衮", "袅", "里", "补", "装", "里", "制", "复", "裈", "袆", "裤", "裢", "褛", "亵", "𫌀", "幞", "裥", "裥", "袯", "𫋹", "袄", "𫋷", "𫋻", "裣", "裆", "褴", "袜", "摆", "衬", "𧝝", "袭", "襕", "𫌇", "覆", "核", "见", "觃", "规", "觅", "视", "觇", "𫌪", "觋", "觍", "觎", "亲", "觊", "觏", "觐", "觑", "𫌭", "觉", "𫌨", "览", "觌", "观", "觞", "觯", "触", "讠", "订", "讣", "计", "讯", "讧", "讨", "讦", "𫍙", "讱", "训", "讪", "讫", "托", "记", "讹", "𫍛", "讶", "𫍚", "讼", "䜣", "诀", "讷", "𫟞", "讻", "访", "设", "许", "诉", "诃", "诊", "注", "证", "𧮪", "诂", "诋", "𫟟", "讵", "诈", "𫍡", "诒", "𫍜", "诏", "评", "诐", "诇", "诎", "诅", "词", "咏", "诩", "询", "诣", "试", "诗", "诧", "诟", "诡", "诠", "诘", "话", "该", "详", "诜", "𫍣", "诙", "诖", "𫍥", "诔", "诛", "诓", "夸", "𫍪", "志", "认", "诳", "诶", "诞", "诱", "诮", "语", "诚", "诫", "诬", "误", "诰", "诵", "诲", "说", "𫍨", "说", "谁", "课", "𫍮", "𫟡", "谇", "𫍬", "诽", "𫍧", "谊", "訚", "调", "谄", "谆", "谈", "诿", "请", "诤", "诹", "诼", "谅", "论", "谂", "谀", "谍", "谞", "谝", "谥", "诨", "𫍩", "谔", "𫍳", "谛", "谐", "谏", "谕", "咨", "𫍱", "𫍰", "讳", "谙", "𫍯", "谌", "讽", "诸", "谚", "谖", "诺", "谋", "谒", "谓", "誊", "诌", "𫍸", "𫍷", "谎", "谜", "𫍲", "谧", "谑", "谡", "谤", "谦", "谥", "讲", "谢", "谣", "谣", "谟", "谪", "谬", "谫", "𫍹", "𫍴", "讴", "𫍵", "谨", "谩", "哗", "𫟠", "䜧", "𫍻", "证", "𫍢", "谲", "讥", "𫍤", "谮", "识", "谯", "谭", "谱", "𫍽", "噪", "𫍦", "谵", "毁", "译", "议", "谴", "护", "诪", "誉", "谫", "读", "谉", "变", "詟", "䜩", "雠", "谗", "让", "谰", "谶", "赞", "谠", "谳", "岂", "竖", "丰", "艳", "猪", "𫎆", "豮", "猫", "𫎌", "䝙", "贝", "贞", "贠", "负", "财", "贡", "贫", "货", "贩", "贪", "贯", "责", "贮", "贳", "赀", "贰", "贵", "贬", "买", "贷", "贶", "费", "贴", "贻", "贸", "贺", "贲", "赂", "赁", "贿", "赅", "资", "贾", "贼", "赈", "赊", "宾", "赇", "赒", "赉", "赐", "𫎩", "赏", "𧹖", "赔", "赓", "贤", "卖", "贱", "赋", "赕", "质", "赍", "账", "赌", "䞐", "赖", "赗", "赚", "赙", "购", "赛", "赜", "𧹗", "贽", "赘", "赟", "赠", "𫎫", "赞", "赝", "赡", "赢", "赆", "𫎬", "赃", "赑", "赎", "赝", "𫎦", "赣", "赃", "赪", "赶", "赵", "趋", "趱", "迹", "践", "逾", "踊", "跄", "𫏐", "跸", "迹", "蹒", "踪", "𫏆", "跷", "𫏋", "跶", "趸", "踌", "跻", "跃", "䟢", "踯", "跞", "踬", "蹰", "𨀁", "跹", "𨅬", "蹑", "蹿", "躜", "躏", "躯", "𨉗", "车", "轧", "轨", "军", "𫐄", "轪", "轩", "轫", "𫐅", "𨐅", "轭", "𫐇", "软", "轷", "𫐉", "轸", "𫐊", "轱", "𫐈", "轴", "轵", "轺", "轲", "轶", "轼", "𫐌", "较", "𨐈", "辂", "辁", "辀", "载", "轾", "𪨶", "辄", "挽", "辅", "轻", "𫐏", "𫐐", "辆", "辎", "辉", "辋", "辍", "𫐎", "辊", "辇", "𫐑", "辈", "轮", "辌", "𫐓", "辑", "辏", "𫐒", "输", "辐", "辒", "辗", "舆", "辒", "毂", "辖", "辕", "辘", "𫐖", "转", "𫐕", "辙", "轿", "𫐗", "辚", "𫐘", "轰", "𫐙", "辔", "轹", "𫐆", "轳", "办", "辞", "辫", "辩", "农", "回", "迳", "这", "连", "周", "进", "游", "运", "过", "达", "违", "遥", "逊", "递", "远", "溯", "适", "𫐷", "迟", "迁", "选", "遗", "辽", "迈", "还", "迩", "边", "逻", "逦", "郏", "邮", "郓", "乡", "邹", "邬", "郧", "𫑘", "邓", "郑", "邻", "郸", "𫑡", "邺", "郐", "邝", "酂", "郦", "腌", "酝", "丑", "酝", "蒏", "糖", "医", "酱", "酦", "𫑷", "酿", "衅", "酾", "酽", "释", "厘", "钅", "钆", "钇", "钌", "钊", "钉", "钋", "𫟲", "针", "𫓥", "钓", "钐", "扣", "钏", "𫓦", "钒", "𫟳", "𨰿", "钗", "钍", "钕", "钎", "䥺", "钯", "钫", "钘", "钭", "钥", "𫓪", "𫓧", "钚", "钠", "𨱂", "钝", "钩", "钤", "钣", "钑", "钞", "钮", "𫟴", "𫟵", "𫓨", "钧", "𨱁", "钟", "钙", "钬", "钛", "钪", "铌", "𨱄", "铈", "𨱃", "钶", "铃", "钴", "钹", "铍", "钰", "钸", "铀", "钿", "钾", "𨱅", "巨", "钻", "铊", "铉", "铇", "铋", "铂", "𫓬", "钷", "钳", "铆", "铅", "𫟷", "钺", "𫓭", "钵", "钩", "钲", "钼", "钽", "锫", "铏", "𫟹", "铰", "铒", "铬", "𫟸", "𫓴", "铪", "银", "𫓲", "𫟻", "铳", "铜", "𫓯", "𫓰", "铚", "𫟶", "铣", "铨", "铢", "铭", "铫", "铦", "衔", "铑", "铷", "铱", "铟", "铵", "铥", "铕", "铯", "铐", "铞", "锐", "𨱇", "销", "锈", "锑", "锉", "铝", "镅", "锒", "锌", "钡", "𨱈", "铤", "铗", "锋", "𫓶", "铻", "锊", "锓", "𫓵", "铘", "锄", "锃", "锔", "锇", "铓", "铺", "锐", "铖", "锆", "锂", "铽", "锍", "锯", "钢", "𬬭", "锞", "𨱋", "录", "锖", "锫", "锩", "铔", "锥", "锕", "锟", "锤", "锱", "铮", "锛", "𫓻", "𫓽", "锬", "锭", "锜", "钱", "𫓹", "𫓾", "锦", "锚", "锠", "锡", "锢", "错", "录", "锰", "表", "铼", "镎", "𫓸", "锝", "锨", "锪", "𨱉", "钫", "钔", "锴", "锳", "𫔂", "炼", "锅", "镀", "𫔄", "锷", "铡", "钖", "锻", "锽", "锸", "锲", "锘", "锹", "𨱎", "锾", "键", "锶", "锗", "针", "钟", "镁", "锿", "镅", "𫟿", "镑", "镰", "𫔅", "镕", "锁", "镉", "𫔈", "锤", "镈", "𨱏", "𫔇", "镃", "钨", "蓥", "镏", "铠", "铩", "锼", "镐", "镇", "镇", "𨱍", "镒", "镋", "镍", "镓", "鿔", "𨰾", "镌", "镎", "镞", "𨱌", "镟", "链", "𨱒", "镆", "镙", "镠", "镝", "铿", "锵", "戚", "镗", "镘", "镛", "铲", "镜", "镖", "镂", "𫔊", "𫓩", "錾", "镚", "铧", "镤", "镪", "䥽", "锈", "𫔌", "铙", "𨱑", "𫔍", "𫓱", "铴", "𫔎", "𨱓", "𨱔", "镣", "铹", "镦", "镡", "锏", "钟", "镫", "镢", "镨", "䦅", "锎", "锏", "镄", "𫓺", "镌", "镰", "䦃", "镯", "镭", "铁", "镮", "铎", "铛", "𫔁", "𫟼", "镱", "锿", "铸", "𫠁", "镬", "镔", "鉴", "鉴", "镲", "锧", "镴", "铄", "镳", "镥", "𬬻", "镧", "钥", "镵", "镶", "𫔔", "镊", "镩", "锣", "钻", "銮", "凿", "镢", "镋", "旋", "长", "门", "闩", "闪", "闫", "闬", "闭", "开", "闶", "𨸂", "闳", "闰", "𨸃", "闲", "闲", "间", "闵", "𫔯", "闸", "𫠂", "𫔰", "阂", "阁", "合", "阀", "闺", "闽", "阃", "阆", "闾", "阅", "阅", "𫔴", "阊", "阉", "阎", "阏", "阍", "阈", "阌", "阒", "板", "暗", "闱", "阔", "阕", "阑", "阇", "阗", "𫔶", "阘", "闿", "阖", "阙", "闯", "关", "阚", "阓", "阐", "辟", "阛", "闼", "阪", "陉", "陕", "升", "阵", "阴", "陈", "陆", "阳", "陧", "队", "阶", "陨", "际", "随", "险", "陦", "隐", "陇", "隶", "只", "隽", "虽", "双", "雏", "杂", "鸡", "离", "难", "云", "电", "霡", "𫕥", "雾", "𪵣", "霁", "雳", "霭", "叇", "灵", "叆", "靓", "静", "靔", "腼", "𫖃", "靥", "鼗", "巩", "绱", "秋", "鞒", "𫖇", "缰", "鞑", "千", "鞯", "韦", "韧", "韨", "韩", "韪", "𫠅", "𫖔", "韬", "鞲", "韫", "𫖒", "韵", "响", "页", "顶", "顷", "项", "顺", "顸", "须", "顼", "颂", "𫠆", "颀", "颃", "预", "顽", "颁", "顿", "颇", "领", "颌", "颉", "颐", "颏", "𫖯", "头", "颒", "颊", "颋", "颕", "𫖳", "颔", "颈", "颓", "频", "颓", "𩓋", "𩖖", "𫖶", "颗", "题", "额", "颚", "颜", "颙", "颛", "颜", "𫖮", "愿", "颡", "颠", "类", "颟", "𫖹", "颢", "顾", "颤", "颥", "显", "颦", "颅", "颞", "颧", "风", "飐", "飑", "飒", "𩙥", "台", "刮", "飓", "𩙪", "飔", "飏", "飖", "飕", "𩙫", "飗", "飘", "飙", "飚", "𫗋", "飞", "饣", "饥", "饤", "饦", "𫗞", "饨", "饪", "饫", "饬", "饭", "飧", "饮", "饴", "𫗢", "𫗣", "饲", "饱", "饰", "饳", "饺", "饸", "饼", "饷", "养", "饵", "饹", "饻", "饽", "馁", "饿", "𫗦", "馂", "饾", "𫗧", "余", "肴", "馄", "馃", "饯", "馅", "𫗠", "𫗪", "馆", "𫗬", "𫗥", "糊", "𫗮", "糇", "饧", "喂", "馉", "馇", "𩠌", "馎", "饩", "馏", "馊", "馌", "馍", "馒", "馐", "馑", "馓", "馈", "馔", "饥", "饶", "飨", "𫗴", "餍", "馋", "𫗵", "𫗩", "馕", "马", "驭", "冯", "𫘛", "驮", "驰", "驯", "驲", "𫘜", "驳", "𫘝", "𫘟", "𩧨", "驻", "驽", "驹", "驵", "驾", "骀", "驸", "𩧫", "驶", "驼", "𫘞", "驷", "骂", "骈", "𫘠", "𩧲", "𩧴", "𫘡", "骇", "骃", "骆", "𩧺", "骎", "𫘣", "骏", "骋", "骍", "𫘤", "𫘧", "骓", "𫘥", "𫘦", "骔", "骒", "骑", "骐", "𩨀", "骛", "骗", "𩨊", "𫘩", "𩨃", "𩨈", "𫘨", "骙", "䯄", "𩨄", "骞", "骘", "骝", "腾", "𫘬", "𫘫", "𫘪", "驺", "骚", "骟", "𫘭", "𫠋", "骡", "蓦", "骜", "骖", "骠", "骢", "驱", "骅", "𩧯", "骕", "骁", "骣", "𫘯", "骄", "验", "𫘰", "惊", "驿", "骤", "驴", "骧", "骥", "骦", "𫘱", "骊", "骉", "肮", "髅", "脏", "体", "髌", "髋", "发", "松", "胡", "𩭹", "须", "𫘽", "鬓", "斗", "闹", "哄", "阋", "阄", "郁", "鬶", "魉", "魇", "鱼", "鱽", "𫚉", "鱾", "𩽹", "𫚌", "鲀", "鲁", "鲂", "𫚍", "鱿", "鲄", "𫠐", "鲅", "鲆", "𫚒", "𫚑", "𫚖", "鲌", "鲉", "鲏", "鲇", "鲐", "鲍", "鲋", "鲊", "𩾀", "鲒", "鲘", "鲞", "鲕", "𩽾", "䲟", "𫚓", "鲖", "鲔", "鲛", "鲑", "鲜", "𫚗", "𫚔", "鲓", "𫚛", "鲪", "𩾃", "鲝", "𫚚", "鲧", "鲠", "𩾁", "𫚙", "鲩", "鲤", "鲨", "鲬", "鲻", "鲯", "鲭", "鲞", "鲷", "鲴", "𫚡", "鲱", "鲵", "鲲", "鲳", "鲸", "鲮", "鲰", "𫚞", "鲶", "𩾇", "鲺", "𩽼", "鳀", "鲫", "𫚣", "鳊", "鳈", "鲗", "鳂", "䲠", "鲽", "鳇", "𫚢", "䲡", "鳅", "鲾", "鳄", "𫚊", "鳆", "鳃", "𫚥", "鳁", "鳒", "鳑", "鳋", "鲥", "𫚕", "鳏", "𫚤", "䲢", "鳎", "鳐", "𫚦", "鳍", "鳁", "鲢", "鳌", "鳓", "鳘", "鲦", "鲣", "鲹", "鳗", "鳛", "𫚧", "鳔", "鳉", "𫚋", "鳙", "𫠒", "𩾌", "鳕", "鳖", "𫚪", "鳟", "鳝", "鳜", "鳞", "鲟", "鲼", "鲎", "鲙", "𫚫", "鳣", "鳡", "鳢", "鲿", "鲚", "𫚈", "鳠", "𫚭", "鳄", "鲈", "鲡", "鸟", "凫", "鸠", "凫", "鸤", "凤", "鸣", "鸢", "𫛛", "𪉃", "𫛚", "䴓", "𫛜", "𫛞", "𫛝", "鸩", "鸨", "鸦", "𫛤", "鸰", "𫛡", "鸵", "𫁡", "鸳", "𪉈", "鸲", "鸮", "鸱", "鸪", "𫛣", "鸯", "鸭", "𫛦", "鸸", "鸹", "𪉆", "𫛩", "鸻", "䴕", "鸿", "𫛪", "鸽", "䴔", "鸺", "鸼", "𫛥", "鹀", "鹃", "鹆", "鹁", "𪉍", "鹈", "鹅", "𫛭", "鹄", "鹉", "𫛨", "𫛳", "鹌", "𫛱", "鹏", "鹐", "鹎", "雕", "鹊", "鹓", "鹍", "䴖", "鸫", "鹑", "鹒", "𫛵", "𫛶", "鹋", "鹙", "𫛸", "鹕", "鹗", "鹖", "鹛", "𫛷", "鹜", "䴗", "鸧", "𫛯", "莺", "𫛫", "鹟", "鹤", "鹠", "鹡", "鹘", "鹣", "鹚", "鹚", "鹢", "鹞", "鸡", "𫛽", "䴘", "䴘", "鹝", "𫜀", "鹧", "𪉑", "鹥", "鸥", "鸷", "鹨", "𫜃", "𫛴", "鸶", "鹪", "𪉊", "𫜁", "鹔", "鹩", "鹫", "鹇", "鹇", "𫜄", "鹬", "鹰", "鹭", "鸴", "䴙", "㶉", "鹯", "䴙", "𫛢", "鹱", "鹲", "鸬", "𫛟", "鹴", "鹦", "鹳", "鹂", "鸾", "卤", "咸", "鹾", "碱", "盐", "丽", "麦", "𪎊", "麸", "面", "面", "𤿲", "曲", "𪎉", "𪎌", "曲", "面", "𫜑", "么", "么", "黄", "黉", "点", "党", "黪", "霉", "黡", "黩", "黾", "鼋", "鼌", "鼍", "冬", "鼹", "齄", "齐", "斋", "赍", "齑", "齿", "龀", "龁", "龂", "龅", "龇", "龃", "龆", "龄", "出", "龈", "啮", "𫜪", "龊", "龉", "𫜭", "𫠜", "𫜬", "龋", "𫜮", "腭", "龌", "𫜰", "龙", "厐", "庞", "䶮", "𫜲", "龚", "龛", "龟", "𩨎", "𨱆", "䜤", "鿒", "𠀾", "𠆿", "𠉗", "𫝋", "㓆", "𠛆", "𠚳", "𪠡", "𪠺", "𪜎", "𪢒", "𪡺", "𠴛", "𪢐", "𠴢", "𠵸", "𠲥", "𪢖", "𡋗", "𪤄", "𡋀", "𡍣", "㛟", "𫝪", "㛿", "𡞱", "㛠", "𡭜", "𡭬", "𡳃", "𪨩", "𪨹", "岁", "𡸃", "㟜", "𪩛", "𪪴", "𢙑", "𪬚", "𢘝", "𢘞", "𪫡", "𢘙", "𪬯", "𪭝", "𪭯", "𢫞", "𢭏", "𢫊", "𢬦", "𪮳", "暅", "𣈣", "𠊉", "㭣", "𪳗", "𣘷", "𣘓", "𣞎", "𣑶", "𣯣", "𣭤", "毶", "𪶮", "㳢", "𣶫", "𣺽", "𪶒", "𣷷", "𤎻", "𪹀", "𤊀", "𪹹", "𪹠", "𤊰", "𪺣", "𤙯", "𫞢", "𪺪", "𪺸", "𤝢", "𢢐", "𫞧", "㻘", "㻏", "𪼴", "𪽝", "𤳄", "𪽭", "𤶧", "𪽴", "𥅿", "𥅘", "𪿊", "𥐰", "𥐯", "𪿞", "𪿵", "𫀓", "𫀌", "𥞦", "䅪", "𫞷", "𫀮", "𥧂", "𥩺", "𫁳", "𫂖", "𫁺", "𥱔", "𥭉", "𫁱", "𥮋", "𫂿", "𥹥", "𥺇", "𫄝", "𦈈", "𫄦", "𦈒", "𦈗", "𫄯", "𫄪", "𫄵", "𫟇", "𫅥", "𫅼", "𡳒", "𫆝", "𫞅", "𫆫", "𣍨", "𦟗", "𫇘", "䑽", "𦨩", "𫇪", "𦶻", "𧌥", "𫊹", "𧒭", "𧉐", "䘞", "䙊", "䘛", "𫌋", "𧝧", "𫌫", "𫌬", "𫍞", "𫍟", "𫍭", "䜥", "𫍶", "𫍺", "𫍼", "𫍾", "𫍐", "𧳕", "䞌", "𧹓", "䞎", "𪠀", "𫎨", "𪥠", "𫎸", "𫏌", "𨀱", "𨁴", "𫏕", "𧿈", "𨅫", "𫏨", "𫏞", "𫏑", "𨂺", "𨄄", "䢀", "䢁", "𨐆", "䢂", "𫐍", "𫐔", "𫐋", "𨐉", "𨐇", "𨐊", "𫟫", "𫟬", "𨡙", "𨡺", "𨟳", "𨠨", "𨤰", "𨱀", "𫓫", "䦀", "𬭊", "䦁", "𫟽", "𨱊", "𬭛", "𫓼", "𫓿", "𫟾", "𫓮", "𨱐", "𫔏", "𬭶", "𬭳", "𫔑", "𫔐", "𨱕", "𫔒", "䥿", "𫔓", "𫔉", "𫓳", "𫔕", "𫔃", "𫔖", "𨸁", "𨸀", "𨸅", "𫔲", "𨸆", "𨸇", "𨸉", "𨸊", "𨸌", "𨸋", "𨸎", "𫔽", "𨸘", "𫕚", "𫕨", "𫖑", "𩏾", "𫖓", "𫖖", "𩏽", "𫃗", "𫖪", "𫖭", "𩖕", "𫖵", "𫖷", "𫖴", "𫠇", "𩙦", "𫗈", "𫗉", "𩙩", "𩙭", "𩙨", "𩙬", "𩙰", "𩟿", "𩠀", "𫗡", "𩠁", "𩠂", "𫗤", "𫗨", "𩠃", "𩠉", "𩠆", "𩠊", "𩠋", "𫗳", "𩠎", "𩠏", "䭪", "𩠅", "𫗚", "𩠠", "𩡖", "𩧦", "𩧬", "𩧵", "𩧳", "𩧮", "𩧶", "䯃", "𩧸", "𩧻", "𩧼", "𩧩", "𩨆", "𩨉", "𩨅", "𩨋", "𩨍", "𩧱", "𩨌", "𫠌", "𩨐", "𩬣", "𫙂", "𩯒", "𩬤", "𩰰", "𩲒", "𩴌", "𫠏", "𩽺", "𩽻", "𫚎", "䲞", "𩽿", "𩽽", "𩾄", "𩾅", "𫚝", "𫚟", "𩾆", "𫚨", "𫚩", "𫚘", "𫚬", "𩾎", "𫠖", "𫛠", "𪉄", "𫛧", "𪉅", "𪉋", "𪉉", "𪉌", "𪉎", "𪉐", "𪉏", "𫛻", "𫛹", "𪉔", "𪉒", "𫜂", "𫛾", "𪉕", "𪉓", "𫜊", "𪎍", "𫜓", "𫜕", "𫜟", "𪔭", "𪚏", "𪚐", "𫜯", "𠛾", "𣶭", "𫓷", "𫜫", "八濛山", "拜复", "毕昇", "变徵", "禀复", "朝乾夕惕", "筹划", "出丑狼借", "答复", "大目乾连冥间救母变文", "荡复", "抵触", "抵牾", "颠乾倒坤", "电复", "貂复额", "笃麽", "锻锺", "发复", "樊於期", "反反复复", "反复", "复按", "复杯", "复瓿", "复成", "复呈", "复帱", "复电", "复鼎", "复醢", "复酱瓿", "复蕉寻鹿", "复鹿寻蕉", "复鹿遗蕉", "复露", "复冒", "复命", "复墓", "复逆", "复上", "复文", "复校", "复信", "复盂", "复育", "复帐", "复住", "复宗", "顾借", "郭子乾", "函复", "狐借虎威", "滑借", "黄锺公", "回复", "借草枕块", "借词", "借此", "借代", "借读", "借端", "借方", "借槁", "借故", "借卉", "借机", "借借", "借据", "借口", "借寇兵", "借甚", "借手", "借条", "借以", "借由", "借喻", "借着", "借助", "借箸代筹", "借资", "计划", "家具", "见复", "角徵", "铰链", "金链", "金吒", "颈链", "酒逢知己千锺少", "钜防", "钜万", "钜子", "钧复", "康乾", "拉链", "李链福", "李乾德", "李泽钜", "李锺郁", "链锤", "链锁", "链条", "链形", "链坠", "链子", "了解", "了然", "了如", "了若指掌", "躏借", "凌借", "流徵", "露复", "买臣复水", "幺麽", "幺麽", "麽些族", "名复金瓯", "明复", "明了", "侔德复载", "木吒", "哪吒", "尼乾陀", "纽扣", "仿佛", "凭借", "颇复", "千锺粟", "乾旦", "乾道", "乾断", "乾纲", "乾卦", "乾红", "乾嘉", "乾坤", "乾陵", "乾隆", "乾清宫", "乾盛世", "乾图", "乾县", "乾象", "乾曜", "乾元", "乾造", "乾宅", "钱锺书", "伤亡枕借", "上链", "射复", "申复", "沉船", "沉积", "沉没", "沉默", "甚钜", "示复", "手链", "素借", "蒜薹", "锁链", "铁链", "万锺", "文锦复阱", "项链", "萧乾", "旋乾转坤", "踅门了户", "雁杳鱼沉", "幺麽", "以功复过", "茵借", "银链", "於敖", "於崇文", "於单", "於乎", "於呼哀哉", "於梨华", "於琳", "於陵子", "於伦", "於穆", "於其一", "於清言", "於世成", "於坦", "於菟", "於惟一", "於戏", "於邑", "於勇明", "於则", "於忠祥", "於仲完", "於竹屋", "芸薹", "酝借", "张法乾", "只得", "只见树木", "徵调", "徵声", "徵弦", "徵弦", "徵音", "锺锻", "锺馗", "锺万梅", "锺重发", "重复"];
    //-----------------------
    var s2tSearch = null; // WordsSearch
    var t2sSearch = null;// WordsSearch
    var t2twSearch = null;// WordsSearch
    var tw2tSearch = null;// WordsSearch
    var t2hkSearch = null;// WordsSearch
    var hk2tSearch = null;// WordsSearch

    /**
     * 转繁体中文
     * @param {any} text 原文本
     * @param {any} type 0、繁体中文，1、港澳繁体，2、台湾正体
     */
    this.ToTraditionalChinese = function (text, type = 0) {
        if (type > 2 || type < 0) { throw "type 不支持该类型"; }

        var s2t = GetWordsSearch(true, 0);
        text = TransformationReplace(text, s2t);
        if (type > 0) {
            var t2 = GetWordsSearch(true, type);
            text = TransformationReplace(text, t2);
        }
        return text;
    }

    /**
     * 转简体中文
     * @param {any} text 原文本
     * @param {any} srcType 0、繁体中文，1、港澳繁体，2、台湾正体
     */
    this.ToSimplifiedChinese = function (text, srcType = 0) {
        if (srcType > 2 || srcType < 0) { throw "srcType 不支持该类型"; }
        if (srcType > 0) {
            var t2 = GetWordsSearch(false, srcType);
            text = TransformationReplace(text, t2);
        }
        var s2t = GetWordsSearch(false, 0);
        text = TransformationReplace(text, s2t);
        return text;
    }

    function TransformationReplace(text, wordsSearch) {
        var ts = wordsSearch.FindAll(text);

        var sb = "";
        var index = 0;
        while (index < text.length) {
            var t = null;
            var max = -1;
            for (var i = 0; i < ts.length; i++) {
                var f = ts[i];
                if (f.Start == index && f.End > max) {
                    max = f.End;
                    t = f;
                }
            }

            if (t == null) {
                sb += text[index];
                index++;
            } else {
                sb += wordsSearch._others[t.Index]
                index = t.End + 1;
            }
        }
        return sb;
    }

    function GetWordsSearch(s2t, srcType) {
        if (s2t) {
            if (srcType == 0) {
                if (s2tSearch == null) {
                    s2tSearch = BuildWordsSearch(_s2t_s, _s2t_t);
                }
                return s2tSearch;
            } else if (srcType == 1) {
                if (t2hkSearch == null) {
                    t2hkSearch = BuildWordsSearch(_t2hk_t, _t2hk_hk);
                }
                return t2hkSearch;
            } else if (srcType == 2) {
                if (t2twSearch == null) {
                    t2twSearch = BuildWordsSearch(_t2tw_t, _t2tw_tw);
                }
                return t2twSearch;
            }
        }
        if (srcType == 0) {
            if (t2sSearch == null) {
                t2sSearch = BuildWordsSearch(_t2s_t, _t2s_s);
            }
            return t2sSearch;
        } else if (srcType == 1) {
            if (hk2tSearch == null) {
                hk2tSearch = BuildWordsSearch(_t2hk_hk, _t2hk_t);
            }
            return hk2tSearch;
        } else if (srcType == 2) {
            if (tw2tSearch == null) {
                tw2tSearch = BuildWordsSearch(_t2tw_tw, _t2tw_t);
            }
            return tw2tSearch;
        }
        return null;
    }
    function BuildWordsSearch(keywords, toWords) {
        var wordsSearch = new WordsSearch();
        wordsSearch.SetKeywords(keywords);
        wordsSearch._others = toWords;
        return wordsSearch;
    }
}